package com.icl.saxon;
import com.icl.saxon.om.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.PreviewManager;
import com.icl.saxon.ExtendedInputSource;
import com.icl.saxon.Context;
import com.icl.saxon.Controller;
import com.icl.saxon.ParserManager;
import com.icl.saxon.output.Emitter;
import com.icl.saxon.output.CharacterSet;
import com.icl.saxon.output.OutputDetails;

import org.xml.sax.*;
import org.xml.sax.helpers.ParserAdapter;

import java.util.*;
import java.io.*;
import java.net.URL;

/**
  * The abstract Builder class is responsible for taking a stream of SAX events
  * and constructing a Document tree. There is one concrete subclass for each
  * tree implementation.
  * @author Michael H. Kay (Michael.Kay@icl.com)
  */
  
public abstract class Builder implements Emitter, ErrorHandler, Locator

{

    public final static int STANDARD_TREE = 0;
    public final static int TINY_TREE = 1;

    //protected static int nextSequenceNr = 0;
    protected int estimatedLength;                // Estimated size of document in bytes

    protected Writer errorOutput = new PrintWriter(System.err);
                                                  // Destination for error messages
    protected Stripper stripper;                  // manages list of elements to be stripped
    protected PreviewManager previewManager = null;
    protected boolean discardComments;            // true if comments and PIs should be ignored
    protected int documentNumber = -1;
    protected Sequencer nodeSequencer = null;

    protected DocumentInfo currentDocument;
    protected XMLReader parser;                   // SAX2-compliant XML parser
    protected ErrorHandler errorHandler = this;   // SAX-compliant XML error handler
    protected Locator locator = this;             // SAX-compliant locator
    protected String baseURI = null;
    protected NamePool namePool = null;

    protected boolean failed = false;
    protected boolean started = false;
    protected boolean timing = false;

    protected boolean inDTD = false;
    protected boolean lineNumbering = false;
    //protected int nodeSequenceNr;
    
    private long startTime;

    /**
    * create a Builder and initialise variables
    */

    public Builder() {}

    /////////////////////////////////////////////////////////////////////////
    // Methods setting and getting options for building the tree
    /////////////////////////////////////////////////////////////////////////

    /**
    * Set timing option on or off
    */

    public void setTiming(boolean on) {
        timing = on;
    }

    /**
    * Get timing option
    */

    public boolean isTiming() {
        return timing;
    }

	/**
	* Set the namepool to be used for all names in the document
	*/
	
	public void setNamePool(NamePool pool) {
		namePool = pool;
	}
	
	/**
	* Get the namepool used for this document
	*/
	
	public NamePool getNamePool() {
		return namePool;
	}

    /**
    * Set the sequence number for the document
    */

    public void setDocumentNumber(int seq) {
        documentNumber = seq;
    }

    /**
    * Get the sequence number for the document
    */

    public int getDocumentNumber() {
        if (documentNumber < 0) {
            documentNumber = Sequencer.getDefaultDocumentSequencer().allocate(1);
        }
        return documentNumber;
    }

    /**
    * Set the sequencer used for allocating node numbers
    */

    public void setNodeSequencer(Sequencer seq) {
        nodeSequencer = seq;
    }

    /**
    * Get the sequencer used for allocating node numbers
    */

    public Sequencer getNodeSequencer() {
        if (nodeSequencer==null) {
            nodeSequencer = Sequencer.getDefaultNodeSequencer();
        }
        return nodeSequencer;
    }

    /**
    * Set the SAX1-compliant parser to use. If no parser is specified, the
    * parser is obtained from the ParserManager.properties file.
    * @param parser The XML parser to use. This must be an instance of a class
    * that implements the org.xml.sax.Parser interface.
    * @see ParserManager
    */

    public void setParser(Parser parser) {
        this.parser = new ParserAdapter(parser);
    }

    /**
    * Set the SAX2-compliant parser to use. If no parser is specified, the
    * parser is obtained from the ParserManager.properties file.
    * @param parser The XML parser to use. This must be an instance of a class
    * that implements the org.xml.sax.XMLReader interface.
    * @see ParserManager
    */

    public void setXMLReader(XMLReader parser) {
        this.parser = parser;
    }

    /**
    * Get the SAX2 parser in use.
    */

    public XMLReader getXMLReader() {
        return parser;
    }

    /**
    * Set line numbering on or off
    */

    public void setLineNumbering(boolean onOrOff) {
        lineNumbering = onOrOff;
    }

    /**
    * Set the Stripper to use
    */

    public void setStripper(Stripper s) {
        stripper = s;
    }


    /**
    * Request stripping of all whitespace text nodes. This is a simple shortcut to
    * avoid allocating a separate Stripper object
    */

    public void setStripAll() {
        stripper = new AllElementsStripper();
    }

    /**
    * Set the PreviewManager to use
    */

    public void setPreviewManager(PreviewManager pm) {
        previewManager = pm;
    }


    /**
    * Indicate whether comments and Processing Instructions should be discarded
    * @params discard true if comments and PIs are to be discarded, false if
    * they are to be added to the tree
    */

    public void setDiscardCommentsAndPIs(boolean discard) {
        discardComments = discard;
    }

    /**
    * Set the error handler to use. If none is specified, SAXON supplies its own,
    * which writes error messages to the selected error output writer.
    * @param eh The error handler to use. It must conform to the interface
    * org.xml.sax.ErrorHandler
    */

    public void setErrorHandler(ErrorHandler eh) {
        this.errorHandler = eh;
    }

    /**
    * Set output for error messages produced by the default error handler.<BR>
    * The default error handler does not throw an exception
    * for parse errors or input I/O errors, rather it returns a result code and
    * writes diagnostics to a user-specified output writer, which defaults to
    * System.err<BR>
    * This call has no effect if setErrorHandler() has been called to supply a
    * user-defined error handler
    * @param writer The Writer to use for error messages
    */

    public void setErrorOutput(Writer writer) {
        errorOutput = writer;
    }

    /**
    * Build the tree from an input source. After building the tree, it can
    * be walked as often as required using run(Document doc).
    * @param source The InputSource to use. InputSource is a SAX-defined class that
    * allows input from a URL, a byte stream, or a character stream. SAXON also
    * provides a subclass, ExtendedInputSource, that allows input directly from a File.
    * @return The DocumentInfo object that results from parsing the input.
    * @throws SAXException if the input document could not be read or if it was not parsed
    * correctly.
    */

    public DocumentInfo build(InputSource in) throws SAXException
    {
        // System.err.println("Builder " + this + " build using parser " + parser);
        if (timing) {
            System.err.println("Building tree for " + in.getSystemId() + " using " + getClass());
            startTime = (new Date()).getTime();
        }

        failed = true;  // until startDocument() called
        started = false;
        String uri = in.getSystemId();
        baseURI = uri;

        if (in instanceof ExtendedInputSource) {
            estimatedLength = ((ExtendedInputSource)in).getEstimatedLength();
            if (estimatedLength < 1) estimatedLength = 4096;
            if (estimatedLength > 1000000) estimatedLength = 1000000;
        } else {
            estimatedLength = 4096;
        }

        // parse the document

        if (parser==null) {
            parser = ParserManager.makeParser();
        }

		ContentEmitter ce = new ContentEmitter();
		ce.setNamePool(namePool);
		parser.setContentHandler(ce);
		parser.setDTDHandler(ce);
        parser.setErrorHandler(errorHandler);

		if (!discardComments) {
			try {
            	parser.setProperty("http://xml.org/sax/properties/lexical-handler", ce);
            } catch (SAXNotSupportedException err) {    // this just means we won't see the comments
        	} catch (SAXNotRecognizedException err) {
        	}
        }        

        if (stripper!=null && !stripper.getPreserveAll()) {
            ce.setEmitter(stripper);
            stripper.setUnderlyingEmitter(this);
        } else {
            ce.setEmitter(this);
        }

        try {
            parser.parse(in);
        } catch (java.io.IOException err) {
            throw new SAXException(err);
        }

        if (!started) {
            // System.err.println("Builder " + this + " failed");
            throw new SAXException("Source document not supplied");
        }

        if (failed) {
            // System.err.println("Builder " + this + " failed");
            throw new SAXException("XML Parsing failed");
        }

        if (timing) {
            long endTime = (new Date()).getTime();
            System.err.println("Tree built in " + (endTime-startTime) + " milliseconds");
            startTime = endTime;
        }

        return currentDocument;
    }

    /**
    * Get the current document
    * @return the document that has been most recently built using this builder
    */

    public DocumentInfo getCurrentDocument() {
        return currentDocument;
    }


  ////////////////////////////////////////////////////////////////////////////////////////
  // Implement the Emitter interface. Abstract methods to be
  // implemented in subclasses.  
  ////////////////////////////////////////////////////////////////////////////////////////

    /**
    * Callback interface for SAX: not for application use
    */

    public abstract void startDocument () throws SAXException;

    /**
    * Callback interface for SAX: not for application use
    */

    public abstract void endDocument () throws SAXException; 

    /**
    * Callback interface for SAX: not for application use
    */
    
    public void setDocumentLocator (Locator locator)
    {
        this.locator = locator;
    }

     /**
    * Callback interface for SAX: not for application use
    */

    public abstract void startElement (
        int nameCode, Attributes attributes, int[] namespaces, int nscount)
        throws SAXException;


    /**
    * Callback interface for SAX: not for application use
    */

    public abstract void endElement (int nameCode) throws SAXException;

    /**
    * Callback interface for SAX: not for application use
    */

    public abstract void characters (char ch[], int start, int length) throws SAXException;

   
    /**
    * Callback interface for SAX: not for application use<BR>
    * Note: because SAX1 does not deliver comment nodes, we get these in the form of a processing
    * instruction with a null name. This requires a specially-adapted SAX driver.
    */

    public abstract void processingInstruction (String name, String remainder)
        throws SAXException;

    /**
    * Callback interface for SAX (part of LexicalHandler interface): not for application use
    */
 
    public abstract void comment (char ch[], int start, int length) throws SAXException;


    ////////////////////////////////////////////////////////////////////////////
    // Implement the org.xml.sax.ErrorHandler interface.
    // The user can supply an alternative implementation using setErrorHandler()
    ////////////////////////////////////////////////////////////////////////////

    /**
    * Callback interface for SAX: not for application use
    */

    public void warning (SAXParseException e) 
    {}

    /**
    * Callback interface for SAX: not for application use
    */

    public void error (SAXParseException e) throws SAXException {
        reportError(e, false);
        failed = true;
    }

    /**
    * Callback interface for SAX: not for application use
    */
    
    public void fatalError (SAXParseException e) throws SAXException {
        reportError(e, true);
        failed = true;
        throw e;
    }

    /**
    * Common routine for errors and fatal errors
    */

    protected void reportError (SAXParseException e, boolean isFatal) {
        try {
            String errcat = (isFatal ? "Fatal error" : "Error");
            errorOutput.write(errcat + " reported by XML parser: " + e.getMessage() + "\n");
            errorOutput.write("  URL:    " + e.getSystemId() + "\n");
            errorOutput.write("  Line:   " + e.getLineNumber() + "\n");
            errorOutput.write("  Column: " + e.getColumnNumber() + "\n");
            errorOutput.flush();
        } catch (Exception e2) {
            System.err.println(e);
            System.err.println(e2);
            e2.printStackTrace();
        };
    }    



    /**
    * Set the URI for an unparsed entity in the document.
    * Abstract method to be implemented in each subclass.
    */

    public abstract void setUnparsedEntity(String name, String uri);

	/**
	* Methods to satisfy the Emitter interface
	*/
	
    public void setWriter (Writer writer) throws SAXException {}

    public void setCharacterSet(CharacterSet charset) throws SAXException {}
    
    public void setEscaping(boolean escaping) throws SAXException {}
    
    public void setOutputDetails(OutputDetails details) throws SAXException {}


    //////////////////////////////////////////////////////////////////////////////
    // Implement Locator interface (default implementation)
    //////////////////////////////////////////////////////////////////////////////

    public void setSystemId(String uri) {
        baseURI = uri;
    }

    public String getSystemId() {
        return baseURI;
    }

    public String getPublicId() {
        return null;
    }

    public int getLineNumber() {
        return -1;
    }

    public int getColumnNumber() {
        return -1;
    }

    ////////////////////////////////////////////////////////////////////////////////////
    // Inner class AllElementsStripper
    // A Stripper that removes all whitespace text nodes.
    ////////////////////////////////////////////////////////////////////////////////////


    private class AllElementsStripper extends Stripper {

        /**
        * Decide whether an element is in the set of white-space preserving element types
        * @param name The element name
        * @return true if the element is in the set of white-space preserving element types
        */

        public boolean isSpacePreserving(Name name) {
            return false;
        }

        /**
        * Return true if all whitespace nodes are to be preserved
        */

        public boolean getPreserveAll() {
            return false;
        }
    }


}   // end of outer class Builder

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
