package com.icl.saxon.axis;
import com.icl.saxon.*;
import com.icl.saxon.om.NodeInfo;
import com.icl.saxon.pattern.NodeTest;
import com.icl.saxon.expr.*;
import org.xml.sax.SAXException;
import java.util.*;
import org.xml.sax.*;


/**
* An axis, that is a direction of navigation in the document structure.
*/

public abstract class Axis  {
   
    protected NodeTest nodeTest;

    public Axis() {}

    /**
    * resolve an axis name into a symbolic constant representing the axis
    */

    public static Axis make(String name) throws SAXException {
        if (name.equals("ancestor"))                return new AncestorAxis();
        if (name.equals("ancestor-or-self"))        return new AncestorOrSelfAxis();
        if (name.equals("attribute"))               return new AttributeAxis();
        if (name.equals("child"))                   return new ChildAxis();
        if (name.equals("descendant"))              return new DescendantAxis();
        if (name.equals("descendant-or-self"))      return new DescendantOrSelfAxis();
        if (name.equals("following"))               return new FollowingAxis();
        if (name.equals("following-sibling"))       return new FollowingSiblingAxis();
        if (name.equals("namespace"))               return new NamespaceAxis();   
        if (name.equals("parent"))                  return new ParentAxis();
        if (name.equals("preceding"))               return new PrecedingAxis();
        if (name.equals("preceding-sibling"))       return new PrecedingSiblingAxis();
        if (name.equals("self"))                    return new SelfAxis();
        throw new SAXException("Unknown axis name: " + name);
    }



    /**
    * Get the principal node type for the axis. This is the default node type, and is
    * Node.ELEMENT for all axes except the Attribute and Namespace axes
    */

    public short getPrincipalNodeType() {
        return NodeInfo.ELEMENT;
    }

    /**
    * Set the name and node type test required by the axis.
    * @param test A NodeTest object
    */

    public void setNodeTest(NodeTest test) {
        nodeTest = test;
    }

    /**
    * Return an enumeration over the nodes reached by the given axis from a specified node
    * @param node NodeInfo representing the node from which the enumeration starts
    * @param nodeType the type(s) of node to be included, e.g. NodeInfo.ELEMENT, NodeInfo.TEXT.
    * The value NodeInfo.NODE means include any type of node.
    * @param nodeNameTest include only nodes with this name (e.g. an element name). Set this to null
    * to include nodes of any name.
    * @return a NodeEnumeration that scans the nodes reached by the axis in turn.
    * @throws SAXException if an invalid axis is specified
    */

    public abstract NodeEnumeration getEnumeration(NodeInfo node)
        throws SAXException;


    /**
    * An Axis has the sorted property if it returns nodes in document order
    */

    public abstract boolean isSorted();

    /**
    * If an axis is not in document order, then it is in reverse document order
    */

    public boolean isReverseSorted() {
        return !isSorted();
    }

    /**
    * An Axis has the sibling property if all the nodes it returns have the same parent
    */

    public abstract boolean isSibling() ;

    /**
    * An axis has the peer property if no node in the result is an ancestor of another
    * (redundant: if this is true the sibling property will also be true)
    */

    public abstract boolean isPeer();

    /**
    * An axis has the withinSubtree property if all the nodes it returns are in the subtree
    * rooted at the start node, that is if it is a subset of the descendants-of-self axis
    */

    public abstract boolean isWithinSubtree();

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
