package com.icl.saxon.expr;
import com.icl.saxon.*;
import com.icl.saxon.om.Namespace;
import com.icl.saxon.om.NodeInfo;
import org.xml.sax.SAXException;
import java.lang.reflect.*;
import org.w3c.dom.*;


/**
* This class acts as a proxy for an extension function defined as a method
* in a user-defined class
*/

public class FunctionProxy extends Function {

    private Class theClass;
    private Object theInstance;
    private Method theMethod;
    private Class[] theParameterTypes;
    private Class theReturnType;
    private SAXException theException = null;
    private String name;
    private boolean usesContext;
    private boolean isStatic;
    private Constructor theConstructor;

    /**
    * Constructor: creates an uncommitted FunctionProxy
    */

    public FunctionProxy() {}

    /**
    * setMethod: locates the external class and method to which this function relates. At this
    * stage addArguments() will have normally been called, so the number of arguments is known. If no
    * method of the required name is located, an exception is saved, but it is not thrown until
    * an attempt is made to evaluate the function. The method is also used before calling
    * function-available(). In this case the number of arguments is not known (it will be
    * set to zero). We return true if a match was found, regardless of the number of arguments.
    * @param fullName This identifies both the local name of the function and its namespace.
	* @return true if the function is available (with some number of arguments).
    */

    public boolean setMethod(String uri, String localName) {
    	boolean isAvailable = false;
        try {
            name = localName;
            String className = getClassName(uri);
            int numArgs = getNumberOfArguments();
            int significantArgs = numArgs;

            theClass = Loader.getClass(className);

            // if the method name is "new", look for a matching constructor

            if (name.equals("new")) {
                Constructor[] constructors = theClass.getConstructors();
                for (int c=0; c<constructors.length; c++) {
                	isAvailable = true;
                    theConstructor = constructors[c];
                    theParameterTypes = theConstructor.getParameterTypes();
                    theReturnType = theClass;
                    isStatic = true;
                    if (theParameterTypes.length == numArgs) {
                        return true;
                    }
                }
                throw new SAXException("No constructor with " + numArgs +
                                     (numArgs==1 ? " parameter" : " parameters") +
                                      " found in class " + className);
            } else {
        
                // look through the methods of this class to find one that matches the local name

                Method[] methods = theClass.getMethods();
                for (int m=0; m<methods.length; m++) {

                    theMethod = methods[m];

					if (matches(theMethod.getName(), name)) {
						isAvailable = true;
	                    theParameterTypes = theMethod.getParameterTypes();
	                    theReturnType = theMethod.getReturnType();            
	                    isStatic = Modifier.isStatic(theMethod.getModifiers());
	
	                    // if the method is not static, the first supplied argument is the instance, so
	                    // discount it
	
	                    significantArgs = (isStatic ? numArgs : numArgs - 1);
	
	                    if (significantArgs>=0) {
	
	                        // System.err.println("Looking for " + name + "(" + significantArgs +"), trying " +
	                        //     methods[m].getName() + "(" + theParameterTypes.length + ")");
	   
	                        if (theParameterTypes.length == significantArgs &&
	                                (significantArgs==0 || theParameterTypes[0]!=Context.class)) {
	                            usesContext = false;
	                            return true;
	                        }
	            
	                        // we allow the method to have an extra parameter if the first parameter is Context
	            
	                        if (theParameterTypes.length == significantArgs+1 &&
	                                theParameterTypes[0]==Context.class) {
	                            usesContext = true;
	                            return true;
	                        }
	                    }
	                }
                }
        
                // No method found.
                  
                throw new SAXException("No method matching " + name +
                                         " with " + significantArgs +
                                         (significantArgs==1 ? " parameter" : " parameters") +
                                          " found in class " + className);
            }
        } catch (SAXException err) {    // Don't fail until the method is called
            theException = err;
            return isAvailable;
        }
    }

    /**
    * Determine the data type of the expression, if possible
    * @return Value.ANY (meaning not known in advance)
    */

    public int getDataType() {
        return Value.ANY;	//TODO: we could do better than this
    }



    private String getClassName(String uri) throws SAXException {

        if (uri.equals(Namespace.SAXON)) {  // recognise the SAXON extension functions specially

            return "com.icl.saxon.functions.Extensions";

        } else {

            // support the URN format java:full.class.Name

            if (uri.startsWith("java:")) {
                return uri.substring(5);
            }   

            // extract the class name as anything in the URI after the last "/"

            int slash = uri.lastIndexOf('/');
            if (slash<0) {
                return uri;
            } else if (slash==uri.length()-1) {
                throw new SAXException("Namespace URI must contain a classname after the final slash (" + uri + ")");
            } else {
                return uri.substring(slash+1);
            }
        }
    }
        

    /**
    * Determine whether the function is available. This supports the function-available
    * function, and as such it is done without regard to the number of arguments.
    */

	// no longer used, done by calling setMethod() instead.
	
    //public boolean isAvailable(String uri, String localName) {
	//
    //    try {
    //        String className = getClassName(uri);
    //        theClass = Loader.getClass(className);
    //    } catch (Exception err) {
    //        return false;
    //    }

        // for the constructor, return true if it seems to be instantiable

    //    if (localName.equals("new")) {
    //        int m = theClass.getModifiers();
    //        if (Modifier.isInterface(m)) return false;
    //        if (Modifier.isAbstract(m)) return false;
    //        if (theClass.getConstructors().length == 0) return false;
    //        return true;
    //    }

        // look through the methods of this class to find one that matches the local name

    //    Method[] methods = theClass.getMethods();
    //    for (int m=0; m<methods.length; m++) {
    //        if (matches(methods[m].getName(), localName)) {
    //            return true;
    //        }
    //    }
    //    return false;
    // }

    /**
    * Determine whether the XPath function name matches the Java method name.
    * This is true if the names are the same ignoring hyphens and case.
    * @param methodName the Java method name
    * @param functionName the XPath function name
    */

    private boolean matches(String methodName, String functionName) {
        if (methodName.equalsIgnoreCase(functionName)) return true;
        int m=0;
        int f=0;
        int ml = methodName.length();
        int fl = functionName.length();
        while (true) {
            if (functionName.charAt(f)=='-') {
                f++;
                if (f>fl) return false;
            }
            if (Character.toUpperCase(methodName.charAt(m)) !=
                     Character.toUpperCase(functionName.charAt(f))) return false;
            f++;
            m++;
            if (f>=fl && m>=ml) return true;
            if (f>=fl || m>=ml) return false;
        }
    }
            
    /**
    * Get the name of the function
    */

    public String getName() {
        return name;
    }

    /**
    * Simplify the function (by simplifying its arguments)
    */

    public Expression simplify() throws SAXException {
        for (int i=0; i<getNumberOfArguments(); i++) {
            argument[i] = argument[i].simplify();
        }
        return this;
    }

    /**
    * Determine which aspects of the context the expression depends on. The result is
    * a bitwise-or'ed value composed from constants such as Context.VARIABLES and
    * Context.CURRENT_NODE
    */

    public int getDependencies() {
        int dep = 0;
        if (usesContext) {
            dep = Context.CONTEXT_NODE | Context.POSITION | Context.LAST;
        } 
        for (int i=0; i<getNumberOfArguments(); i++) {
            dep |= argument[i].getDependencies();
        }
        return dep;
    } 

    /**
    * Perform a partial evaluation of the expression, by eliminating specified dependencies
    * on the context.
    * @param dependencies The dependencies to be removed
    * @param context The context to be used for the partial evaluation
    * @return a new expression that does not have any of the specified
    * dependencies
    */

    public Expression reduce(int dependencies, Context context) throws SAXException {

        // only safe thing is to evaluate it now
        if (usesContext &&
            (dependencies & (Context.CONTEXT_NODE |
                                 Context.POSITION | Context.LAST)) != 0) {
            return evaluate(context);
            
        } else {

            // eliminate dependencies in the arguments
            FunctionProxy fp = new FunctionProxy();
            fp.theClass = theClass;
            fp.theInstance = theInstance;
            fp.theMethod = theMethod;
            fp.theParameterTypes = theParameterTypes;
            fp.theReturnType = theReturnType;
            fp.theException = theException;
            fp.name = name;
            fp.usesContext = usesContext;
            fp.isStatic = isStatic;
            fp.theConstructor = theConstructor;
            fp.argument = new Expression[getNumberOfArguments()];
            for (int a=0; a<getNumberOfArguments(); a++) {
                fp.addArgument(argument[a].reduce(dependencies, context));
            }
            return fp;
        }
    }

    /**
    * Evaluate the function. <br>
    * @param context The context in which the function is to be evaluated
    * @return a Value representing the result of the function. 
    * @throws SAXException if the function cannot be evaluated.
    */    

    public Value evaluate(Context context) throws SAXException {

        // Fail now if no method was found

        if (theException!=null) {
            throw theException;
        }

        if (name.equals("new")) {       
            int requireArgs = theParameterTypes.length;
            int numArgs = checkArgumentCount(requireArgs, requireArgs);
            
            Object[] params = new Object[theParameterTypes.length];
            
            setupParams(params, theParameterTypes, 0, 0, context);           

            try {
                Object obj = theConstructor.newInstance(params);
                return new ObjectValue(obj);
            } catch (InstantiationException err0) {
                throw new SAXException ("Cannot instantiate class", err0);
            } catch (IllegalAccessException err1) {
                throw new SAXException ("Constructor access is illegal", err1);
            } catch (IllegalArgumentException err2) {
                throw new SAXException ("Argument is of wrong type", err2);
            } catch (InvocationTargetException err3) {
                Throwable ex = err3.getTargetException();
                if (ex instanceof SAXException) {
                    throw (SAXException)ex;
                } else {
                	if (context.getController().isTracing()) {
                		err3.getTargetException().printStackTrace();
                	}
                    throw new SAXException ("Exception in extension function " +
                                            err3.getTargetException().toString());
                }
            }
        } else {

            if (isStatic) {
                theInstance = null;
            } else {
                int actualArgs = getNumberOfArguments();
                if (actualArgs==0) {
                    throw new SAXException("Must supply an argument for an instance-level extension function");
                }
                Value arg0 = argument[0].evaluate(context);
                if (!(arg0 instanceof ObjectValue)) {
                    throw new SAXException("First argument is not an object instance");
                }
                theInstance = ((ObjectValue)arg0).getObject();
            }

            int requireArgs = theParameterTypes.length -
                                 (usesContext ? 1 : 0) +
                                 (isStatic ? 0 : 1);
                                 
            checkArgumentCount(requireArgs, requireArgs);
            Object[] params = new Object[theParameterTypes.length];

            if (usesContext) {
                context.setStaticContext(getStaticContext());
                params[0] = context;
            }

            setupParams(params, theParameterTypes,
                            (usesContext ? 1 : 0),
                            (isStatic ? 0 : 1),
                            context);
      
            try {
                Object result = theMethod.invoke(theInstance, params);
                return convertJavaObjectToXPath(result);
            } catch (IllegalAccessException err1) {
                throw new SAXException ("Method access is illegal", err1);
            } catch (IllegalArgumentException err2) {
                throw new SAXException ("Argument is of wrong type", err2);
            } catch (InvocationTargetException err3) {
                Throwable ex = err3.getTargetException();
                if (ex instanceof SAXException) {
                    throw (SAXException)ex;
                } else {
                	if (context.getController().isTracing()) {
                		err3.getTargetException().printStackTrace();
                	}
                    throw new SAXException ("Exception in extension function " +
                                            err3.getTargetException().toString());
                }
            }
        }
    }

    public static Value convertJavaObjectToXPath(Object result) throws SAXException {
        if (result instanceof Boolean) {
            return new BooleanValue(((Boolean)result).booleanValue());
        } else if (result instanceof Double) {
            return new NumericValue(((Double)result).doubleValue());
        } else if (result instanceof Float) {
            return new NumericValue((double)((Float)result).floatValue());
        } else if (result instanceof Short) {
            return new NumericValue((double)((Short)result).shortValue());
        } else if (result instanceof Integer) {
            return new NumericValue((double)((Integer)result).intValue());
        } else if (result instanceof Long) {
            return new NumericValue((double)((Long)result).longValue());
        } else if (result instanceof String) {
            return new StringValue((String)result);
        } else if (result instanceof Value) {
            return (Value)result;
        } else if (result instanceof NodeInfo) {
            return new SingletonNodeSet((NodeInfo)result);
        } else if (result instanceof org.w3c.dom.NodeList) {
            NodeList list = ((NodeList)result);
            NodeInfo[] nodes = new NodeInfo[list.getLength()];
            for (int i=0; i<list.getLength(); i++) {
                if (list.item(i) instanceof NodeInfo) {
                    nodes[i] = (NodeInfo)list.item(i);
                } else {
                    throw new SAXException("Result NodeList contains non-Saxon DOM Nodes");
                }
            }
            return new NodeSetExtent(nodes);
        } else if (result instanceof org.w3c.dom.Node) {
            throw new SAXException("Result is a non-Saxon DOM Node");
        } else {
            return new ObjectValue(result);
        }
    }
                
    private void setupParams(Object[] params,
                             Class[] paramTypes,
                             int firstParam,
                             int firstArg,
                             Context context) throws SAXException {
        int j=firstParam;
        for (int i=firstArg; i<getNumberOfArguments(); i++) {
            Value arg = argument[i].evaluate(context);
            Class pclass = paramTypes[j];

            // try direct mappings of scalar values
        
            if (pclass==boolean.class) {
                params[j++] = new Boolean(arg.asBoolean());
            } else if (pclass==double.class) {
                params[j++] = new Double(arg.asNumber());
            } else if (pclass==float.class) {
                params[j++] = new Float((float)arg.asNumber());
            } else if (pclass==int.class) {
                params[j++] = new Integer((int)arg.asNumber());
            } else if (pclass==long.class) {
                params[j++] = new Long((long)arg.asNumber());
            } else if (pclass==short.class) {
                params[j++] = new Short((short)arg.asNumber());
            } else if (pclass==String.class) {
                params[j++] = new String(arg.asString());
            } else if (pclass==Value.class) {
                params[j++] = arg;


            // try the DOM types
            
            } else if (arg instanceof NodeSetValue &&
                            (pclass==org.w3c.dom.Node.class ||
                             pclass==org.w3c.dom.Document.class ||
                             pclass==org.w3c.dom.Element.class ||
                             pclass==org.w3c.dom.Attr.class ||
                             pclass==org.w3c.dom.Text.class ||
                             pclass==org.w3c.dom.Comment.class ||
                             pclass==org.w3c.dom.ProcessingInstruction.class)) {
                params[j++] = ((NodeSetValue)arg).getFirst();
            } else if (arg instanceof NodeSetValue &&
                             pclass==org.w3c.dom.NodeList.class) {
                params[j++] = arg;
            //} else if (arg instanceof FragmentValue &&
            //                 pclass==org.w3c.dom.DocumentFragment.class) {
            //    params[j++] = ((FragmentValue)arg);

            // following attempts will fail if the parameter is declared with the wrong type,
            // but they are useful where the type is, e.g. "Object"
            
            } else if (arg instanceof BooleanValue) {
                params[j++] = new Boolean(arg.asBoolean());
            } else if (arg instanceof NumericValue) {
                params[j++] = new Double(arg.asNumber());
            } else if (arg instanceof StringValue) {
                params[j++] = new String(arg.asString());
            } else if (arg instanceof ObjectValue) {
                Object obj = ((ObjectValue)arg).getObject();
                params[j++] = obj;
            } else {
                params[j++] = arg;
            }
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
