package com.icl.saxon.expr;
import com.icl.saxon.om.NodeInfo;
import com.icl.saxon.Context;

import org.xml.sax.SAXException;


/**
* A TailExpression is a filter expression of the form "xxx[position()>n]"
*/

class TailExpression extends NodeSetExpression {

    private Expression baseExpression;
    private int startPosition;

    /**
    * Constructor
    * @param start A node-set expression denoting the absolute or relative set of nodes from which the
    * navigation path should start.
    * @param filter An expression defining the filter predicate
    */

    public TailExpression(Expression base, int start) {
        this.baseExpression = base;
        this.startPosition = start;
    }

    /**
    * Simplify the expression
    */

    public Expression simplify() throws SAXException {        
        baseExpression = baseExpression.simplify();

        // following code is designed to catch the case where we recurse over a node-set
        // setting $ns := $ns[position()>1]
        
        if (baseExpression instanceof NodeSetIntent) {
            NodeSetIntent b = (NodeSetIntent)baseExpression;
            if (b.getNodeSetExpression() instanceof TailExpression) {
                TailExpression t = (TailExpression)b.getNodeSetExpression();
                return new TailExpression(t.baseExpression,
                                          t.startPosition+this.startPosition-1);
            }
        }
        return this;
    }

    /**
    * Evaluate the tail expression in a given context to return a Node Enumeration
    * @param context the evaluation context
    * @param sort true if the result must be in document order
    */

    public NodeEnumeration enumerate(Context context, boolean sort) throws SAXException {
        NodeEnumeration base = baseExpression.enumerate(context, true);
        return new TailEnumerator(base, startPosition);
    }

    /**
    * Return a string representation of the expression
    * Used for diagnostics, and also for reconstituting the expression in the compiler
    */

    public String toString() {
        return baseExpression.toString() + "[position() > " + (startPosition-1) + "]";
    }

    /**
    * Determine which aspects of the context the expression depends on. The result is
    * a bitwise-or'ed value composed from constants such as Context.VARIABLES and
    * Context.CURRENT_NODE
    */

    public int getDependencies() {
        return baseExpression.getDependencies();
                    
    }

    /**
    * Perform a partial evaluation of the expression, by eliminating specified dependencies
    * on the context.
    * @param dependencies The dependencies to be removed
    * @param context The context to be used for the partial evaluation
    * @return a new expression that does not have any of the specified
    * dependencies
    */

    public Expression reduce(int dependencies, Context context) throws SAXException {
        if ((dependencies & baseExpression.getDependencies()) != 0) {
            Expression newbase = baseExpression.reduce(dependencies, context);
            Expression e = new TailExpression(newbase, startPosition);
            e.setStaticContext(getStaticContext());
            return e.simplify();
        } else {
            return this;
        }
    }

    /**
    * TailEnumerator class
    */

private class TailEnumerator implements NodeEnumeration {

    private NodeEnumeration base;
    
    /**
    * Constructor
    * @param base A node-set expression denoting the absolute or relative set of nodes from which the
    * @param startAt The first node to be returned, counting from one.
    */

    public TailEnumerator(NodeEnumeration base, int startAt) throws SAXException {
        this.base = base;
        // discard the leading nodes in the base enumeration
        while (startAt > 1 && base.hasMoreElements()) {
            NodeInfo discard = base.nextElement();
            startAt--;
        }
    }

    /**
    * Test whether there are any more nodes available in the enumeration
    */

    public boolean hasMoreElements() throws SAXException {
        return base.hasMoreElements();
    }

    /**
    * Get the next node if there is one
    */

    public NodeInfo nextElement() throws SAXException {
        return base.nextElement();
    }

    /**
    * Determine whether the nodes are guaranteed to be in document order
    */

    public boolean isSorted() throws SAXException {
        return base.isSorted();
    }

    public boolean isReverseSorted() throws SAXException {
        return base.isReverseSorted();
    }

    /**
    * Determine whether the nodes are guaranteed to be peers
    */

    public boolean isPeer() throws SAXException {
        return base.isPeer();
    }
}    

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
