package com.icl.saxon.functions;
import com.icl.saxon.*;
import com.icl.saxon.om.NodeInfo;
import com.icl.saxon.om.DocumentInfo;
import com.icl.saxon.expr.*;
import org.xml.sax.SAXException;
import java.util.*;
import java.text.*;



public class Key extends Function {

    private DocumentInfo boundDocument = null;
    private Controller boundController = null;

    public String getName() {
        return "key";
    };

    /**
    * Determine the data type of the expression
    * @return Value.NODESET
    */

    public int getDataType() {
        return Value.NODESET;
    }

    /**
    * Simplify and validate
    */

    public Expression simplify() throws SAXException {
    	// check that key() function is allowed in this context
    	if (!getStaticContext().allowsKeyFunction()) {
    		throw new SAXException("key() function cannot be used here");
    	}
        checkArgumentCount(2, 2);
        argument[0] = argument[0].simplify();
        argument[1] = argument[1].simplify();
        return this;
    }

    /**
    * Evaluate the expression
    */

    public Value evaluate(Context context) throws SAXException {
        return new NodeSetExtent(enumerate(context));
    }

    /**
    * Enumerate the expression
    */

    public NodeEnumeration enumerate(Context context) throws SAXException {   
        String givenkeyname = argument[0].evaluateAsString(context);
        Value keyvalue = argument[1].evaluate(context);
        
        Controller controller = boundController;
        if (controller==null) controller = context.getController();

        DocumentInfo doc = boundDocument;
        if (doc==null) doc = context.getContextNode().getDocumentRoot();
        
        int fingerprint = getStaticContext().getFingerprint(givenkeyname, false);  
        return findKey(controller, doc, fingerprint, keyvalue);
    }

    /**
    * Construct an enumeration of nodes that satisfy the given key
    * @param controller The controller (to get the key definitions)
    * @param doc The document to search
    * @param keyname The absolute (expanded) name of the key
    * @param arg1 The value of the key (or nodeset containing the values)
    */

    private NodeEnumeration findKey(
                            Controller controller,
                            DocumentInfo doc,
                            int fingerprint,
                            Value arg1) throws SAXException {
      
        KeyManager keyManager = controller.getKeyManager();

        boolean sorted;

        if ((arg1 instanceof NodeSetValue) &&
        		!(arg1 instanceof FragmentValue)) {

            NodeSetValue keyvals = (NodeSetValue)arg1;

            NodeEnumeration supplied = keyvals.enumerate();
            NodeEnumeration result = null;
                    
            int inNodes = 0;
            while (supplied.hasMoreElements()) {
                inNodes++;
                NodeEnumeration onekeyresult =
                    keyManager.selectByKey(fingerprint,
                                           doc,
                                           supplied.nextElement().getValue(),
                                           controller);
                if (inNodes==1) {
                    result = onekeyresult;
                } else {
                    result = new UnionEnumeration(result, onekeyresult);
                }
            }
            if (inNodes==0) {
                return new SingletonEnumeration(null);
            } else {
                return result;
            }

        } else {
            return keyManager.selectByKey(fingerprint, doc, arg1.asString(), controller);
        }
    }

    /**
    * Determine which aspects of the context the expression depends on. The result is
    * a bitwise-or'ed value composed from constants such as Context.VARIABLES and
    * Context.CURRENT_NODE
    */

    public int getDependencies() {
        int dependencies = argument[0].getDependencies();
        dependencies |= argument[1].getDependencies();
        if (boundDocument == null) {
            dependencies |= Context.CONTEXT_NODE;
        } 
        if (boundController == null) {
            dependencies |= Context.CONTROLLER;
        }
        return dependencies;
    }

    /**
    * Remove specified dependencies. 
    */

    public Expression reduce(int dep, Context context)
            throws SAXException {
        Key key = new Key();
        key.addArgument(argument[0].reduce(dep, context));
        key.addArgument(argument[1].reduce(dep, context));
        if (boundDocument==null && ((dep & Context.CONTEXT_NODE)!=0)) {
            key.boundDocument = context.getContextNode().getDocumentRoot();
        } else {
            key.boundDocument = boundDocument;
        }
        if (boundController==null && ((dep & Context.CONTROLLER)!=0)) {
            key.boundController = context.getController();
        } else {
            key.boundController = boundController;
        }
        key.setStaticContext(getStaticContext());
        return key.simplify();

    }

}





//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
