package com.icl.saxon.om;

import com.icl.saxon.pattern.Pattern;
import com.icl.saxon.Context;
import com.icl.saxon.expr.NodeEnumeration;
import com.icl.saxon.PIGrabber;
import com.icl.saxon.pattern.NodeTypeTest;
import com.icl.saxon.pattern.NameTest;
import com.icl.saxon.pattern.NodeTest;
import java.util.Vector;

import org.xml.sax.SAXException;


/**
  * Navigator provides helper classes for navigating a tree, irrespective
  * of its implementation
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  */



public class Navigator {

    /**
    * Get the URIs of the stylesheet associated with this document by means of an xml-stylesheet
    * processing instruction.
    * @param media The required medium, or null to match any medium
    * @param title The required title, or null to match the preferred stylesheet
    * @return null if there is no such processing instruction
    * @throws SAXException if there is such a processing instruction and it is invalid
    */

    public static String[] getAssociatedStylesheets(
            DocumentInfo doc, String media, String title) throws SAXException {

        PIGrabber grabber = new PIGrabber();
        grabber.setCriteria(media, title, null);
        grabber.setBaseURI(doc.getSystemId());
        
        NodeEnumeration children = doc.enumerateChildren();

        while (children.hasMoreElements()) {
            NodeInfo node = children.nextElement();
            if (node instanceof ProcInstInfo) {
                grabber.processingInstruction(node.getDisplayName(), node.getValue());
            }
            if (node instanceof ElementInfo) {
                break;
            }
        }

        return grabber.getStylesheetURIs();
    }


       
    /**
     * Determine whether this element is the first in a consecutive
     * group. A consecutive group is a group of elements of the same
     * type subordinate to the same parent element; there can be intervening
     * character data, white space, or processing instructions, but no
     * elements of a different type.
     * @param element The element being tested
     * @return True if this is the first child of its parent, or if the
     * previous child was a different element type, or if this element 
     * is the root.
     */

    public static boolean isFirstInGroup(ElementInfo element) {
        NodeInfo prev = (NodeInfo)element.getPreviousSibling();
        while(prev!=null && !(prev instanceof ElementInfo))
            prev = (NodeInfo)prev.getPreviousSibling();
        if (prev==null) return true;
        return prev.getFingerprint() != element.getFingerprint();
    }

    /**
     * Determine whether this element is the last in a consecutive
     * group. A consecutive group is a group of elements of the same
     * type subordinate to the same parent element; there can be intervening
     * character data, white space, or processing instructions, but no
     * elements of a different type.
     * @param element The element being tested
     * @return True if this is the last child of its parent, or if the
     * next child is a different element type, or if this element 
     * is the root.
     */

     public static boolean isLastInGroup(ElementInfo element) {
        NodeInfo next = (NodeInfo)element.getNextSibling();
        while(next!=null && !(next instanceof ElementInfo))
            next = (NodeInfo)next.getNextSibling();
        if (next==null) return true;
        return next.getFingerprint() != element.getFingerprint();
    }
    
    /**
     * Determine whether this element is the first element child of its parent.
     * @param element The element being tested
     * @return True if this element is the first element child of its parent, or
     * if it is the document element.
     */
     
    public static boolean isFirstChild(ElementInfo element) {
        NodeInfo prev = (NodeInfo)element.getPreviousSibling();
        while(prev!=null && !(prev instanceof ElementInfo))
            prev = (NodeInfo)prev.getPreviousSibling();
        return (prev==null);
    }

    /**
     * Determine whether this element is the last child element of its parent.
     * @param element The element being tested
     * @return True if this element is the last child element of its parent, or
     * if it is the root element.
     */
     
    public static boolean isLastChild(ElementInfo element) {
        NodeInfo next = (NodeInfo)element.getNextSibling();
        while(next!=null && !(next instanceof ElementInfo))
            next = (NodeInfo)next.getNextSibling();
        return (next==null);
    }
    
     /**
     * Get the nearest ancestor element with a given name
     * @param node The starting node
     * @param test NodeTest that the required ancestor must satisfy. 
     * @return The NodeInfo for the nearest ancestor with the
     * given tag; null if there is no such ancestor
     */

    public static ElementInfo getAncestor(NodeInfo node, NodeTest test) {
        NodeInfo ancestor = (NodeInfo)node.getParentNode();
        while (ancestor!=null && (ancestor instanceof ElementInfo) &&
                              !test.matches(ancestor)) {
            ancestor = (NodeInfo)ancestor.getParentNode();
        }
        if (ancestor instanceof DocumentInfo) return null;
        return (ElementInfo)ancestor;
    }

     /**
     * Get the nearest ancestor node that matches the given pattern
     * @param node The starting node
     * @param pattern A pattern that the ancestor node must satisfy.
     * @return The NodeInfo for the nearest ancestor node that matches the
     * given pattern; null if there is no such ancestor
     */

    public static NodeInfo getAncestor(NodeInfo node,
                Pattern pat, Context c) throws SAXException
    {
        NodeInfo p = (NodeInfo)node.getParentNode();
        while (p!=null) {
            if (pat.matches(p, c)) return p;
            else p=(NodeInfo)p.getParentNode();
        }
        return null;
    }

    /**
    * Determine whether this node is an ancestor of another node
    * @param node the putative ancestor node
    * @param other the other node (the putative descendant of this node)
    * @return true of this node is an ancestor of the other node
    */

    public static boolean isAncestor(NodeInfo node, NodeInfo other) {
        NodeInfo parent = (NodeInfo)other.getParentNode();
        if (parent==null) return false;
        if (parent.isSameNode(node)) return true;
        return isAncestor(node, parent);
    }

    /**
    * Get the next sibling node that matches a given pattern.
    * @param node The starting node
    * @param pattern The match-pattern that the required sibling must match
    * @return The NodeInfo object describing the next node at the same level.
    * Returns null if there is no subsequent child of the same parent that matches the supplied
    * pattern.
    */
    
    public static NodeInfo getNextSibling(NodeInfo node,
            Pattern pattern, Context c) throws SAXException {
        NodeInfo next=(NodeInfo)node.getNextSibling();
        while (next!=null) {
            if (pattern.matches(next, c)) return next;
            next=(NodeInfo)next.getNextSibling();
        } 
        return null;
    }

    /**
    * Get the first child node matching a given pattern
    * @param node The starting node
    * @param pattern the pattern to be matched
    * @return the the first child node of the required type, or null if there is no such
    * child
    */

    public static NodeInfo getFirstChild(
            NodeInfo node, Pattern pattern, Context c) throws SAXException {
        NodeInfo n = (NodeInfo)node.getFirstChild();
        while (n!=null) {
            if (pattern.matches(n, c)) return n;
            n = (NodeInfo)n.getNextSibling();
        }
        return null;
    }

    /**
    * Get the previous sibling of the node that matches a given pattern.
    * @param node The starting node
    * @param pattern The match-pattern that the required sibling must match
    * @return The NodeInfo object describing the previous node at the same level that
    * matches the pattern. Returns null if there is no previous child of the same
    * parent that matches the supplied pattern.
    */
    
    public static NodeInfo getPreviousSibling(
            NodeInfo node, Pattern pattern, Context c) throws SAXException {
        NodeInfo prev=(NodeInfo)node.getPreviousSibling();
        while (prev!=null) {
            if (pattern.matches(prev, c)) return prev;
            prev=(NodeInfo)prev.getPreviousSibling();
        } 
        return null;
    }

    /**
    * Get the previous node in document order
    * @param node the starting node
    * @param pattern: identifies a pattern the required node
    * @return the previous node in the document of the required type, or null if there is no
    * such node
    */

    public static NodeInfo getPreviousInDocument(
            NodeInfo node, Pattern pattern, Context c) throws SAXException {
        NodeInfo next = node;
        while(true) {
            next = (NodeInfo)next.getPreviousInDocument();
            if (next==null) return null;
            if (pattern.matches(next, c)) return next;
        }
    }

    /**
    * Get the last child node matching a given pattern
    * @param node the starting node
    * @param pattern the pattern to be matched
    * @return the the last child node of the required type, or null if there is no such
    * child
    */

    public static NodeInfo getLastChild(
            NodeInfo node, Pattern pattern, Context c) throws SAXException{
        NodeInfo n = (NodeInfo)node.getLastChild();
        while (n!=null) {
            if (pattern.matches(n, c)) return n;
            n = (NodeInfo)n.getPreviousSibling();
        }
        return null;
    }

    /**
    * Get simple node number. This is defined as one plus the number of previous siblings of the
    * same node type and name. It is not accessible directly in XSL.
    * @param context Used for remembering previous result, for performance
    */

    public static int getNumberSimple(NodeInfo node, Context context) throws SAXException {

        //checkNumberable(node);

        int i=1;
        int type = node.getNodeType();
        NodeInfo prev = (NodeInfo)node.getPreviousSibling();

        if (prev!=null && prev.getNodeType()==type &&
        			 prev.getFingerprint()==node.getFingerprint()) {
            int memo = context.getRememberedNumber(prev);
            if (memo>0) {
                context.setRememberedNumber(node, ++memo);
                return memo;
            }
        }
        
        while (prev!=null) {
            if (prev.getNodeType()==type &&
            	 (prev.getFingerprint()==node.getFingerprint())) i++;
            prev = (NodeInfo)prev.getPreviousSibling();
        }
        context.setRememberedNumber(node, i);
        return i;
    }

    /**
    * Get simple node number. This is defined as one plus the number of previous siblings of the
    * same node type and name. It is not accessible directly in XSL. This version doesn't require
    * the context, and therefore doesn't remember previous results
    */

    public static int getNumberSimple(NodeInfo node) {

        //checkNumberable(node);

        int i=1;
        int type = node.getNodeType();
        NodeInfo prev = (NodeInfo)node.getPreviousSibling();

        while (prev!=null) {
            if (prev.getNodeType()==type && (prev.getFingerprint()==node.getFingerprint())) i++;
            prev = (NodeInfo)prev.getPreviousSibling();
        }

        return i;
    }

    /**
    * Get node number (level="single"). If the current node matches the supplied pattern, the returned
    * number is one plus the number of previous siblings that match the pattern. Otherwise,
    * return the element number of the nearest ancestor that matches the supplied pattern.
    * @param count Pattern that identifies which nodes should be counted. Default (null) is the element
    * name if the current node is an element, or "node()" otherwise.
    * @param from Pattern that specifies where counting starts from. Default (null) is the root node.
    * (This parameter does not seem useful but is included for the sake of XSLT conformance.)
    * @return the node number established as follows: go to the nearest ancestor-or-self that
    * matches the 'count' pattern and that is a descendant of the nearest ancestor that matches the
    * 'from' pattern. Return one plus the nunber of preceding siblings of that ancestor that match
    * the 'count' pattern. If there is no such ancestor, return 0.
    */

    public static int getNumberSingle(NodeInfo node, Pattern count,
                    Pattern from, Context context) throws SAXException {

        checkNumberable(node);

        if (count==null && from==null) {
            return getNumberSimple(node, context);
        }
        
        if (count==null) {
            if (node.getNameCode()==-1) {	// unnamed node
                count = new NodeTypeTest(node.getNodeType());
            } else {
                count = new NameTest(node);
            }
        }

        NodeInfo searchFrom;
        if (from==null) {
            searchFrom = node.getDocumentRoot();
        } else {
            searchFrom = getAncestor(node, from, context);
        }
        
        NodeInfo curr = node;
        while (!count.matches(curr, context)) {
            curr = (NodeInfo)curr.getParentNode();
            if (curr==null || curr==searchFrom) return 0;
        }
        
        int num = 0;
        NodeInfo prev = curr;
        while (prev!=null) {
            num++;
            prev = getPreviousSibling(prev, count, context);
        }

        return num;
    }

    /**
    * Get node number (level="any").
    * Return one plus the number of previous nodes in the
    * document that match the supplied pattern
    * @param count Pattern that identifies which nodes should be counted. Default (null) is the element
    * name if the current node is an element, or "node()" otherwise.
    * @param from Pattern that specifies where counting starts from. Default (null) is the root node.
    * Only nodes after the first (most recent) node that matches the 'from' pattern are counted.
    * @return one plus the number of nodes that precede the current node, that match the count pattern,
    * and that follow the first node that matches the from pattern if specified.
    */

    public static int getNumberAny(NodeInfo node, Pattern count,
                    Pattern from, Context context) throws SAXException {

        //checkNumberable(node);

        if (count==null) {
            if (node.getNameCode()==-1) {	// unnamed node
                count = new NodeTypeTest(node.getNodeType());
            } else {
                count = new NameTest(node);
            }
        }

        NodeInfo curr = node;
        int num = 0;

        while(true) {
            if (curr==null) break;
            if (from!=null && from.matches(curr, context)) break;
            if (count.matches(curr, context)) num++;
            curr = (NodeInfo)curr.getPreviousInDocument();
        }

        return num;
    }

    /**
    * Get node number (level="multiple").
    * Return a vector giving the hierarchic position of this node. See the XSLT spec for details.
    * @param count Pattern that identifies which nodes (ancestors and their previous siblings)
    * should be counted. Default (null) is the element
    * name if the current node is an element, or "node()" otherwise.
    * @param from Pattern that specifies where counting starts from. Default (null) is the root node.
    * Only nodes below the first (most recent) node that matches the 'from' pattern are counted.
    * @return a vector containing for each ancestor-or-self that matches the count pattern and that
    * is below the nearest node that matches the from pattern, an Integer which is one greater than
    * the number of previous siblings that match the count pattern.
    */

    public static Vector getNumberMulti(NodeInfo node, Pattern count,
                    Pattern from, Context context) throws SAXException {

        //checkNumberable(node);

        Vector v = new Vector();

        if (count==null) {
            if (node.getNameCode()==-1) {    // unnamed node
                count = new NodeTypeTest(node.getNodeType());
            } else {
                count = new NameTest(node);
            }
        }

        NodeInfo curr = node;

        while(true) {
            if (count.matches(curr, context)) {
                int num = getNumberSingle(curr, count, null, context);
                v.insertElementAt(new Integer(num), 0);
            }
            curr = (NodeInfo)curr.getParentNode();
            if (curr==null) break;
            if (from!=null && from.matches(curr, context)) break;
        }

        return v;
    }

    private static void checkNumberable(NodeInfo node) throws SAXException {
        if (node instanceof AttributeInfo) {
            throw new SAXException("Attribute nodes cannot be numbered");
        }
        if (node instanceof NamespaceInfo) {
            throw new SAXException("Namespace nodes cannot be numbered");
        }
    }

    /**
    * Translate numeric node type to a string representation
    */

    public static String getNodeTypeName(int type) {
        switch(type) {
            case NodeInfo.NODE:      return "NODE";
            case NodeInfo.ELEMENT:   return "ELEMENT";
            case NodeInfo.TEXT:      return "TEXT";
            case NodeInfo.ATTRIBUTE: return "ATTRIBUTE";
            case NodeInfo.ROOT:  return "DOCUMENT";
            case NodeInfo.PI:        return "PI";
            case NodeInfo.COMMENT:   return "COMMENT";
            default:                 return "NODE";
        }
    }

    /**
    * Construct an alphabetic key from an positive integer; the key collates in the same sequence
    * as the integer
    * @param value The positive integer key value (negative values are treated as zero).
    */

    public static String alphaKey(int value) {
        if (value<1) return "a";
        if (value<10) return "b" + value;
        if (value<100) return "c" + value;
        if (value<1000) return "d" + value;
        if (value<10000) return "e" + value;
        if (value<100000) return "f" + value;
        if (value<1000000) return "g" + value;
        if (value<10000000) return "h" + value;
        if (value<100000000) return "i" + value;
        if (value<1000000000) return "j" + value;
        return "k" + value;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
