package com.icl.saxon.om;
import com.icl.saxon.handlers.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.output.Outputter;
import com.icl.saxon.Context;

import java.util.*;
import java.io.Writer;
import java.net.*;
import org.xml.sax.AttributeList;
import org.xml.sax.SAXException;

import org.w3c.dom.*;


/**
  * A node in the XML parse tree representing an XML element, character content, or attribute.<P>
  * This is the top class in the interface hierarchy for nodes; see NodeImpl for the implementation
  * hierarchy.
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  * @version 20 July 1999
  */

public interface NodeInfo extends Node {

    // Node types. "NODE" means any type.
    // These node numbers should be kept aligned with those defined in the DOM.

    public static final short NODE = 0;       // matches any kind of node
    public static final short ELEMENT = 1;
    public static final short ATTRIBUTE = 2;
    public static final short TEXT = 3;
    public static final short ROOT = 4;
    public static final short PI = 7;
    public static final short COMMENT = 8;
    public static final short NAMESPACE = 9;
    public static final short NUMBER_OF_TYPES = 10;
    public static final short NONE = 9999;    // a test for this node type will never be satisfied

	/**
	* Array of names of node types. You can index into this with the numeric node type.
	*/
	
	public static final String[] NODE_TYPE_NAMES =
		{"Node", "Element", "Attribute", "Text", "Root", "", "", 
			"Processing Instruction", "Comment", "Namespace"};

    /**
    * Return the type of node.
    * @return one of the values Node.ELEMENT, Node.TEXT, Node.ATTRIBUTE, etc.
    */

    public short getNodeType();
    

    /**
    * Determine whether the node is of a given type. <P>
    * Note, this can also be done by testing the node using "instanceof". But this is
    * inconvenient when passing the class as a parameter to another routine.
    * @param nodeType One of the specific node types such as ELEMENT or TEXT, or the general
    * node type NODE
    * @return true if the node is an instance of the specified node type
    */

    //public boolean isa(int nodeType);

    /**
    * Determine whether this is the same node as another node
    * @return true if this Node object and the supplied Node object represent the
    * same node in the tree.
    */

    public boolean isSameNode(NodeInfo other);

    /**
    * Get the base URL for the node.
    * @return the System Identifier of the entity in the source document containing the node,
    * or null if not known
    */

    public String getSystemId();

    /**
    * Get line number
    * @return the line number of the node in its original source document; or -1 if not available
    */

    public int getLineNumber();

    /**
    * Get the node sequence number (in document order). Sequence numbers are monotonic but not
    * consecutive. In the current implementation, parent nodes (elements and roots) have a zero
    * least-significant word, while namespaces, attributes, text nodes, comments, and PIs have
    * the top word the same as their owner and the bottom half reflecting their relative position.
    */

    public long getSequenceNumber(); 
    
    /**
    * Return the character value of the node. The interpretation of this depends on the type
    * of node. For an element it is the accumulated character content of the element,
    * including descendant elements.
    * @return the string value of the node
    */

    public String getValue();

    /**
    * Get the name of this node, following the DOM rules 
    * @return The name of the node. For an element this is the element name, for an attribute
    * it is the attribute name, as a QName. Other node types return conventional names such
    * as "#text" or "#comment"
    */

    public String getNodeName();

	/**
	* Get name code. The name code is a coded form of the node name: two nodes
	* with the same fingerprint have the same namespace URI, the same local name,
	* and the same prefix.
	*/
	
	public int getNameCode();


	/**
	* Get fingerprint. The fingerprint is a coded form of the node name: two nodes
	* with the same fingerprint have the same namespace URI and the same local name.
	* The fingerprint does not include information about the prefix.
	*/
	
	public int getFingerprint();

    /**
    * Get the local part of the name of this node. This is the name after the ":" if any.
    * @return the local part of the name. For an unnamed node, return an empty string.
    */

    public String getLocalName();

    /**
    * Get the prefix part of the name of this node. This is the name before the ":" if any.
    * @return the prefix part of the name. For an unnamed node, return an empty string.
    */

    public String getPrefix();

    /**
    * Get the URI part of the name of this node. This is the URI corresponding to the
    * prefix, or the URI of the default namespace if appropriate.
    * @return The URI of the namespace of this node. For an unnamed node, return null.
    * For a node with an empty prefix, return an empty string.
    */

    public String getURI();

	/**
	* Get the URI code of the name
	*/
	
	public short getURICode();

    /**
    * Get the display name of this node. For elements and attributes this is [prefix:]localname.
    * For unnamed nodes, it is an empty string.
    * @return The display name of this node. 
    * For a node with no name, return an empty string.
    */

    public String getDisplayName();

    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param uri the namespace uri of an attribute ("" if no namespace)
     * @param localname the local name of the attribute
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue(String uri, String localName); 

    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param name the name of an attribute. This must be an unqualified attribute name,
     * i.e. one with no namespace.
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue(String name); 
    
    /**
    * Get the value of a given attribute of this node
    * @param fingerprint The fingerprint of the attribute name
    * @return the attribute value if it exists or null if not
    */
    
    public String getAttributeValue(int fingerprint);
       
    /**
     * Determine whether this element is the outermost element.
     * @return True if this element is the outermost element.
     */
     
    public boolean isDocumentElement(); 
    
    /**
     * Get the outermost element.
     * @return the Element for the outermost element of the document.
     */
     
    public Element getDocumentElement(); 

    /**
    * Get the root (document) node
    * @return the DocumentInfo representing the containing document
    */

    public DocumentInfo getDocumentRoot(); 

    /**
    * Get the next node in document order
    * @param anchor: the scan stops when it reaches a node that is not a descendant of the specified
    * anchor node
    * @return the next node in the document, or null if there is no such node
    */

    public NodeInfo getNextInDocument(NodeInfo anchor); 

    /**
    * Get the previous node in document order
    * @return the previous node in the document, or null if there is no such node
    */

    public NodeInfo getPreviousInDocument(); 

    /**
    * Get the number of children.
    */

    public int getNumberOfChildren();

    /**
    * Get all child nodes of the element (child elements and character nodes)
    * @return an array containing a NodeInfo for each child node
    */

    public NodeInfo[] getAllChildNodes(); 

    /**
    * Get an enumeration of the children of this node
    */

    public NodeEnumeration enumerateChildren();

    /**
    * Get index: that is, the number of preceding sibling nodes at the same level.
    * Note this is not defined for attribute nodes.
    * @return the number of preceding sibling nodes at the same level
    */

    public int getIndex(); 

    /**
    * Get a character string that uniquely identifies this node and that collates nodes
    * into document order
    * @return a string. 
    */

    public String getSequentialKey() ; //throws SAXException;

    /**
    * Perform default action for this kind of node (built-in template rule)
    */

    public void defaultAction(Context c) throws SAXException;

    /**
    * Copy this node to a given outputter
    */

    public void copy(Outputter out) throws SAXException;

    /**
    * Copy the string-value of this node to a given outputter
    */

    public void copyStringValue(Outputter out) throws SAXException;

    /**
    * Generate a path to this node
    */

    public String getPath(); 
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
