package com.icl.saxon.output;
import com.icl.saxon.*;
//import com.icl.saxon.om.Name;
import com.icl.saxon.om.Namespace;
import com.icl.saxon.om.NamePool;
import org.xml.sax.*;
import java.io.*;
import java.util.*;

/**
  * NamespaceEmitter is a ProxyEmitter responsible for removing duplicate namespace
  * declarations
  */
  
public class NamespaceEmitter extends ProxyEmitter
{
    protected NamePool namePool;
    protected int nscodeXML;
    protected int nscodeNull;

    // We keep track of namespaces to avoid outputting duplicate declarations. The namespaces
    // vector holds a list of all namespaces currently declared (organised as pairs of entries,
    // prefix followed by URI). The stack contains an entry for each element currently open; the
    // value on the stack is an Integer giving the size of the namespaces vector on entry to that
    // element.

    private int[] namespaces = new int[30];          // all namespace codes currently declared
    private int namespacesSize = 0;                  // all namespaces currently declared
    private int[] namespaceStack = new int[100];
    private int nsStackTop = 0;
    private int nsCount = 0;                // number of namespaces for current element

    
	/**
	* Set the name pool to be used for all name codes
	*/
	
	public void setNamePool(NamePool pool) {
		namePool = pool;
		nscodeXML = pool.getNamespaceCode("xml", Namespace.XML);
		nscodeNull = pool.getNamespaceCode("", "");
		super.setNamePool(pool);
	}
	

    /**
    * startElement. This call removes redundant namespace declarations.
    */

    public void startElement(int nameCode, Attributes attList, 
    						 int[] namespaceCodes, int nscount) throws SAXException {

		int[] uniqueNamespaces = new int[namespaceCodes.length];
		int uniqueNamespacePos = 0;
		
		for (int n=0; n<nscount; n++) {
			
			int nscode = namespaceCodes[n];
			boolean use = true;

	        if (nscode==nscodeXML) {
	        		// Ignore the XML namespace
	            use = false;
	
	        } else {
	        	
	        		// Ignore xmlns="" unless there is already a different
	        		// xmlns= declaration on the stack
	            boolean needed = (nscode != nscodeNull);
	            for (int i=namespacesSize-1; i>=0; i--) {
	            	if (namespaces[i]==nscode) {
	            		// it's a duplicate so we don't need it
	            		needed = false;
	            		break;
	            	}
	            	if ((namespaces[i]>>16) == (nscode>>16)) {
	            		// same prefix, different URI, so we do need it
	            		needed = true;
	            		break;
		            }
	            }
	            use = needed;
	            
	        } 
	
			if (use) {
	            
		        // Add the declaration to the stack, expanding the stack if necessary
		
		        if (namespacesSize+1 >= namespaces.length) {
		            int[] newlist = new int[namespacesSize*2];
		            System.arraycopy(namespaces, 0, newlist, 0, namespacesSize);
		            namespaces = newlist;
		        }
		        namespaces[namespacesSize++] = nscode;
		        nsCount++;
		        
		    	// include it in the list to be passed to the underlying Emitter
		    	
		    	uniqueNamespaces[uniqueNamespacePos++] = nscode;
		    }
	    }

        // remember how many namespaces there were so we can unwind the stack later

        if (nsStackTop>=namespaceStack.length) {
            int[] newstack = new int[nsStackTop*2];
            System.arraycopy(namespaceStack, 0, newstack, 0, nsStackTop);
            namespaceStack = newstack;
        }

        namespaceStack[nsStackTop++] = nsCount;

        nsCount = 0;


		// finally, pass the namespace on to the underlying Emitter        
        super.startElement(nameCode, attList, uniqueNamespaces, uniqueNamespacePos);
    }        

    /**
    * endElement: Discard the namespaces declared on this element.
    */


    public void endElement (int nameCode) throws SAXException
    {
        if (nsStackTop-- == 0) {
            throw new SAXException("Attempt to output end tag with no matching start tag");
        }
        
        int nscount = namespaceStack[nsStackTop];
        namespacesSize -= nscount;
        
        super.endElement(nameCode);

    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
