package com.icl.saxon.output;
import com.icl.saxon.*;
//import com.icl.saxon.om.Name;
import com.icl.saxon.om.NamePool;
import com.icl.saxon.tree.AttributeCollection;
import org.xml.sax.*;

/**
  * This class allows output to be generated. It channels output requests to an
  * Emitter which does the actual writing.
  * 
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A>
  * @version 19 April 2000: responsibility for eliminating duplicate namespaces has moved
  * to the Emitter.
  */

  // This class is really just a wrapper around the Emitter. It would be better if output
  // were done directly to the emitter, but this interface has been around for a while and
  // is widely used; and the class does add some logic, while Emitter is just an interface.
  

public class Outputter {

	private NamePool namePool;
    private Emitter emitter;
    private OutputDetails outputDetails;
    
    private int pendingStartTag = -1;
    private AttributeCollection pendingAttList;
    
    private int[] pendingNSList = new int[20];
    private int pendingNSListSize = 0;

	public Outputter(NamePool pool) {
		namePool = pool;
		pendingAttList = new AttributeCollection(namePool, 10);
	}
		
    /**
    * Set the emitter that will deal with this output
    */

    public void setEmitter(Emitter handler) {
        this.emitter = handler;
        //emitter.setNamePool(namePool);
    }

    /**
    * Get emitter. This is used by xsl:copy-of, a fragment is copied directly to the
    * Emitter rather than going via the Outputter.
    */
    
    public Emitter getEmitter() throws SAXException {
        if (pendingStartTag != -1) flushStartTag();
        return emitter;
    }

    public void setOutputDetails(OutputDetails details) throws SAXException {
        outputDetails = details;
        emitter.setOutputDetails(details);
    }

    public OutputDetails getOutputDetails() {
        return outputDetails;
    }

    /**
    * Switch escaping (of special characters) on or off.
    * @param escaping: true if special characters are to be escaped, false if not.
    */

    public void setEscaping(boolean escaping) throws SAXException {
        emitter.setEscaping(escaping);
    }

    /**
    * Start the output process
    */

    public void open() throws SAXException {
        //System.err.println("Open " + this + " using emitter " + emitter.getClass());
        emitter.startDocument();
    }

    /**
    * Produce literal output. This is written as is, without any escaping. 
    * The method is provided for Java applications that wish to output literal HTML text.
    * It is not used by the XSL system, which always writes using specific methods such as
    * writeStartTag().
    */

    char[] charbuffer = new char[1024];
    public void write(String s) throws SAXException {
        if (pendingStartTag != -1) flushStartTag();
        emitter.setEscaping(false);
        int len = s.length();
        if (len>charbuffer.length) {
            charbuffer = new char[len];
        }
        s.getChars(0, len, charbuffer, 0);
        emitter.characters(charbuffer, 0, len);
        emitter.setEscaping(true);
    }

    /**
    * Produce text content output. <BR>
    * Special characters are escaped using XML/HTML conventions if the output format
    * requires it.
    * @param s The String to be output
    * @exception SAXException for any failure
    */

    public void writeContent(String s) throws SAXException {
        if (s==null) return;
        int len = s.length();
        if (len>charbuffer.length) {
            charbuffer = new char[len];
        }
        s.getChars(0, len, charbuffer, 0);
        writeContent(charbuffer, 0, len);
    }
    
    /**
    * Produce text content output. <BR>
    * Special characters are escaped using XML/HTML conventions if the output format
    * requires it.
    * @param chars Character array to be output
    * @param start start position of characters to be output
    * @param length number of characters to be output
    * @exception SAXException for any failure
    */

    public void writeContent(char[] chars, int start, int length) throws SAXException {
        //System.err.println("WriteContent " + this + ":" + new String(chars, start, length) );
        if (length==0) return;
        if (pendingStartTag != -1) {
            flushStartTag();
        }
        emitter.characters(chars, start, length);
    }

    /**
    * Produce text content output. <BR>
    * Special characters are escaped using XML/HTML conventions if the output format
    * requires it.
    * @param chars StringBuffer containing to be output
    * @param start start position of characters to be output
    * @param len number of characters to be output
    * @exception SAXException for any failure
    */

    public void writeContent(StringBuffer chars, int start, int len) throws SAXException {
        //System.err.println("WriteContent " + this + ":" + chars.substring(start, start+length) );
        if (len==0) return;
        if (pendingStartTag != -1) {
            flushStartTag();
        }
        if (len>charbuffer.length) {
            charbuffer = new char[len];
        }
        chars.getChars(start, start+len, charbuffer, 0);
        emitter.characters(charbuffer, 0, len);
    }

    /**
    * Output an element start tag. <br>
    * The actual output of the tag is deferred until all attributes have been output
    * using writeAttribute(). 
    * @param nameCode The element name code
    */

    public void writeStartTag(int nameCode) throws SAXException {
        //System.err.println("Write start tag " + this + " : " + name + " to emitter " + emitter);
        if (pendingStartTag != -1) flushStartTag();
        pendingAttList.clear();
        pendingNSListSize = 0;
        pendingStartTag = nameCode;
        
        // make sure there is a namespace declaration for this start tag
        writeNamespaceDeclaration(namePool.allocateNamespaceCode(nameCode));
    }

	/**
	* Check that the prefix for an attribute is acceptable, returning a substitute
	* prefix if not. The prefix is acceptable unless a namespace declaration has been
	* written that assignes this prefix to a different namespace URI. This method
	* also checks that the attribute namespace has been declared, and declares it
	* if not.
	*/
	
	public int checkAttributePrefix(int nameCode) throws SAXException {
		int nscode = namePool.allocateNamespaceCode(nameCode);
        for (int i=0; i<pendingNSListSize; i++) {
        	if ((nscode>>16) == (pendingNSList[i]>>16)) {
        		// same prefix
        		if ((nscode & 0xffff) == (pendingNSList[i] & 0xffff)) {
        			// same URI
        			return nameCode;	// all is well
        		} else {
        			String prefix = getSubstitutePrefix(nscode);
        			int newCode = namePool.allocate(
        								prefix,
        								namePool.getURI(nameCode),
        								namePool.getLocalName(nameCode));
        			writeNamespaceDeclaration(namePool.allocateNamespaceCode(newCode));
        			return newCode;
        		}
        	}
        }
        // no declaration of this prefix: declare it now
        writeNamespaceDeclaration(nscode);
        return nameCode;
    }
    		
    /**
    * Output a namespace declaration. <br>
    * This is added to a list of pending namespaces for the current start tag.
    * If there is already another declaration of the same prefix, this one is
    * ignored.
    * Note that unlike SAX2 startPrefixMapping(), this call is made AFTER writing the start tag.
    * @param nscode The namespace code 
    * @throws SAXException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void writeNamespaceDeclaration(int nscode)
    throws SAXException {
        
        // System.err.println("Write namespace prefix=" + (nscode>>16) + " uri=" + (nscode&0xffff));
        if (pendingStartTag==-1) {
            throw new SAXException("Cannot write a namespace declaration when there is no open start tag");
        }

        // elimination of namespaces already present on an outer element of the
        // result tree is now done by the NamespaceEmitter. 

        // Ignore declarations whose prefix is duplicated for this element.

        for (int i=0; i<pendingNSListSize; i++) {
        	if ((nscode>>16) == (pendingNSList[i]>>16)) {
        		// same prefix, do a quick exit
        		return;
        	}
        }

        // if it's not a duplicate namespace, add it to the list for this start tag

        if (pendingNSListSize+1 > pendingNSList.length) {
            int[] newlist = new int[pendingNSListSize * 2];
            System.arraycopy(pendingNSList, 0, newlist, 0, pendingNSListSize);
            pendingNSList = newlist;
        }
        pendingNSList[pendingNSListSize++] = nscode;
    }

	/**
	* Copy a namespace node to the current element node
	* (Rules defined in XSLT 1.0 errata)
	*/
	
	public void copyNamespaceNode(int nscode) throws SAXException {
        if (pendingStartTag==-1) {
            throw new SAXException("Cannot copy a namespace node when there is no containing element node");
        }
        if (pendingAttList.getLength()>0) {
        	throw new SAXException("Cannot copy a namespace node to an element after attributes have been added");
        }
        
        // check for duplicates
        for (int i=0; i<pendingNSListSize; i++) {
        	if ((nscode>>16) == (pendingNSList[i]>>16)) {
        		// same prefix
        		if (nscode==pendingNSList[i]) {		// same URI
        			return;							// ignore the duplicate
        		} else {
					// may need to handle default namespace differently
        			throw new SAXException("Cannot create two namespace nodes with the same name");
				}
        	}
        }
        writeNamespaceDeclaration(nscode);
    }        		

    /**
    * It is possible for a single output element to use the same prefix to refer to different
    * namespaces. In this case we have to generate an alternative prefix for uniqueness. The
    * one we generate is based on the uri part of the namespace code itself,
    * which is almost certain to be unique.
    */

    private String getSubstitutePrefix(int nscode) {
    	String prefix = namePool.getPrefixFromNamespaceCode(nscode);
        return prefix + "." + (nscode&0xffff);
    }

    /**
    * Output an attribute value. <br>
    * This is added to a list of pending attributes for the current start tag, overwriting
    * any previous attribute with the same name. <br>
    * This method should NOT be used to output namespace declarations.
    * @param nameCode The name code of the attribute
    * @param value The value of the attribute
    * @throws SAXException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void writeAttribute(int nameCode, String value) throws SAXException {
        writeAttribute(nameCode, value, false);
    }

    /**
    * Output an attribute value. <br>
    * This is added to a list of pending attributes for the current start tag, overwriting
    * any previous attribute with the same name. <br>
    * This method should NOT be used to output namespace declarations.<br>
    * Before calling this, checkAttributePrefix() should be called to ensure the namespace
    * is OK.
    * @param name The name of the attribute
    * @param value The value of the attribute
    * @param noEscape True if it's known there are no special characters in the value. If
    * unsure, set this to false.
    * @throws SAXException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void writeAttribute(int nameCode, String value, boolean noEscape) throws SAXException {
        //System.err.println("Write attribute " + name + "=" + value + " (" + noEscape + ")");
        if (pendingStartTag==-1) {
            throw new SAXException("Cannot write an attribute when there is no open start tag");
        }

        pendingAttList.setAttribute(nameCode, 
              (noEscape ? "NO-ESC" : "CDATA"),  // dummy attribute type to indicate no special chars
              value); 

    }


    /**
    * Output an element end tag.<br>
    * @param nameCode The element name code
    */

    public void writeEndTag(int nameCode) throws SAXException {
        //System.err.println("Write end tag " + this + " : " + name);
        if (pendingStartTag != -1) {
            flushStartTag();
        }
        
        // write the end tag
        emitter.endElement(nameCode);
    }

    /**
    * Write a comment
    */

    public void writeComment(String comment) throws SAXException {
        if (pendingStartTag != -1) flushStartTag();
        emitter.comment(comment.toCharArray(), 0, comment.length());
    }

    /**
    * Write a processing instruction
    */

    public void writePI(String target, String data) throws SAXException {
        if (pendingStartTag != -1) flushStartTag();
        emitter.processingInstruction(target, data);
    }

    /**
    * Close the output
    */

    public void close() throws SAXException {
        //System.err.println("Close " + this + " using emitter " + emitter.getClass());
        emitter.endDocument();
    }

    /**
    * Flush out a pending start tag
    */

    protected void flushStartTag() throws SAXException {
          
        emitter.startElement(pendingStartTag, pendingAttList,
        						 pendingNSList, pendingNSListSize);
        pendingNSListSize = 0;
        pendingStartTag = -1;
            
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
