package com.icl.saxon.style;
import com.icl.saxon.expr.StaticContext;
import com.icl.saxon.expr.Function;
import com.icl.saxon.expr.ExpressionParser;
import com.icl.saxon.expr.FunctionProxy;
import com.icl.saxon.Binding;
import com.icl.saxon.om.Name;
import com.icl.saxon.om.NamePool;
import com.icl.saxon.pattern.NameTest;
import com.icl.saxon.pattern.NamespaceTest;
import org.xml.sax.SAXException;

/**
* An ExpressionContext represents the context for an XPath expression written
* in the stylesheet.
*/

public class ExpressionContext implements StaticContext {
	
	private StyleElement element;
	private NamePool namePool;

	public ExpressionContext(StyleElement styleElement) {
		element = styleElement;
		namePool = styleElement.getNamePool();	// the stylesheet namepool
	}

	/**
	* Create a context for parsing XPath expressions at runtime, using this styelsheet
	* element for namespace declarations, variables, etc, but using the runtime NamePool
	*/
	
	public StaticContext makeRuntimeContext(NamePool pool) {
		ExpressionContext ec = new ExpressionContext(element);
		ec.namePool = pool;
		return ec;
	}

    /**
    * Get the Base URI of the container of the expression
    */

    public String getSystemId() {
    	return element.getSystemId();
    }

    /**
    * Get the line number of the expression within its containing entity
    * Returns -1 if no line number is available
    */

    public int getLineNumber() {
    	return element.getLineNumber();
    }

    /**
    * Get the URI for a prefix, using this Element as the context for namespace resolution
    * @param prefix The prefix
    * @throw SAXException if the prefix is not declared
    */

    public String getURIForPrefix(String prefix) throws SAXException {
    	short uriCode = element.getURICodeForPrefix(prefix);
    	return namePool.getURIFromURICode(uriCode);
    }

    /**
    * Make a NameCode, using this Element as the context for namespace resolution
    * @param qname The name as written, in the form "[prefix:]localname"
    * @boolean useDefault Defines the action when there is no prefix. If true, use
    * the default namespace URI (as for element names). If false, use no namespace URI
    * (as for attribute names).
    */

    public final int makeNameCode(String qname, boolean useDefault) throws SAXException {

        String prefix = Name.getPrefix(qname);
        if (prefix.equals("")) {
            short uriCode = 0;
            
            if (!Name.isNCName(qname)) {
                throw new SAXException("Name " + qname + " contains invalid characters");
            }
            
            if (useDefault) {
                uriCode = element.getURICodeForPrefix(prefix);
            }
            
			return namePool.allocate(prefix, uriCode, qname);
            
        } else {
            String localName = Name.getLocalName(qname);
            short uriCode = element.getURICodeForPrefix(prefix);
			return namePool.allocate(prefix, uriCode, localName);
        }

	}   	

    /**
    * Get a fingerprint for a name, using this as the context for namespace resolution
    * @param qname The name as written, in the form "[prefix:]localname"
    * @boolean useDefault Defines the action when there is no prefix. If true, use
    * the default namespace URI (as for element names). If false, use no namespace URI
    * (as for attribute names).
    * @return -1 if the name is not already present in the name pool
    */

    public int getFingerprint(String qname, boolean useDefault) throws SAXException {

        String prefix = Name.getPrefix(qname);
        if (prefix.equals("")) {
            String uri = "";
            
            if (useDefault) {
                uri = getURIForPrefix(prefix);
            }
            
			return namePool.getFingerprint(uri, qname);
            
        } else {
            String localName = Name.getLocalName(qname);
            String uri = getURIForPrefix(prefix);
			return namePool.getFingerprint(uri, localName);
        }
    }

	/**
	* Make a NameTest, using this element as the context for namespace resolution
	*/
	
	public NameTest makeNameTest(short nodeType, String qname, boolean useDefault)
			throws SAXException {
        int nameCode = makeNameCode(qname, useDefault);			
		return new NameTest(nodeType, nameCode);
	}

	/**
	* Make a NamespaceTest, using this element as the context for namespace resolution
	*/
	
	public NamespaceTest makeNamespaceTest(short nodeType, String prefix)
			throws SAXException {
        short uriCode = element.getURICodeForPrefix(prefix);			
		return new NamespaceTest(nodeType, uriCode);
	}

    /**
    * Bind a variable to an object that can be used to refer to it
    * @param fingerprint The fingerprint of the variable name
    * @return a Binding object that can be used to identify it in the Bindery
    * @throws SAXException if the variable has not been declared, or if the context
    * does not allow the use of variables
    */

    public Binding bindVariable(int fingerprint) throws SAXException {
    	return element.bindVariable(fingerprint);
    }   

    /**
    * Determine whether a given URI code identifies an extension element namespace
    */

    public boolean isExtensionNamespace(short uriCode) throws SAXException {
    	return element.isExtensionNamespace(uriCode);
    }

    /**
    * Determine whether forwards-compatible mode is enabled
    */

    public boolean forwardsCompatibleModeIsEnabled() throws SAXException {
    	return element.forwardsCompatibleModeIsEnabled();
    }

	/*
    * Get a Function declared using a saxon:function element in the stylesheet
    * @param fingerprint the fingerprint of the name of the function
    * @return the Function object represented by this saxon:function; or null if not found
    */

    public Function getStyleSheetFunction(int fingerprint) throws SAXException {
    	return element.getStyleSheetFunction(fingerprint);
    }

    /**
    * Determine if an extension element is available
    */
    
    public boolean isElementAvailable(String qname) throws SAXException {
    	String prefix = Name.getPrefix(qname);
    	String localName = Name.getLocalName(qname);
    	String uri = getURIForPrefix(prefix);
    	return element.getPreparedStyleSheet().
    						getStyleNodeFactory().isElementAvailable(uri, localName);
    }

    /**
    * Determine if a function is available
    */
    
    public boolean isFunctionAvailable(String qname) throws SAXException {
    	String prefix = Name.getPrefix(qname);
    	if (prefix.equals("")) {
    		return ExpressionParser.makeSystemFunction(qname)!=null;
    	}

		int fingerprint = getFingerprint(qname, false);
		if (fingerprint>=0) {
			Function f = getStyleSheetFunction(fingerprint);
			if (f!=null) return true;
		}
	    	
    	String uri = getURIForPrefix(prefix);
    	String localName = Name.getLocalName(qname);
	    	    	
    	FunctionProxy fp = new FunctionProxy();
    	return fp.setMethod(uri, localName);
	}    	

	/**
	* Determine whether the key() function is permmitted in this context
	*/
	
	public boolean allowsKeyFunction() {
		return !(element instanceof XSLKey);
	}

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
