package com.icl.saxon.style;
import com.icl.saxon.*;
//import com.icl.saxon.om.Name;
import com.icl.saxon.om.NamePool;
import com.icl.saxon.om.ExtendedAttributes;
import com.icl.saxon.expr.*;
import com.icl.saxon.output.*;
import org.xml.sax.*;
import java.io.*;
import java.net.URL;
import java.util.StringTokenizer;

/**
* A saxon:output element in the stylesheet. <BR>
* The saxon:output element takes an attribute file="filename". The filename will
* often contain parameters, e.g. {@saxon:nr} to ensure that a different file is produced
* for each element instance. <BR>
* There is a further attribute method=xml|html|text which determines the format of the
* output file (default XML).
* Alternatively the xsl:output element may take a next-in-chain attribute in which case
* output is directed to another stylesheet.
* There is a user-data attribute which is used only to pass information to a user-specified
* Emitter (as part of the OutputDetails). It is an attribute value template.
*/

public class SAXONOutput extends StyleElement {

    Expression file = null  ;
    Expression userData = null;
    Expression method = null;
    Expression version = null;
    Expression indent = null;
    Expression encoding = null;
    Expression mediaType = null;
    Expression doctypeSystem = null;
    Expression doctypePublic = null;
    Expression omitDeclaration = null;
    Expression standalone = null;
    Expression cdataElements = null;
    Expression nextInChain = null;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws SAXException {
        
		StandardNames sn = getStandardNames();
		ExtendedAttributes atts = getAttributeList();
		
	    String methodAtt = null;
	    String versionAtt = null;
	    String indentAtt = null;
	    String encodingAtt = null;
	    String mediaTypeAtt = null;
	    String doctypeSystemAtt = null;
	    String doctypePublicAtt = null;
	    String omitDeclarationAtt = null;
	    String standaloneAtt = null;
	    String cdataElementsAtt = null;
	    String nextInChainAtt = null;
		String userDataAtt = null;
		String fileAtt = null;		
        
        for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.METHOD) {
        		methodAtt = atts.getValue(a);
        	} else if (f==sn.VERSION) {
        		versionAtt = atts.getValue(a);
        	} else if (f==sn.ENCODING) {
        		encodingAtt = atts.getValue(a);
        	} else if (f==sn.OMIT_XML_DECLARATION) {
        		omitDeclarationAtt = atts.getValue(a);
        	} else if (f==sn.STANDALONE) {
        		standaloneAtt = atts.getValue(a);        		        		
        	} else if (f==sn.DOCTYPE_PUBLIC) {
        		doctypePublicAtt = atts.getValue(a);     
        	} else if (f==sn.DOCTYPE_SYSTEM) {
        		doctypeSystemAtt = atts.getValue(a);     
        	} else if (f==sn.CDATA_SECTION_ELEMENTS) {
        		cdataElementsAtt = atts.getValue(a);     
        	} else if (f==sn.INDENT) {
        		indentAtt = atts.getValue(a);     
        	} else if (f==sn.MEDIA_TYPE) {
        		mediaTypeAtt = atts.getValue(a);     
        	} else if (f==sn.FILE) {
        		fileAtt = atts.getValue(a);     
        	} else if (f==sn.USER_DATA) {
        		userDataAtt = atts.getValue(a);     
        	} else if (f==sn.NEXT_IN_CHAIN) {
        		nextInChainAtt = atts.getValue(a);     
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (fileAtt != null) {
            file = makeAttributeValueTemplate(fileAtt);
        }

        if (userDataAtt != null) {
            userData = makeAttributeValueTemplate(userDataAtt);
        }

        if (methodAtt != null) {
            method = makeAttributeValueTemplate(methodAtt);
        }

        if (versionAtt != null) {
            version = makeAttributeValueTemplate(versionAtt);
        }

        if (indentAtt != null) {
            indent = makeAttributeValueTemplate(indentAtt);
        }        

        if (encodingAtt != null) {
            encoding = makeAttributeValueTemplate(encodingAtt);
        }

        if (mediaTypeAtt != null) {
            mediaType = makeAttributeValueTemplate(mediaTypeAtt);
        }

        if (doctypeSystemAtt != null) {
            doctypeSystem = makeAttributeValueTemplate(doctypeSystemAtt);
        }

        if (doctypePublicAtt != null) {
            doctypePublic = makeAttributeValueTemplate(doctypePublicAtt);
        }

        if (omitDeclarationAtt != null) {
            omitDeclaration = makeAttributeValueTemplate(omitDeclarationAtt);
        }

        if (standaloneAtt != null) {
            standalone = makeAttributeValueTemplate(standaloneAtt);
        }

        if (cdataElementsAtt != null) {
            cdataElements = makeAttributeValueTemplate(cdataElementsAtt);
        }

        if (nextInChainAtt != null) {
            nextInChain = makeAttributeValueTemplate(nextInChainAtt);
        }
        
        if (nextInChain!=null && fileAtt!=null) {
            throw styleError("The file attribute and the next-in-chain attribute must not both be present");
        }
        

    }

    public void validate() throws SAXException {
        checkWithinTemplate();
    }
    
    public void process( Context context ) throws SAXException
    {
        Controller c = context.getController();
        OutputDetails prev = c.getCurrentOutputDetails();
        OutputDetails details =  new OutputDetails(prev);
        OutputManager manager = c.getOutputManager();

        if (file != null) {

            // following code to create any directory that doesn't exist is courtesy of
            // Brett Knights [brett@knightsofthenet.com]
            // Modified by MHK to work with JDK 1.1
            
            String outFileName = file.evaluateAsString(context);
            try {
			    File outFile = new File(outFileName);
		        if (!outFile.exists()) {
		            String parent = outFile.getParent();        // always returns null with Microsoft JVM?
		            if (parent!=null) {
        				File parentPath = new File(parent);
        				if (parentPath != null) {
        					if (!parentPath.exists()) {
        						parentPath.mkdirs();
        					}
        				}
    				    outFile.createNewFile();
		            }
			    }

                details.setOutputStream(new FileOutputStream(outFile));
                details.setWriter(null);
            } catch (java.io.IOException err) {
                throw new SAXException("Failed to create output file " + outFileName, err);
            }
        }

        if (userData != null) {
            String data = userData.evaluateAsString(context);
            details.setUserData(data);
        }

        if (method != null) {
            String data = method.evaluateAsString(context);
            if (data.equals("xml") || data.equals("html") || data.equals("text"))  {
                // OK
            } else {
                int methodNameCode = makeNameCode(data, false);
                if (getNamePool().getURICode(methodNameCode)==0) {
                    throw styleError("method must be xml, html, or text, or a prefixed name");
                }
                String localName = getNamePool().getLocalName(methodNameCode);  // don't care what the prefix is
                
                if (localName.equals("fop")) { 
                    details.setEmitter(new FOPEmitter());
                    data="saxon:user";
                } else if (localName.equals("xhtml")) { 
                    data="xhtml";
                } else {
                    details.setEmitter(XSLOutput.makeEmitter(localName));
                    data="saxon:user";
                } 
            }
            details.setMethod(data);
        }

        if (version != null) {
            String data = version.evaluateAsString(context);
            details.setVersion(data);
        }

        if (indent != null) {
            String data = indent.evaluateAsString(context);
            if (data==null || data.equals("yes") || data.equals("no")) {
                details.setIndent(data);
            } else {
                try {
                    int indentSpaces = Integer.parseInt(data);
                    details.setIndent("yes");
                    details.setIndentSpaces(indentSpaces);    
                } catch (NumberFormatException err) {
                    throw styleError("indent must be yes or no or an integer");
                }
            }
        }         

        if (encoding != null) {
            String data = encoding.evaluateAsString(context);
            details.setEncoding(data);
        }    

        if (mediaType != null) {
            String data = mediaType.evaluateAsString(context);
            details.setEncoding(data);
        }    

        if (doctypeSystem != null) {
            String data = doctypeSystem.evaluateAsString(context);
            details.setDoctypeSystem(data);
        }
        
        if (doctypePublic != null) {
            String data = doctypePublic.evaluateAsString(context);
            details.setDoctypePublic(data);
        }   

        if (omitDeclaration != null) {
            String data = omitDeclaration.evaluateAsString(context);
            if (data.equals("yes") || data.equals("no")) {
                details.setOmitDeclaration(data);
            } else {
                throw styleError("omit-xml-declaration attribute must be yes or no");
            }
        }   

        if (standalone != null) {
            String data = standalone.evaluateAsString(context);
            if (data.equals("yes") || data.equals("no")) {
                details.setStandalone(data);
            } else {
                throw styleError("omit-xml-declaration attribute must be yes or no");
            }
        }  

        if (cdataElements != null) {
            String data = cdataElements.evaluateAsString(context);
            StringTokenizer st = new StringTokenizer(data);
            while (st.hasMoreTokens()) {
                String displayname = st.nextToken();
                int namecode = makeNameCode(displayname, true);
                details.addCdataElement(nameCode);
            }
        }

        if (nextInChain != null) {
            String data = nextInChain.evaluateAsString(context);
            Controller nextStyleSheet = prepareNextStylesheet(data, context);
            ContentHandlerProxy emitter = new ContentHandlerProxy();
            emitter.setUnderlyingContentHandler(nextStyleSheet.getInputContentHandler());
            emitter.setRequireWellFormed(false);
            details.setEmitter(emitter);
            details.setMethod("saxon:user");
        }  
        
        if (details.getMethod()==null) {             
            details.setMethod("saxon:uncommitted");
        }

        
        manager.setOutputDetails(details);        
        processChildren(context);
        manager.resetOutputDetails();     
    }

    /**
    * Prepare another stylesheet to handle the output of this one
    */

    private Controller prepareNextStylesheet(String href, Context context) throws SAXException {

        URL baseURL;
        URL includedURL;
        try {
            baseURL = new URL(getSystemId());
        } catch (java.net.MalformedURLException err) {
            throw new SAXException("saxon:output - invalid base URL " + getSystemId());
        }
        try {
            includedURL = new URL(baseURL, href);
        } catch (java.net.MalformedURLException err) {
            throw new SAXException("xsl:include - invalid target URL " + href);
        }
        
		NamePool pool = context.getController().getNamePool();
        ExtendedInputSource nextSheet = new ExtendedInputSource(includedURL.toString());

        PreparedStyleSheet next = new PreparedStyleSheet();
                        // use the same stylesheet parser and namepool as before
        next.setXMLReader( getPreparedStyleSheet().getXMLReader() );
        next.setNamePool(pool); 
        next.prepare(nextSheet);

        Controller nextTransformer = (Controller)next.newTransformer();
        nextTransformer.setNamePool(pool);
        return nextTransformer;
    }
       
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Additional Contributor(s): Brett Knights [brett@knightsofthenet.com]
//
