package com.icl.saxon.style;
import com.icl.saxon.*;
import com.icl.saxon.om.*;
import com.icl.saxon.tree.ElementWithAttributes;
import com.icl.saxon.tree.DocumentImpl;
import com.icl.saxon.expr.*;
import com.icl.saxon.pattern.Pattern;
import com.icl.saxon.handlers.NodeHandler;
import com.icl.saxon.output.*;
import com.icl.saxon.trace.*;  // e.g.

import org.xml.sax.*;
import org.w3c.dom.Node;

import java.util.*;
import java.io.*;
import java.text.*;

/**
* Abstract superclass for all element nodes in the stylesheet. <BR>
* Note: this class implements Locator. The element
* retains information about its own location in the stylesheet, which is useful when
* an XSL error is found. 
*/

public abstract class StyleElement extends ElementWithAttributes
        implements Locator {

    protected Vector attributeSets = null;
    private short[] extensionNamespaces = null;		// a list of URI codes
    private short[] excludedNamespaces = null;		// a list of URI codes
    protected String version = null;
    protected SAXException validationError = null;	// used for fallback processing

    /**
    * Constructor creates an empty attribute list
    */

    public StyleElement() {}

    /**
    * Make this node a substitute for a temporary one previously added to the tree. See
    * StyleNodeFactory for details. "A node like the other one in all things but its class".
    * Note that at this stage, the node will not yet be known to its parent, though it will
    * contain a reference to its parent; and it will have no children.
    */

    public void substituteFor(StyleElement temp) throws SAXException {
        this.parent = temp.parent;
        this.attributeList = temp.attributeList;
        this.namespaceList = temp.namespaceList;
        this.nameCode = temp.nameCode;
        this.sequence = temp.sequence;
        this.attributeSets = temp.attributeSets;
        this.extensionNamespaces = temp.extensionNamespaces;
        this.excludedNamespaces = temp.excludedNamespaces;
        this.version = temp.version;
        this.root = temp.root;
    }       

	/**
	* Set a validation error
	*/
	
	protected void setValidationError(SAXException reason) {
		validationError = reason;
	}

    /**
    * Determine whether this node is an instruction. The default implementation says it isn't.
    */

    public boolean isInstruction() {
        return false;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return false;
    }

	/**
	* Get the containing XSLStyleSheet element
	*/
	
	public XSLStyleSheet getContainingStyleSheet() {
		NodeInfo next = this;
		while (!(next instanceof XSLStyleSheet)) {
			next = (NodeInfo)next.getParentNode();
		}
		return (XSLStyleSheet)next;
	}

    /**
    * Get the import precedence of this stylesheet element. 
    */

    public int getPrecedence() throws SAXException {
        return getContainingStyleSheet().getPrecedence();
    }

	/**
	* Get the StandardNames object
	*/
	
	public final StandardNames getStandardNames() throws SAXException {
		DocumentImpl root = (DocumentImpl)getDocumentRoot();
		return ((StyleNodeFactory)root.getNodeFactory()).getStandardNames();
	}

    /**
    * Process the attributes of this element and all its children
    */

    public void processAllAttributes() throws SAXException {
        processAttributes();
        NodeInfo child = (NodeInfo)getFirstChild();
        while (child != null) {
            if (child instanceof StyleElement) {
                ((StyleElement)child).processAllAttributes();          
            }
            child = (NodeInfo)child.getNextSibling();
        }
    }

    /**
    * Process the attribute list for the element. This is a wrapper method that calls
    * prepareAttributes (provided in the subclass) and traps any exceptions
    */

    public final void processAttributes() throws SAXException {
        try {
            prepareAttributes();
        } catch (SAXException err) {
        	if (forwardsCompatibleModeIsEnabled()) {
        		validationError = err;
        	} else {
            	throw styleError(err);
            }
        }
    }

    /**
    * Check whether an unknown attribute is permitted.
    * @param nc The name code of the attribute name
    */
    
    protected void checkUnknownAttribute(int nc) throws SAXException {
    	if (forwardsCompatibleModeIsEnabled()) {
    		// then unknown attributes are permitted and ignored
    		return;
    	}
    	String attributeURI = getNamePool().getURI(nc);
    	String elementURI = getURI();
    	int attributeFingerprint = nc & 0xfffff;
    	StandardNames sn = getStandardNames();
    	
    	// allow xsl:extension-element-prefixes etc on an extension element

    	if (isInstruction() &&
    		 attributeURI.equals(Namespace.XSLT) &&
    		 !(elementURI.equals(Namespace.XSLT)) &&
    		 (attributeFingerprint == sn.XSL_EXTENSION_ELEMENT_PREFIXES ||
    		  attributeFingerprint == sn.XSL_EXCLUDE_RESULT_PREFIXES ||
    		  attributeFingerprint == sn.XSL_VERSION)) {
    		return;
    	}
    	
    	if (attributeURI.equals("") || attributeURI.equals(Namespace.XSLT)) {
			throw styleError("Attribute " + getNamePool().getDisplayName(nc) +
				 " is not allowed on this element");
        }
    }    		
    

    /**
    * Set the attribute list for the element. This is called to process the attributes (note
    * the distinction from processAttributes in the superclass).
    * Must be supplied in a subclass
    */

    public abstract void prepareAttributes() throws SAXException;

	/**
	* Make an expression in the context of this stylesheet element
	*/
	
	public Expression makeExpression(String expression) throws SAXException {
		ExpressionContext env = new ExpressionContext(this);
		return Expression.make(expression, env);
	}

	/**
	* Make a pattern in the context of this stylesheet element
	*/
	
	public Pattern makePattern(String pattern) throws SAXException {
		ExpressionContext env = new ExpressionContext(this);
		return Pattern.make(pattern, env);
	}

	/**
	* Make an attribute value template in the context of this stylesheet element
	*/
	
	public Expression makeAttributeValueTemplate(String expression) throws SAXException {
		ExpressionContext env = new ExpressionContext(this);
		return AttributeValueTemplate.make(expression, env);
	}

    /**
    * Process the [xsl:]extension-element-prefixes attribute if there is one
    * @param nc the name code of the attribute required
    */

    protected void processExtensionElementAttribute(int nc) throws SAXException {
        String ext = getAttributeValue(nc & 0xfffff);
        if (ext!=null) {
        	// go round twice, once to count the values and next to add them to the array
        	int count = 0;
            StringTokenizer st1 = new StringTokenizer(ext);
            while (st1.hasMoreTokens()) {
                String s = st1.nextToken();
                count++;
            }       
			extensionNamespaces = new short[count];
			count = 0;
            StringTokenizer st2 = new StringTokenizer(ext);
            while (st2.hasMoreTokens()) {
                String s = st2.nextToken();
                if (s.equals("#default")) {
                	s = "";
                }
                short uriCode = getURICodeForPrefix(s);
                extensionNamespaces[count++] = uriCode;
            }            
        }
    }

    /**
    * Process the [xsl:]exclude-result-prefixes attribute if there is one
    * @param nc the name code of the attribute required
    */

    protected void processExcludedNamespaces(int nc) throws SAXException {
        String ext = getAttributeValue(nc & 0xfffff);
        if (ext!=null) {
        	// go round twice, once to count the values and next to add them to the array
        	int count = 0;
            StringTokenizer st1 = new StringTokenizer(ext);
            while (st1.hasMoreTokens()) {
                String s = st1.nextToken();
                count++;
            }       
			excludedNamespaces = new short[count];
			count = 0;
            StringTokenizer st2 = new StringTokenizer(ext);
            while (st2.hasMoreTokens()) {
                String s = st2.nextToken();
                if (s.equals("#default")) {
                	s = "";
                }
                short uriCode = getURICodeForPrefix(s);
                excludedNamespaces[count++] = uriCode;
            }            
        }
    }

    /**
    * Process the [xsl:]version attribute if there is one
    * @param nc the name code of the attribute required
    */

    protected void processVersionAttribute(int nc) throws SAXException {
        version = getAttributeValue(nc & 0xfffff);
    }

    /**
    * Get the version number on this element, or inherited from its ancestors
    */

    public String getVersion() throws SAXException {
        if (version==null) {
            NodeInfo node = (NodeInfo)getParentNode();
            if (node instanceof StyleElement) {
                version = ((StyleElement)node).getVersion();
            } else {
                version = "1.0";
            }
        } 
        return version;
    }

    /**
    * Determine whether forwards-compatible mode is enabled for this element
    */

    public boolean forwardsCompatibleModeIsEnabled() throws SAXException {
        return !(getVersion().equals("1.0"));
    }
        
    /**
    * Check whether a particular extension element namespace is defined on this node.
    * This checks this node only, not the ancestor nodes.
    * The implementation checks whether the prefix is included in the
    * [xsl:]extension-element-prefixes attribute.
    * @param uriCode the namespace URI code being tested
    */

    protected boolean definesExtensionElement(short uriCode) throws SAXException {
    	if (extensionNamespaces==null) {
    		return false;
    	}
    	for (int i=0; i<extensionNamespaces.length; i++) {
    		if (extensionNamespaces[i] == uriCode) {
    			return true;
    		}
    	}
        return false;
    }

    /**
    * Check whether a namespace uri defines an extension element. This checks whether the
    * namespace is defined as an extension namespace on this or any ancestor node.
    * @param uriCode the namespace URI code being tested
    */

    public boolean isExtensionNamespace(short uriCode) throws SAXException {
        NodeInfo anc = this;
        while (anc instanceof StyleElement) {
            if (((StyleElement)anc).definesExtensionElement(uriCode)) {
                return true;
            }
            anc = (NodeInfo)anc.getParentNode();
        }
        return false;
    }

    /**
    * Check whether this node excludes a particular namespace from the result.
    * This method checks this node only, not the ancestor nodes.
    * @param uriCode the code of the namespace URI being tested
    */

    protected boolean definesExcludedNamespace(short uriCode) throws SAXException {
    	if (excludedNamespaces==null) {
    		return false;
    	}
    	for (int i=0; i<excludedNamespaces.length; i++) {
    		if (excludedNamespaces[i] == uriCode) {
    			return true;
    		}
    	}
        return false;
    }

    /**
    * Check whether a namespace uri defines an namespace excluded from the result.
    * This checks whether the namespace is defined as an excluded namespace on this
    * or any ancestor node.
    * @param uriCode the code of the namespace URI being tested
    */

    public boolean isExcludedNamespace(short uriCode) throws SAXException {
		if (uriCode==Namespace.XSLT_CODE) return true;
        if (isExtensionNamespace(uriCode)) return true;        
        NodeInfo anc = this;
        while (anc instanceof StyleElement) {
            if (((StyleElement)anc).definesExcludedNamespace(uriCode)) {
                return true;
            }
            anc = (NodeInfo)anc.getParentNode();
        }
        return false;
    }

    /**
    * Check that the element is valid. This is called once for each element, after
    * the entire tree has been built. As well as validation, it can perform first-time
    * initialisation. The default implementation does nothing; it is normally overriden
    * in subclasses.
    */

    public void validate() throws SAXException {}

    /**
    * Default preprocessing method does nothing. It is implemented for those top-level elements
    * that can be evaluated before the source document is available, for example xsl:key,
    * xsl:attribute-set, xsl:template, xsl:locale
    */

    public void preprocess() throws SAXException {}

    /**
    * Recursive walk through the stylesheet to validate all nodes
    */

    public void validateSubtree() throws SAXException {
        try {
            validate();
        } catch (SAXException err) {
        	if (forwardsCompatibleModeIsEnabled()) {
        		validationError = err;
        	} else {
            	throw styleError(err);
            }
        }
        NodeInfo child = (NodeInfo)getFirstChild();
        while (child != null) {
            if (child instanceof StyleElement) {
                ((StyleElement)child).validateSubtree();
            }
            child = (NodeInfo)child.getNextSibling();
        }
    }

    /**
    * Get the principal XSLStyleSheet node. This gets the principal style sheet, i.e. the
    * one originally loaded, that forms the root of the import/include tree
    */

    protected XSLStyleSheet getPrincipalStyleSheet() throws SAXException {
        XSLStyleSheet sheet = getContainingStyleSheet();
        while (true) {
            XSLStyleSheet next = sheet.getImporter();
            if (next==null) return sheet;
            sheet = next;
        }
    }

    /**
    * Get the PreparedStyleSheet object.
    * @return the PreparedStyleSheet to which this stylesheet element belongs
    */

    public PreparedStyleSheet getPreparedStyleSheet() throws SAXException {
        return getPrincipalStyleSheet().getPreparedStyleSheet();
    }

    /**
    * Check that the stylesheet element is within a template body
    * @throws SAXException if not within a template body
    */

    public void checkWithinTemplate() throws SAXException {
        StyleElement parent = (StyleElement)getParentNode();
        if (!parent.mayContainTemplateBody()) {
            throw styleError("Element must only be used within a template body");
        }
    }

    /**
    * Convenience method to check that the stylesheet element is at the top level
    * @throws SAXException if not at top level
    */

    public void checkTopLevel() throws SAXException {
        if (!(getParentNode() instanceof XSLStyleSheet)) {
            throw styleError("Element must only be used at top level of stylesheet");
        }
    }

    /**
    * Convenience method to check that the stylesheet element is not at the top level
    * @throws SAXException if it is at the top level
    */

    public void checkNotTopLevel() throws SAXException {
        if (getParentNode() instanceof XSLStyleSheet) {
            throw styleError("Element must not be used at top level of stylesheet");
        }
    }

    /**
    * Convenience method to check that the stylesheet element is empty
    * @throws SAXException if it is not empty
    */

    public void checkEmpty() throws SAXException {
        if (getFirstChild()!=null) {
            throw styleError("Element must be empty");
        }
    }

    /**
    * Convenience method to report the absence of a mandatory attribute
    * @throws SAXException if the attribute is missing
    */

    public void reportAbsence(String attribute) throws SAXException {
        throw styleError("Element must have a \"" + attribute + "\" attribute");
    }

    /**
    * Process: called to do the real work of this stylesheet element. This method
    * must be implemented in each subclass.
    * @param context The context in the source XML document, giving access to the current node,
    * the current variables, etc.
    */

    public abstract void process(Context context) throws SAXException;

    /**
    * Process the children of this node in the stylesheet
    * @param context The context in the source XML document, giving access to the current node,
    * the current variables, etc.
    */

    public void processChildren(Context context) throws SAXException {

    	if (context.getController().isTracing()) { // e.g.
    	    TraceListener listener = context.getController().getTraceListener();

    	    NodeInfo node = (NodeInfo)getFirstChild();
    	    while (node!=null) {

        		listener.enter(node, context);

        		if (node instanceof TextInfo) {
        		    node.copy(context.getOutputter());
        		} else if (node instanceof StyleElement) {
        		    StyleElement snode = (StyleElement)node;
        		    if (snode.validationError != null) {
        		    	fallbackProcessing(snode, context);
        		    } else {
	        		    try {
	        			    snode.process(context);
	        		    } catch (SAXException err) {
	        			    throw snode.styleError(err);
	        		    }
	        		}                    
        		}

        		listener.leave(node, context);
        		node = (NodeInfo)node.getNextSibling();
    	    }

    	} else {

    	    NodeInfo node = (NodeInfo)getFirstChild();
    	    while (node!=null) {

        		if (node instanceof TextInfo) {
        		    node.copy(context.getOutputter());
        		} else if (node instanceof StyleElement) {
        		    StyleElement snode = (StyleElement)node;
        		    if (snode.validationError != null) {
        		    	fallbackProcessing(snode, context);
        		    } else {
	        		    try {
	        			    snode.process(context);
	        		    } catch (SAXException err) {
	        			    throw snode.styleError(err);
	        		    }
	        		}                    
        		}
        		node = (NodeInfo)node.getNextSibling();
    	    }

    	}
    }

    // diagnostic version of process()

    private void processX(Context context) throws SAXException {
        System.err.println("Processing " + context.getCurrentNode() + " using " + this);
        process(context);
    }

	/**
	* Perform fallback processing
	*/
	
	protected void fallbackProcessing(StyleElement instruction, Context context) throws SAXException {
        // process any xsl:fallback children; if there are none, report the original failure reason
        
        XSLFallback fallback = null;
        Node child = instruction.getFirstChild();
        while (child!=null) {
            if (child instanceof XSLFallback) {
                fallback = (XSLFallback)child;
                break;
            }
            child = child.getNextSibling();
        }
        
        if (fallback==null) {
        	throw instruction.styleError(instruction.validationError);
        }

        boolean tracing = context.getController().isTracing();

        while (child!=null) {
            if (child instanceof XSLFallback) {
                XSLFallback f = (XSLFallback)child;
                
                if (tracing) {
                    TraceListener listener = context.getController().getTraceListener();
            		listener.enter(f, context);
            		f.process(context);
            		listener.leave(f, context);
                } else {
                    f.process(context);
                }
            }
            child = child.getNextSibling();
        }
		
		
	}

    /**
    * Modify the "select" expression to include any sort keys specified. Used in XSLForEach
    * and XSLApplyTemplates
    */

    protected Expression handleSortKeys(Expression select) throws SAXException {
        // handle sort keys if any

        SortedSelection sortExpression = null;
        Node child = getFirstChild();

        while(child!=null) {
            if (child instanceof XSLSort) {
                if (sortExpression==null) {
                    sortExpression = new SortedSelection(select);                
                } 
                sortExpression.addSortKey(
                    ((XSLSort)child).getSortKeyDefinition());               
            }
            child = child.getNextSibling();
        }
        if (sortExpression == null) {
            return new NodeListExpression(select);  // sorts into document order
        } else {
            return sortExpression;
        }
    }

    /**
    * Expand the stylesheet elements subordinate to this one, returning the result
    * as a string. The expansion must not generate any element or attribute nodes.
    * @param context The context in the source document
    */

    public String expandChildren(Context context) throws SAXException {
        Controller c = context.getController();
        TextFragment details = new TextFragment();
        c.setNewOutputDetails(details);
        processChildren(context);
        c.resetOutputDetails();
        return details.getText();
    }


    /**
    * Determine the list of attribute-sets associated with this element.
    * This is used for xsl:element, xsl:copy, xsl:attribute-set, and on literal
    * result elements
    */

    protected void findAttributeSets(String use) throws SAXException {

        attributeSets = new Vector();
        
        XSLStyleSheet stylesheet = getPrincipalStyleSheet();
        Vector toplevel = stylesheet.getTopLevel();

        StringTokenizer st = new StringTokenizer(use);
        while (st.hasMoreTokens()) {
            String asetname = st.nextToken();
            int fprint = makeNameCode(asetname, false) & 0xfffff;
            boolean found = false;

            // search for the named attribute set, using all of them if there are several with the
            // same name

            for (int i=0; i<toplevel.size(); i++) {
                if (toplevel.elementAt(i) instanceof XSLAttributeSet) {
                    XSLAttributeSet t = (XSLAttributeSet)toplevel.elementAt(i);
                    if (t.getAttributeSetFingerprint() == fprint) {  
                        attributeSets.addElement(t);
                        found = true;
                    }
                }
            }
	        
            if (!found) {
                throw styleError("No attribute-set exists named " + asetname);
            }            
        }
    }

    /**
    * Expand the attribute sets referenced in this element's use-attribute-sets attribute
    */

    protected void processAttributeSets(Context context) throws SAXException {
        if (attributeSets==null) return;
        for (int i=0; i<attributeSets.size(); i++) {
            XSLAttributeSet aset = (XSLAttributeSet)attributeSets.elementAt(i);
            aset.expand(context);
        }
    }

    /**
    * Construct an exception with diagnostic information
    */

    protected SAXException styleError(SAXException error) {
        if (error instanceof StyleException) return error;
        if (error instanceof TerminationException) return error;
        if (error.getException()!=null) {
            return new StyleException(error.getException(),
                                      positionMessage(error.getMessage()));
        }
        return new StyleException(error, positionMessage(error.getMessage()));
    }

    protected SAXException styleError(String message) {
        return new StyleException(positionMessage(message));
    }

    private String positionMessage(String message) {
        String name = getDisplayName();
        if (name.equals("")) {
            name = "element";
        }
        String systemId = getSystemId();
        return "At " + name +
               (systemId==null ? "" : " on line " + getLineNumber() + " of " + systemId) +
               ": " + message;
    }

    /**
    * Test whether this is a top-level element
    */

    public boolean isTopLevel() throws SAXException {
        return (getParentNode() instanceof XSLStyleSheet);
    }

    /**
    * Bind a variable used in this element to the XSLVariable element in which it is declared
    * @param fingerprint The fingerprint of the name of the variable 
    * @return a Binding for the variable
    * @throw SAXException if the variable has not been declared
    */

    public Binding bindVariable(int fingerprint) throws SAXException {
        Binding binding = getVariableBinding(fingerprint);
        if (binding==null) {
            throw styleError("Variable " + getNamePool().getDisplayName(fingerprint) + " has not been declared");
        }
        return binding;
    }

    /**
    * Bind a variable used in this element to the XSLVariable element in which it is declared
    * @name The absolute name of the variable (prefixed by namespace URI)
    * @return a Binding for the variable, or null if it has not been declared
    */

    public Binding getVariableBinding(int fprint) throws SAXException {
        NodeInfo curr = this;
        NodeInfo prev = this;

        // first search for a local variable declaration
        
        if (!isTopLevel()) {
            while (true) {
                curr = (NodeInfo)curr.getPreviousSibling();
                while (curr==null) {
                    curr = (NodeInfo)prev.getParentNode();
                    prev = curr;
                    if (curr.getParentNode() instanceof XSLStyleSheet) break;   // top level
                    curr = (NodeInfo)curr.getPreviousSibling();
                }
                if (curr.getParentNode() instanceof XSLStyleSheet) break;
                if (curr instanceof Binding) {
                    int var = ((Binding)curr).getVariableFingerprint();
                    if (var==fprint) {
                        return (Binding)curr;
                    }
                }
            }
        }
       
        // Now check for a global variable
        // we rely on the search following the order of decreasing import precedence.
        
        XSLStyleSheet root = getPrincipalStyleSheet();
        Vector toplevel = root.getTopLevel();
        for (int i=toplevel.size()-1; i>=0; i--) {
            Object child = toplevel.elementAt(i);
            if (child instanceof Binding && child != this) {
                int var = ((Binding)child).getVariableFingerprint();
                if (var==fprint) {
                    return (Binding)child;
                }
            }
        }
        
        return null;
    }

    /**
    * List the variables that are in scope for this stylesheet element. 
    * Designed for a debugger, not used by the processor.
    * @return two Enumeration of Strings, the global ones [0] and the local ones [1]
    */
    
    public Enumeration[] getVariableNames() throws SAXException {  // e.g.
        Hashtable local = new Hashtable();
        Hashtable global = new Hashtable();

        NodeInfo curr = this;
        NodeInfo prev = this;
        NamePool pool = getNamePool();

        // first collect the local variable declarations
        
        if (!isTopLevel()) {
            while (true) {
                curr = (NodeInfo)curr.getPreviousSibling();
                while (curr==null) {
                    curr = (NodeInfo)prev.getParentNode();
                    prev = curr;
                    if (curr.getParentNode() instanceof XSLStyleSheet) break;   // top level
                    curr = (NodeInfo)curr.getPreviousSibling();
                }
                if (curr.getParentNode() instanceof XSLStyleSheet) break;
                if (curr instanceof Binding) {
                	int fprint = ((Binding)curr).getVariableFingerprint();
                	String uri = pool.getURI(fprint);
                	String lname = pool.getLocalName(fprint);
                    String varname = uri + "^" + lname;
        		    if (local.get(varname)==null) {
        			    local.put(varname, varname);
        		    }
                }
            }
        }

        // Now collect the global variables
        // we rely on the search following the order of increasing import precedence.
        
        XSLStyleSheet root = getPrincipalStyleSheet();
        Vector toplevel = root.getTopLevel();
        for (int i=0; i<toplevel.size(); i++) {
            Object child = toplevel.elementAt(i);
            if (child instanceof Binding && child != this) {
            	int fprint = ((Binding)curr).getVariableFingerprint();
            	String uri = pool.getURI(fprint);
            	String lname = pool.getLocalName(fprint);
                String varname = uri + "^" + lname;
        		if (local.get(varname)==null) {
        		    global.put(varname, varname);
        		}
            }
        }

    	Enumeration info[] = new Enumeration[2];
    	info[0] = global.keys();
    	info[1] = local.keys();
        return info;
    }


    /**
    * Get a Function declared using a saxon:function element in the stylesheet
    * @param fingerprint the fingerprint of the name of the function
    * @return the Function object represented by this saxon:function; or null if not found
    */

    public Function getStyleSheetFunction(int fingerprint) throws SAXException {

        // we rely on the search following the order of decreasing import precedence.
        
        XSLStyleSheet root = getPrincipalStyleSheet();
        Vector toplevel = root.getTopLevel();
        for (int i=toplevel.size()-1; i>=0; i--) {
            Object child = toplevel.elementAt(i);
            if (child instanceof SAXONFunction &&
                    ((SAXONFunction)child).getFunctionFingerprint() == fingerprint) {
                StyleSheetFunctionCall fc = new StyleSheetFunctionCall();
                fc.setFunction((SAXONFunction)child);
                return fc;
            }
        }
        return null;
    }

	/** needed to satisfy Locator interface
	*/
	
	public int getColumnNumber() {
		return -1; 
	}
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): 
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
