package com.icl.saxon.style;
import com.icl.saxon.*;
import com.icl.saxon.om.*;
import com.icl.saxon.handlers.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.output.*;
import org.xml.sax.SAXException;
import java.io.*;
import java.util.*;

/**
* xsl:attribute element in stylesheet.<BR>
*/

public class XSLAttribute extends StyleElement {

    private Expression attributeName;
    private Expression namespace=null;
    private boolean disable = false;
    
    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws SAXException {

		StandardNames sn = getStandardNames();
		ExtendedAttributes atts = getAttributeList();
		
		String nameAtt = null;
		String namespaceAtt = null;
		String disableAtt = null;
		
		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		nameAtt = atts.getValue(a);
        	} else if (f==sn.NAMESPACE) {
        		namespaceAtt = atts.getValue(a);
        	} else if (f==sn.SAXON_DISABLE_OUTPUT_ESCAPING) {
        		disableAtt = atts.getValue(a);
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (nameAtt==null) {
            reportAbsence("name");
        }
        attributeName = makeAttributeValueTemplate(nameAtt);

        if (namespaceAtt!=null) {
            namespace = makeAttributeValueTemplate(namespaceAtt);
        }

        disable = (disableAtt != null && disableAtt.equals("yes"));

    }

    public void validate() throws SAXException {
        if (getParentNode() instanceof XSLAttributeSet) return;
        checkWithinTemplate();
    }

    public void process( Context context ) throws SAXException
    {       
        String expandedName = attributeName.evaluateAsString(context);
        Controller controller = context.getController();
        NamePool pool = controller.getNamePool();      
        
        if (!Name.isQName(expandedName)) {
            controller.reportRecoverableError(
                "Invalid attribute name: " + expandedName, this);
            return;
        }

        if (expandedName.equals("xmlns") || 
        		(expandedName.length()>6 && expandedName.substring(0,6).equals("xmlns:"))) {
            controller.reportRecoverableError(
                "Invalid attribute name: " + expandedName, this);
            return;
        }

        String prefix = Name.getPrefix(expandedName);
        short uriCode;
        
        if (namespace==null) {
        	
        	// NB, we can't just call makeNameCode() because that would use the wrong
        	// name pool
        	
        	if (prefix.equals("")) {
        		uriCode = 0;
        	} else {
        		uriCode = getURICodeForPrefix(prefix);  // fails if not declared
        	}
        	        	
        } else {

            // generate a name using the supplied namespace URI
            
            String uri = namespace.evaluateAsString(context);
            uriCode = pool.allocateCodeForURI(uri);   // allocate code in run-time name pool

            if (prefix.equals("")) {
                prefix = getPrefixForURI(uri);
                if (prefix==null) {
                    prefix="ns0";       // arbitrary generated prefix; will be changed later if
                                        // not unique
                }
            } 
        }
            
        String localName = Name.getLocalName(expandedName);
        int nameCode = pool.allocate(prefix, uriCode, localName);

        Outputter out = controller.getOutputter();

    	// we may need to change the namespace prefix if the one we chose is
    	// already in use with a different namespace URI
    	if (((nameCode>>20)&0xff) != 0) {	// prefix!=""
        	nameCode = out.checkAttributePrefix(nameCode);
        }   
        out.writeAttribute(nameCode, expandChildren(context), disable);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
