package com.icl.saxon.style;
import com.icl.saxon.*;
import com.icl.saxon.om.ExtendedAttributes;
import com.icl.saxon.expr.*;
import com.icl.saxon.output.*;
import org.xml.sax.*;
import java.io.*;
import java.util.*;
import java.text.*;

/**
* An xsl:output element in the stylesheet. <BR>
* This is not to be confused with the xsl:output supported in earlier versions of SAXON,
* which was a precursor to the present saxon:output element. <br>
*/

public class XSLOutput extends StyleElement {

    String method = null;
    String version = null;
    String indent = null;
    String encoding = null;
    String mediaType = null;
    String doctypeSystem = null;
    String doctypePublic = null;
    String omitDeclaration = null;
    String standalone = null;
    Vector cdataElements = null;
    String omitMetaTag = null;
    String representation = null;
    Emitter handler = null;

    public void prepareAttributes() throws SAXException {

		StandardNames sn = getStandardNames();
		ExtendedAttributes atts = getAttributeList();
		
		String cdataAtt = null;		
        
        for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.METHOD) {
        		method = atts.getValue(a);
        	} else if (f==sn.VERSION) {
        		version = atts.getValue(a);
        	} else if (f==sn.ENCODING) {
        		encoding = atts.getValue(a);
        	} else if (f==sn.OMIT_XML_DECLARATION) {
        		omitDeclaration = atts.getValue(a);
        	} else if (f==sn.STANDALONE) {
        		standalone = atts.getValue(a);        		        		
        	} else if (f==sn.DOCTYPE_PUBLIC) {
        		doctypePublic = atts.getValue(a);     
        	} else if (f==sn.DOCTYPE_SYSTEM) {
        		doctypeSystem = atts.getValue(a);     
        	} else if (f==sn.CDATA_SECTION_ELEMENTS) {
        		cdataAtt = atts.getValue(a);     
        	} else if (f==sn.INDENT) {
        		indent = atts.getValue(a);     
        	} else if (f==sn.MEDIA_TYPE) {
        		mediaType = atts.getValue(a);     
        	} else if (f==sn.SAXON_OMIT_META_TAG) {
        		omitMetaTag = atts.getValue(a);     
        	} else if (f==sn.SAXON_CHARACTER_REPRESENTATION) {
        		representation = atts.getValue(a);     
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (method != null) {
            if (method.equals("xml") || method.equals("html") || method.equals("text"))  {
                // OK
            } else {
                int methodNameCode = makeNameCode(method, false);
                if (getNamePool().getURICode(methodNameCode)==0) {
                    throw styleError("method must be xml, html, or text, or a prefixed name");
                }
                String localName = getNamePool().getLocalName(methodNameCode);  // don't care what the prefix is
                
                if (localName.equals("fop")) { 
                    handler = new FOPEmitter();
                    method="saxon:user";
                } else if (localName.equals("xhtml")) { 
                    method="xhtml";
                } else {
                    handler = makeEmitter(localName);
                    method="saxon:user";
                } 
            }
        }

        indent = getIndent(indent);    // different for xsl:output and saxon:output

        if (omitDeclaration != null) {
            if (omitDeclaration.equals("yes") || omitDeclaration.equals("no")) {
                // OK
            } else {
                throw styleError("omit-xml-declaration attribute must be yes or no");
            }
        }

        if (standalone != null) {
            if (standalone.equals("yes") || standalone.equals("no")) {
                // OK
            } else {
                throw styleError("standalone attribute must be yes or no");
            }
        }
        
        if (cdataAtt != null) {
            cdataElements = new Vector();
            StringTokenizer st = new StringTokenizer(cdataAtt);
            while (st.hasMoreTokens()) {
                String displayname = st.nextToken();
                int nameCode = makeNameCode(displayname, true);
                cdataElements.addElement(new Integer(nameCode));
            }
        }
    }

    /**
    * Get indent value. 
    */

    protected String getIndent(String indentAtt) throws SAXException {
        String att = getAttributeValue("indent");
        if (indentAtt==null || indentAtt.equals("yes") || indentAtt.equals("no")) {
        	return indentAtt;
        }
        throw styleError("indent must be yes or no");
    }

    public void validate() throws SAXException {
        checkTopLevel();
        checkEmpty();
    }
    
    public void process( Context context ) throws SAXException {}

    public void setDetails(OutputDetails details) throws SAXException {
        if (method!=null) details.setMethod(method);
        if (version!=null) details.setVersion(version);
        if (indent!=null) details.setIndent(indent);
        if (encoding!=null) details.setEncoding(encoding);
        if (mediaType!=null) details.setMediaType(mediaType);
        if (doctypeSystem!=null) details.setDoctypeSystem(doctypeSystem);
        if (doctypePublic!=null) details.setDoctypePublic(doctypePublic);
        if (omitDeclaration!=null) details.setOmitDeclaration(omitDeclaration);
        if (standalone!=null) details.setStandalone(standalone);
        if (cdataElements!=null) details.addCdataElements(cdataElements);
        if (handler!=null) details.setEmitter(handler);
        if (representation!=null) details.setCharacterRepresentation(representation);
        if (omitMetaTag!=null) details.setIncludeHtmlMetaTag(!omitMetaTag.equals("yes"));
    }
    
    /**
    * load a named output emitter or document handler and check it is OK.
    */

    public static Emitter makeEmitter (String className) throws SAXException
    {

        Object handler = Loader.getInstance(className);            

        if (handler instanceof Emitter) {
            return (Emitter)handler;
        } else if (handler instanceof DocumentHandler) {
            DocumentHandlerProxy emitter = new DocumentHandlerProxy();
            emitter.setUnderlyingDocumentHandler((DocumentHandler)handler);
            return emitter;
        } else if (handler instanceof ContentHandler) {
            ContentHandlerProxy emitter = new ContentHandlerProxy();
            emitter.setUnderlyingContentHandler((ContentHandler)handler);
            return emitter;
        } else {
            throw new SAXException("Failed to load emitter " + className +
                        ": it is not a SAX DocumentHandler or SAX2 ContentHandler");
        }      

     }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
