package com.icl.saxon.style;
import com.icl.saxon.*;
import com.icl.saxon.om.*;
import com.icl.saxon.handlers.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.pattern.Pattern;
import com.icl.saxon.trace.*;  // e.g.

import org.xml.sax.SAXException;
import org.xml.sax.AttributeList;

/**
* An xsl:template element in the style sheet.
*/
    
public class XSLTemplate extends StyleElement {

    protected int modeNameCode = -1;
    protected int templateFingerprint = -1;
    protected Pattern match;
    protected boolean prioritySpecified;
    protected double priority;
    protected Procedure procedure = new Procedure();

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    /**
    * Return the fingerprint for the name of this template
    */

    public int getTemplateFingerprint() throws SAXException {
    	
    	//We use -1 to mean "not yet evaluated"
    	
    	if (templateFingerprint==-1) {
    		// allow for forwards references
    		StandardNames sn = getStandardNames();
    		String nameAtt = getAttributeList().getValue(sn.NAME);
    		if (nameAtt!=null) {
    			templateFingerprint = 
            		makeNameCode(nameAtt, false) & 0xfffff;
            }
        }
        return templateFingerprint;
    }

    public int getMinImportPrecedence() throws SAXException {
        return ((XSLStyleSheet)getDocumentElement()).getMinImportPrecedence();
    }

    public void prepareAttributes() throws SAXException {

        //checkAllowedAttributes();

		String modeAtt = null;
		String nameAtt = null;
		String priorityAtt = null;
		String matchAtt = null;				

		StandardNames sn = getStandardNames();
		ExtendedAttributes atts = getAttributeList();

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.MODE) {
        		modeAtt = atts.getValue(a);
			} else if (f==sn.NAME) {
        		nameAtt = atts.getValue(a);
			} else if (f==sn.MATCH) {
        		matchAtt = atts.getValue(a);
			} else if (f==sn.PRIORITY) {
        		priorityAtt = atts.getValue(a);        		        		
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }        
        
        if (modeAtt!=null) {
            modeNameCode = makeNameCode(modeAtt, false);
        }
        
        if (nameAtt!=null) {
            templateFingerprint = makeNameCode(nameAtt, false) & 0xfffff;
        }
        
        prioritySpecified = (priorityAtt != null);
        if (prioritySpecified) { 
            try {
                priority = new Double(priorityAtt.trim()).doubleValue();
            } catch (NumberFormatException err) {
                throw styleError("Invalid numeric value for priority (" + priority + ")");
            }
        }

        if (matchAtt != null) {
            match = makePattern(matchAtt);
        }
        
        if (match==null && nameAtt==null)
            throw styleError("xsl:template must have a a name or match attribute (or both)");

	}
	

    public void validate() throws SAXException {
        checkTopLevel();

        // it is in error if there is another template with the same name and precedence

        if (templateFingerprint!=-1) {
            NodeInfo node = (NodeInfo)getPreviousSibling();
            while (node!=null) {
                if (node instanceof XSLTemplate) {
                    XSLTemplate t = (XSLTemplate)node;
                    if (t.getTemplateFingerprint()==templateFingerprint &&
                            t.getPrecedence()==this.getPrecedence()) {
                        throw styleError("There is another template with the same name and precedence");
                    }
                }
                node = (NodeInfo)node.getPreviousSibling();
            }
        }
                    
    }

    /**
    * Preprocess: this registers the template with the rule manager, and ensures
    * space is available for local variables
    */

    public void preprocess() throws SAXException
    {
        NodeHandler eh = new TemplateExpander(this);
        RuleManager mgr = getPrincipalStyleSheet().getRuleManager();
        Mode mode = mgr.getMode(modeNameCode);
        
        if (match!=null) {
            if (prioritySpecified) { 
                mgr.setHandler(match, eh, mode, getPrecedence(), priority);
            } else {
                mgr.setHandler(match, eh, mode, getPrecedence());
            }
        }

        getPrincipalStyleSheet().allocateLocalSlots(procedure.getNumberOfVariables());
      
    }

    /**
    * Process template. This is called while all the top-level nodes are being processed in order,
    * so it does nothing.
    */

    public void process(Context context) throws SAXException {
    }

    /**
    * Expand the template. Called from TemplateExpander (for apply-templates) and directly from
    * XSLCallTemplate
    */

    public void expand(Context context) throws SAXException {

    	if (context.getController().isTracing()) { // e.g. FIXME: trace tail recursion
    	    TraceListener listener = context.getController().getTraceListener();

    	    listener.enter(this, context);
    	    realExpand(context);
    	    listener.leave(this, context);

    	} else {
    	    realExpand(context);
    	}
    }

    private void realExpand(Context context) throws SAXException {
	    ParameterSet p = null;
	    do {
    		context.setTailRecursion(null);
    		processChildren(context);
    		p = context.getTailRecursion();
    		if (p!=null) {
    		    context.getBindery().closeStackFrame();
    		    context.getBindery().openStackFrame(p);
    		}
	    } while (p!=null);
    }

    /**
    * Disallow variable references in the match pattern
    */

    public Binding bindVariable(String name) throws SAXException {
        throw styleError("The match pattern in xsl:template may not contain references to variables");
    }

    /**
    * Get associated Procedure (for details of stack frame)
    */

    public Procedure getProcedure() {
        return procedure;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): 
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
