package com.icl.saxon.tinytree;
import com.icl.saxon.om.*;
//import org.xml.sax.Attributes;


    /**
    * AttributeCollection is an implementation of the SAX interface AttributeList
    * that also provides the ability to manipulate namespaces and to convert attributes
    * into Nodes.
    *
    * It is extremely similar (both in interface and in implementation) to the SAX2 Attributes
    * class, but was defined before SAX2 was available.
    *
    * The type of attribute is not maintained: all attributes are regarded as CDATA.
    */

public class TinyAttributeList implements ExtendedAttributes
{

    TinyDocumentImpl doc;
    int element;
    int firstAtt;
    int length;
    boolean empty;
    
    /**
    * Create an attribute list.
    */
    
    public TinyAttributeList (TinyDocumentImpl doc, int element) {
        this.doc = doc;
        this.element = element;
        empty = doc.offset[element]<0;
        if (empty) {
            length = 0;
        } else {
            length = 0;
            firstAtt = doc.offset[element];
            while (firstAtt+length < doc.numberOfAttributes &&
                    doc.attParent[firstAtt+length]==element) {
                length++;
            }
        }
    }


    //////////////////////////////////////////////////////////////////////
    // Implementation of org.xml.sax.Attributes
    //////////////////////////////////////////////////////////////////////


    /**
    * Return the number of attributes in the list.
    * @return The number of attributes in the list.
    */

    public int getLength ()
    {
        return length;
    }


    /**
    * Get the name code of an attribute (by position).
    *
    * @param i The position of the attribute in the list.
    * @return The local name of the attribute as a string, or null if there
    *         is no attribute at that position.
    */

    public int getNameCode (int i)
    {
        if (i>=length) return -1;
        if (i<0) return -1;
        
		return doc.attCode[firstAtt+i];
    }

    /**
    * Get the display name of an attribute (by position).
    *
    * @param i The position of the attribute in the list.
    * @return The display name of the attribute as a string, or null if there
    *         is no attribute at that position.
    */

    public String getQName (int i)
    {
        if (i>=length) return null;
        if (i<0) return null;
        
		return doc.getNamePool().getDisplayName(doc.attCode[firstAtt+i]);

    }

    /**
    * Get the local name of an attribute (by position).
    *
    * @param i The position of the attribute in the list.
    * @return The local name of the attribute as a string, or null if there
    *         is no attribute at that position.
    */

    public String getLocalName (int i)
    {
        if (i>=length) return null;
        if (i<0) return null;
        
		return doc.getNamePool().getLocalName(doc.attCode[firstAtt+i]);
    }

    /**
    * Get the URI of an attribute (by position).
    *
    * @param i The position of the attribute in the list.
    * @return The local name of the attribute as a string, or null if there
    *         is no attribute at that position.
    */

    public String getURI (int i)
    {
        if (i>=length) return null;
        if (i<0) return null;
        
		return doc.getNamePool().getURI(doc.attCode[firstAtt+i]);
    }

	/**
	* Get the fingerprint, given the index
	*/
	
	public int getFingerprint(int i) {
        if (i>=length) return -1;
        if (i<0) return -1;
        
		return doc.attCode[firstAtt+i] & 0xfffff;
    }		


    /**
    * Get the type of an attribute (by position).
    * @param i The position of the attribute in the list.
    * @return The attribute type as a string (always "CDATA")
    */
    
    public String getType (int i)
    {
		return "CDATA";
    }

    /**
    * Get the type of an attribute (by name).
    *
    * @param uri The namespace uri of the attribute.
    * @param localname The local name of the attribute.
    * @return The type of the attribute: always "CDATA"
    */
    
    public String getType (String uri, String localname)
    {
		return "CDATA";
    }

    /**
    * Get the value of an attribute (by position).
    *
    * @param i The position of the attribute in the list.
    * @return The attribute value as a string, or null if
    *         there is no attribute at that position.
    */
    
    public String getValue (int i)
    {
        if (i>=length) return null;
        if (i<0) return null;
        
        return doc.attValue[firstAtt+i];
    }

    /**
    * Get the value of an attribute (by name).
    *
    * @param uri The namespace uri of the attribute.
    * @param localname The local name of the attribute.
    * @return The index position of the attribute
    */
    
    public String getValue (String uri, String localname)
    {
        int offset = findByName(uri, localname);
        return ( offset<0 ? null : doc.attValue[offset]);
    }

	/**
	* Get the value of an attribute by Fingerprint
	*/
	
	public String getValueByFingerprint (int fingerprint) {
		int offset = findByFingerprint(fingerprint);
        return ( offset<0 ? null : doc.attValue[offset]);
    }		
		

    /**
    * Get the index of an attribute (by name).
    *
    * @param name The display name of the attribute.
    * @return The index position of the attribute
    */
    
    public int getIndex (String name)
    {
        int offset = findByDisplayName(name);
        return ( offset<0 ? -1 : offset - firstAtt);
    }

    /**
    * Get the index of an attribute (by name).
    *
    * @param uri The namespace uri of the attribute.
    * @param localname The local name of the attribute.
    * @return The index position of the attribute
    */
    
    public int getIndex (String uri, String localname)
    {
        int offset = findByName(uri, localname);
        return ( offset<0 ? -1 : offset - firstAtt);
    }

    /**
    * Get the index of an attribute (by fingerprint).
    *
    * @param uri The namespace uri of the attribute.
    * @param localname The local name of the attribute.
    * @return The index position of the attribute
    */
    
    public int getIndexByFingerprint (int fingerprint)
    {
        int offset = findByFingerprint(fingerprint);
        return ( offset<0 ? -1 : offset - firstAtt);
    }

    /**
    * Get the type of an attribute (by name).
    *
    * @param name The display name of the attribute.
    * @return The attribute type: always "CDATA"
    */
    
    public String getType (String name)
    {
		return "CDATA";
    }


    /**
    * Get the value of an attribute (by name).
    *
    * @param name The attribute name.
    */
    
    public String getValue (String name)
    {
        int offset = findByDisplayName(name);
        return ( offset<0 ? null : doc.attValue[offset]);
    }

    //////////////////////////////////////////////////////////////////////
    // Supporting methods
    //////////////////////////////////////////////////////////////////////



    private int findByName(String uri, String localName) {
        if (length==0) return -1;
        int fingerprint = doc.getNamePool().getFingerprint(uri, localName);
		return findByFingerprint(fingerprint);
    }

    private int findByFingerprint(int fingerprint) {
        if (length==0) return -1;
        for (int i=0; i<length; i++) {
            if (fingerprint == (doc.attCode[firstAtt+i] & 0xfffff)) {
                return i+firstAtt;
            }
        }
        return -1;
    }

    /**
    * Find an attribute by display name
    * @return the offset of the attribute 
    */

    private int findByDisplayName(String name) {
        if (length==0) return -1;

        for (int i=0; i<length; i++) {
        	String dn = getQName(i);
        	if (dn.equals(name)) {
                return i+firstAtt;
            }
        }
        return -1;

    }
    

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
