package com.icl.saxon.tinytree;
import com.icl.saxon.om.*;
import com.icl.saxon.tree.DOMExceptionImpl;
import com.icl.saxon.expr.NodeSetExtent;
import com.icl.saxon.output.Outputter;
import com.icl.saxon.pattern.NameTest;
import com.icl.saxon.pattern.NamespaceTest;

import java.util.Vector;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.w3c.dom.DOMException;

/**
  * A node in the XML parse tree representing an XML element.<P>
  * This class is an implementation of ElementInfo
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  * @version 8 August 2000: separated from ElementImpl
  */

public class TinyElementImpl extends TinyParentNodeImpl
    implements ElementInfo, Element, NamedNodeMap {

    //protected String[] namespaceList = null;          // organised as prefix/uri pairs
            // note that this namespace list includes only the namespaces actually defined on
            // this element, not those inherited from outer elements.

    /**
    * Constructor
    */
    
    public TinyElementImpl(TinyDocumentImpl doc, int nodeNr) {
        this.document = doc;
        this.nodeNr = nodeNr;
    }

    /**
    * Make a NameCode, using this Element as the context for namespace resolution
    * @param qname The name as written, in the form "[prefix:]localname"
    * @boolean useDefault Defines the action when there is no prefix. If true, use
    * the default namespace URI (as for element names). If false, use no namespace URI
    * (as for attribute names).
    */

    public final int makeNameCode(String qname, boolean useDefault) throws SAXException {
		NamePool namePool = document.getNamePool();
		String prefix = Name.getPrefix(qname);
        if (prefix.equals("")) {
            short uriCode = 0;
            
            if (useDefault) {
                uriCode = getURICodeForPrefix(prefix);
            }
            
			return namePool.allocate(prefix, uriCode, qname);

        } else {
            String localName = Name.getLocalName(qname);
            short uriCode = getURICodeForPrefix(prefix);
			return namePool.allocate(prefix, uriCode, localName);
        }

	}   	

	/**
	* Make a NameTest, using this element as the context for namespace resolution
	*/
	
	public NameTest makeNameTest(short nodeType, String qname, boolean useDefault)
			throws SAXException {
        int nameCode = makeNameCode(qname, useDefault);			
		return new NameTest(nodeType, nameCode);
	}

	/**
	* Make a NamespaceTest, using this element as the context for namespace resolution
	*/
	
	public NamespaceTest makeNamespaceTest(short nodeType, String prefix)
			throws SAXException {
		NamePool pool = document.getNamePool();
        short uriCode = getURICodeForPrefix(prefix);			
		return new NamespaceTest(nodeType, uriCode);
	}

    /**
    * Return the type of node.
    * @return NodeInfo.ELEMENT
    */

    public final short getNodeType() {
        return ELEMENT;
    }

    /**
    * Search the NamespaceList for a given prefix, returning the corresponding URI.
    * @param prefix The prefix to be matched. To find the default namespace, supply ""
    * @return The URI code corresponding to this namespace. If it is an unnamed default namespace,
    * return "". 
    * @throws SAXException if the prefix has not been declared on this NamespaceList.
    */

    // Note: this method is not actually used; it is needed only on the
    // stylesheet tree. So we aren't too worried about efficiency!

    public short getURICodeForPrefix(String prefix) throws SAXException {
        Vector v = new Vector();
        addNamespaceNodes(this, v, true);
        for (int i=0; i<v.size(); i++) {
            NamespaceInfo ns = (NamespaceInfo)v.elementAt(i);
            if (ns.getNamespacePrefix().equals(prefix)) {
                return (short)(ns.getNamespaceCode() & 0xffff);
            }
        }
        throw new SAXException("Namespace prefix " + prefix + " has not been declared");
    }

    /**
    * Search the NamespaceList for a given URI, returning the corresponding prefix.
    * @param uri The URI to be matched. 
    * @return The prefix corresponding to this URI. If not found, return null. If there is
    * more than one prefix matching the URI, the first one found is returned. If the URI matches
    * the default namespace, return an empty string.
    */

    // This method is not actually used: it is needed only for stylesheet trees.
    // So performance is unimportant.

    public String getPrefixForURI(String uri) {
        Vector v = new Vector();
        addNamespaceNodes(this, v, true);
        for (int i=0; i<v.size(); i++) {
            NamespaceInfo ns = (NamespaceInfo)v.elementAt(i);
            if (ns.getNamespaceURI().equals(uri)) {
                return ns.getNamespacePrefix();
            }
        }
        return null;
    }

    /**
    * Make the set of all namespace nodes associated with this element.
    * @param owner The element owning these namespace nodes.
    * @param list a Vector containing NamespaceInfo objects representing the namespaces
    * in scope for this element; the method appends nodes to this Vector, which should
    * initially be empty. The list WILL include
    * an entry for the undeclaration xmlns=""; it is the job of NamespaceEnumeration
    * to ignore this, since it doesn't represent a true namespace node.
    * @param addXML Add the node for the XML Namespace
    */

    public void addNamespaceNodes(ElementInfo owner, Vector list, boolean addXML) {
        int ns = document.length[nodeNr]; // by convention
        int max = list.size();
        
        if (ns>=0) {
            while (document.namespaceParent[ns] == nodeNr && ns < document.numberOfNamespaces) {
				TinyNamespaceImpl nsi = new TinyNamespaceImpl(document, ns);
                String prefix = nsi.getNamespacePrefix();

                boolean found = false;

                // Don't add a node if the prefix is already in the list
                for (int j=0; j<max; ) {
                    NamespaceInfo nsj = (NamespaceInfo)list.elementAt(j++);
                    if (nsj.getNamespacePrefix().equals(prefix)) {
                        found=true;
                        break;
                    }
                }
                if (!found) {
                    nsi.setParentNode(((TinyElementImpl)owner).nodeNr);
                    list.addElement(nsi);
                }

                ns++;            
            }
        }
        
        // now add the namespaces defined on the ancestor nodes
        
        NodeInfo parent = (NodeInfo)getParentNode();
        if (!(parent instanceof DocumentInfo)) {
            ((TinyElementImpl)parent).addNamespaceNodes(owner, list, false);
        }

        if (addXML) {
                // XML namespace is node 0 by convention
                TinyNamespaceImpl nsi = new TinyNamespaceImpl(document, 0);
                nsi.setParentNode(nodeNr);
                list.addElement(nsi);            
        }

    }
    
    /**
    * Output all namespace nodes associated with this element.
    * @param out The relevant outputter
    * @param includeAncestors True if namespaces associated with ancestor
    * elements must also be output; false if these are already known to be
    * on the result tree.
    */

    public void outputNamespaceNodes(Outputter out, boolean includeAncestors)
                throws SAXException {

        int ns = document.length[nodeNr]; // by convention
        if (ns>0 ) {        
            while (document.namespaceParent[ns] == nodeNr &&
                         ns < document.numberOfNamespaces) {
                int nscode = document.namespaceCode[ns];
                out.writeNamespaceDeclaration(nscode);
                ns++;
            }
        }

        // now add the namespaces defined on the ancestor nodes. We rely on the outputter
        // to eliminate multiple declarations of the same prefix
        
        if (includeAncestors && document.isUsingNamespaces()) {
            NodeInfo parent = (NodeInfo)getParentNode();
            if (!(parent instanceof DocumentInfo)) {
                ((ElementInfo)parent).outputNamespaceNodes(out, true);
            }
        }
    }
    
    
    /**
    * Get the attribute list for this element.
    * @return The attribute list. This will not include any
    * namespace attributes. 
    */
    
    public ExtendedAttributes getAttributeList() {
        return new TinyAttributeList(document, nodeNr);
    }

    /**
     * Returns whether this node (if it is an element) has any attributes.
     * @return <code>true</code> if this node has any attributes, 
     *   <code>false</code> otherwise.
     * @since DOM Level 2
     */
    
    public boolean hasAttributes() {
        return document.offset[nodeNr] >= 0;
    }    
    
    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param uri the namespace uri of an attribute
     * @param localName the local name of an attribute
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String uri, String localName ) {
        int f = document.getNamePool().getFingerprint(uri, localName);
		return getAttributeValue(f);
    }

    /**
     *  Find the value of a given attribute of this element. <BR>
     *  This is a short-cut method; the full capability to examine
     *  attributes is offered via the getAttributeList() method. <BR>
     *  The attribute may either be one that was present in the original XML document,
     *  or one that has been set by the application using setAttribute(). <BR>
     *  @param name the name of an attribute. There must be no prefix in the name.
     *  @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String name ) {
		return getAttributeValue( "", name ); 
    }

    /**
    * Get the value of a given attribute of this node
    * @param fingerprint The fingerprint of the attribute name
    * @return the attribute value if it exists or null if not
    */
    
    public String getAttributeValue(int fingerprint) {
        int a = document.offset[nodeNr];
        if (a<0) return null;
        while (document.attParent[a] == nodeNr && a < document.numberOfAttributes) {
            if ((document.attCode[a] & 0xfffff) == fingerprint ) {
                return document.attValue[a];
            }
            a++;
        }        
        return null;
    }    

    /**
    * Make an attribute node for a given attribute of this element
    * @param index The relative position of the attribute, counting from zero. This
    * is trusted to be in range.
    */

    public AttributeInfo makeAttributeNode(int index) {
        int a = document.offset[nodeNr];
        if (a<0) return null;
        return document.getAttributeNode(a+index);
    }

    /**
    * Make an attribute node for a given attribute of this element
    * @param fingerprint The fingerprint of the attribute name
    */
    
    public AttributeInfo makeAttributeNodeFS(int fingerprint) {
        int a = document.offset[nodeNr];
        if (a<0) return null;
        while (a < document.numberOfAttributes && document.attParent[a] == nodeNr) {
            if ((document.attCode[a] & 0xfffff) == fingerprint ) {
                return document.getAttributeNode(a);
            }
            a++;
        }        
        return null;
    }

    
    /**
    * Set the value of an attribute on the current element. This affects subsequent calls
    * of getAttribute() for that element. 
    * @param name The name of the attribute to be set. Any prefix is interpreted relative
    * to the namespaces defined for this element. 
    * @param value The new value of the attribute. Set this to null to remove the attribute.
    */
    
    public void setAttribute(String name, String value ) throws DOMException {
        throw new DOMExceptionImpl((short)9999, "Saxon DOM is not updateable");  
    }

    /**
    * Copy this node to a given outputter (supporting xsl:copy-of)
    */

    public void copy(Outputter out) throws SAXException {
        copy(out, true);
    }

    /**
    * Copy this node to a given outputter
    * @param allNamespaces true if all namespace nodes must be copied; false
    * if namespace nodes for the parent element are already on the result tree
    */

    public void copy(Outputter out, boolean allNamespaces) throws SAXException {
        int nc = getNameCode();
        out.writeStartTag(nc);

        // output the namespaces
        
        outputNamespaceNodes(out, allNamespaces);

        // output the attributes

        int a = document.offset[nodeNr];
        if (a >= 0) {
            while (document.attParent[a] == nodeNr && a < document.numberOfAttributes) {
            	document.getAttributeNode(a).copy(out);
                a++;
            }
        }

        // output the children

        NodeInfo next = (NodeInfo)getFirstChild();
        while (next!=null) {
            if (next instanceof TinyElementImpl) {
                ((TinyElementImpl)next).copy(out, false);
                          // no need to do all the namespaces again
            } else {
                next.copy(out);
            }
            next = (NodeInfo)next.getNextSibling();
        }

        out.writeEndTag(nc);
    }    

    /**
    * Generate a path to this node
    */

    public String getPath() {
        String pre = ((NodeInfo)getParentNode()).getPath();
        return (pre.equals("/") ? "" : pre) +
               "/" + getDisplayName() +
               "[" + Navigator.getNumberSimple(this) + "]";
    }


    ////////////////////////////////////////////////////////////////////////////
    // Following interfaces are provided to implement the DOM Element interface
    ////////////////////////////////////////////////////////////////////////////
    

    /**
     *  The name of the element (DOM interface). 
     */
     
    public String getTagName() {
        return getDisplayName();
    }

    /**
     *  The value of this node (as defined in DOM). Always null.
     */

    public String getNodeValue() {
        return null;
    }

    /**
     * Retrieves an attribute value by name. Namespace declarations will not
     * be retrieved.
     * @param name  The QName of the attribute to retrieve.
     * @return  The <code>Attr</code> value as a string, or the empty string if
     *    that attribute does not have a specified or default value.
     */
     
    public String getAttribute(String name) {
    	String val;
        try {
        	if (name.indexOf(':') < 0) {
        		val = getAttributeValue(name);
        	} else {
            	int nameCode = makeNameCode(name, false);
            	val = getAttributeValue(nameCode & 0xfffff);
            }
            return (val==null ? "" : val);
        } catch (SAXException err) {
            return "";
        }
    }

    /**
     * A <code>NamedNodeMap</code> containing the attributes of this element.
     * Namespace declarations will not be retrieved.
     */
     
    public NamedNodeMap getAttributes() {
        return this;
    }

    /**
     * Removes an attribute by name.
     * @param name  The name of the attribute to remove.
     */
     
    public void removeAttribute(String name) {
        setAttribute(name, null);
    }
    
    /**
     * Retrieves an attribute node by name.
     * Namespace declarations will not be retrieved.
     * <br> To retrieve an attribute node by qualified name and namespace URI, 
     * use the <code>getAttributeNodeNS</code> method.
     * @param name  The name (<code>nodeName</code> ) of the attribute to 
     *   retrieve.
     * @return  The <code>Attr</code> node with the specified name (
     *   <code>nodeName</code> ) or <code>null</code> if there is no such 
     *   attribute.
     */
     
    public Attr getAttributeNode(String name) {
        int index = getAttributeList().getIndex(name);
        if (index<0) return null;
        return (TinyAttributeImpl)makeAttributeNode(index);
    }

    /**
     * Adds a new attribute node. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
     
    public Attr setAttributeNode(Attr newAttr) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     * Removes the specified attribute node. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */

    public Attr removeAttributeNode(Attr oldAttr) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     *  Returns a <code>NodeList</code> of all descendant <code>Elements</code>
     *  with a given tag name, in the order in which they are encountered in 
     * a preorder traversal of this <code>Element</code> tree.
     * @param name  The name of the tag to match on. The special value "*" 
     *   matches all tags.
     * @return  A list of matching <code>Element</code> nodes.
     */
     
    public NodeList getElementsByTagName(String name) {
        Vector v = new Vector();
        int depth = document.depth[nodeNr];
        int next = nodeNr+1;

        while(document.depth[next]>depth) {
        	NodeInfo node = document.getNode(next);
        	String dn = node.getDisplayName();
            if (document.nodeType[next] == NodeInfo.ELEMENT) {
                if (name.equals("*") || name.equals(dn)) {
                    v.addElement(node);
                }
            }
            next++;
        }
        return new NodeSetExtent(v);        
    }

    /**
     *  Returns a <code>NodeList</code> of all descendant <code>Elements</code>
     *  with a given tag name, in the order in which they are encountered in 
     * a preorder traversal of this <code>Element</code> tree.
     * @param namespaceURI  The  namespace URI of the elements to match on. 
     *   The special value "*" matches all namespaces.
     * @param localName  The  local name of the elements to match on. The 
     *   special value "*" matches all local names.
     * @return  A list of matching <code>Element</code> nodes.
     */
     
    public NodeList getElementsByTagNameNS(String namespaceURI, String localName) {
        Vector v = new Vector();
        int depth = document.depth[nodeNr];
        int next = nodeNr+1;
        int fingerprint = document.getNamePool().getFingerprint(namespaceURI, localName);

        while(document.depth[next]>depth) {
            if (document.nodeType[next] == NodeInfo.ELEMENT) {
            	NodeInfo node = document.getNode(next);
            	String local = node.getLocalName();
            	String uri = node.getURI();
                if (namespaceURI.equals("*") ||
                            namespaceURI.equals(uri) &&
                    (localName.equals("*") ||
                            localName.equals(local))) {
                    v.addElement(node);
                }
            }
            next++;
        }
        return new NodeSetExtent(v);        
    }


    /**
     * Retrieves an attribute value by local name and namespace URI. 
     * HTML-only DOM implementations do not need to implement this method.
     * @param namespaceURI  The  namespace URI of the attribute to retrieve.
     * @param localName  The  local name of the attribute to retrieve.
     * @return  The <code>Attr</code> value as a string, or the empty string if
     *    that attribute does not have a specified or default value.
     * @since DOM Level 2
     */
     
    public String getAttributeNS(String namespaceURI, String localName) {
    	String val = getAttributeValue(namespaceURI, localName);
    	if (val==null) return "";
    	return val; 
    }                                           

    /**
     * Adds a new attribute. Always fails.
     * @param namespaceURI  The  namespace URI of the attribute to create or 
     *   alter.
     * @param qualifiedName  The  qualified name of the attribute to create or 
     *   alter.
     * @param value  The value to set in string form.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
     
    public void setAttributeNS(String namespaceURI, 
                               String qualifiedName, 
                               String value)
                               throws DOMException {
        disallowUpdate();
    }

    /**
     * Removes an attribute by local name and namespace URI. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     * @since DOM Level 2
     */
     
    public void removeAttributeNS(String namespaceURI, 
                                  String localName)
                                  throws DOMException{
        disallowUpdate();
    }
    
    /**
     * Retrieves an <code>Attr</code> node by local name and namespace URI. 
     * DOM method, so namespace declarations count as attributes.
     * @param namespaceURI  The  namespace URI of the attribute to retrieve.
     * @param localName  The  local name of the attribute to retrieve.
     * @return  The <code>Attr</code> node with the specified attribute local 
     *   name and namespace URI or <code>null</code> if there is no such 
     *   attribute.
     * @since DOM Level 2
     */
     
    public Attr getAttributeNodeNS(String namespaceURI, String localName) {
		return (Attr)makeAttributeNode(
			getAttributeList().getIndex(namespaceURI, localName));
    }                                    

    /**
     * Add a new attribute. Always fails.
     * @param newAttr  The <code>Attr</code> node to add to the attribute list.
     * @return  If the <code>newAttr</code> attribute replaces an existing 
     *   attribute with the same  local name and  namespace URI , the 
     *   replaced <code>Attr</code> node is returned, otherwise 
     *   <code>null</code> is returned.
     * @exception DOMException
     *   <br> NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     * @since DOM Level 2
     */
     
    public Attr setAttributeNodeNS(Attr newAttr)
                                   throws DOMException{
        disallowUpdate();
        return null;
    }

    /**
     * Returns <code>true</code> when an attribute with a given name is 
     * specified on this element or has a default value, <code>false</code> 
     * otherwise. 
     * Namespace declarations will not be retrieved.
     * @param name  The name of the attribute to look for.
     * @return <code>true</code> if an attribute with the given name is 
     *   specified on this element or has a default value, <code>false</code> 
     *   otherwise.
     * @since DOM Level 2
     */
     
    public boolean hasAttribute(String name) {
        try {
        	if (name.indexOf(':')<0) {
        		return getAttributeValue(name) != null;
        	} else {
            	int nameCode = makeNameCode(name, false);
            	return getAttributeValue(nameCode & 0xfffff) != null;
            }
        } catch (SAXException err) {
            return false;
        }
    }

    /**
     * Returns <code>true</code> when an attribute with a given local name 
     * and namespace URI is specified on this element or has a default value, 
     * <code>false</code> otherwise.
     * Namespace declarations will not be retrieved.     
     * @param namespaceURI  The  namespace URI of the attribute to look for.
     * @param localName  The  local name of the attribute to look for.
     * @return <code>true</code> if an attribute with the given local name and 
     *   namespace URI is specified or has a default value on this element, 
     *   <code>false</code> otherwise.
     * @since DOM Level 2
     */
     
    public boolean hasAttributeNS(String namespaceURI, String localName) {
		return (getAttributeValue(namespaceURI, localName) != null);
    }    

    //////////////////////////////////////////////////////////////////////
    // Methods to implement DOM NamedNodeMap (the set of attributes)
    //////////////////////////////////////////////////////////////////////

    /**
    * Get named attribute (DOM NamedNodeMap method)
    */

    public Node getNamedItem(String name) {
        return getAttributeNode(name);
    }

    /**
    * Set named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node setNamedItem(Node arg) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Remove named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node removeNamedItem(String name) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Get n'th attribute (DOM NamedNodeMap method).
    * Namespace declarations are not retrieved.
    */

    public Node item(int index) {
        if (index<0) {
            return null;
        }
        if (index>=getLength()) {
            return null;
        }
        return makeAttributeNode(index);
    }

    /**
    * Get number of attributes (DOM NamedNodeMap method). 
    */

    public int getLength() {
        Attributes atts = getAttributeList();
        return atts.getLength();
    }

    /**
    * Get named attribute (DOM NamedNodeMap method)
    */

    public Node getNamedItemNS(String uri, String localName) {
        return getAttributeNodeNS(uri, localName);
    }

    /**
    * Set named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node setNamedItemNS(Node arg) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Remove named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node removeNamedItemNS(String uri, String localName) throws DOMException {
        disallowUpdate();
        return null;
    }

        
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
