package com.icl.saxon.tinytree;
import com.icl.saxon.om.*;
import com.icl.saxon.Context;
import com.icl.saxon.expr.NodeSetExtent;
import com.icl.saxon.expr.NodeEnumeration;
import com.icl.saxon.expr.SingletonEnumeration;
import com.icl.saxon.pattern.NodeTest;
import com.icl.saxon.output.Outputter;
import com.icl.saxon.tree.DOMExceptionImpl;

import java.util.Vector;
import java.util.Hashtable;
import org.xml.sax.AttributeList;
import org.xml.sax.SAXException;
import org.w3c.dom.*;



/**
  * A node in the XML parse tree representing an XML element, character content, or attribute.<P>
  * This is the top-level class in the implementation class hierarchy; it essentially contains
  * all those methods that can be defined using other primitive methods, without direct access
  * to data.
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  */

abstract class TinyNodeImpl implements NodeInfo {

    private static SingletonEnumeration emptyEnumeration = new SingletonEnumeration(null);
    
    protected TinyDocumentImpl document;
    protected int nodeNr;
    protected TinyNodeImpl parent = null;

    /**
    * Set the parent of this node
    */

    protected void setParentNode(TinyNodeImpl parent) {
        this.parent = parent;
    }

    /**
    * Determine whether this is the same node as another node
    * @return true if this Node object and the supplied Node object represent the
    * same node in the tree.
    */

    public boolean isSameNode(NodeInfo other) {
        if (this==other) return true;
        if (!(other instanceof TinyNodeImpl)) return false;
        return this.document==((TinyNodeImpl)other).document &&
             this.nodeNr==((TinyNodeImpl)other).nodeNr;
    }

    /**
    * Get the base URI for the node.
    */

    public String getSystemId() {
        return document.getBaseURI(nodeNr);
    }

    /**
    * Set the line number of the node within its source document entity
    */

    public void setLineNumber(int line) {
        document.setLineNumber(nodeNr, line);
    }


    /**
    * Get the line number of the node within its source document entity
    */

    public int getLineNumber() {
        return document.getLineNumber(nodeNr);
    }

    /**
    * Get the node sequence number (in document order). Sequence numbers are monotonic but not
    * consecutive. In this implementation, the top half is the 
    */

    public long getSequenceNumber() {
        return (document.getSequenceNumber() + nodeNr) << 32;
    }

	/**
	* Get the fingerprint of the node, used for matching names
	*/
	
	public int getFingerprint() {
		return document.nameCode[nodeNr] & 0xfffff;
	}

	/**
	* Get the name code of the node, used for matching names
	*/
	
	public int getNameCode() {
		return document.nameCode[nodeNr];
	}

    /**
    * Get the name of this node, following the DOM rules 
    * @return The name of the node. For an element this is the element name, for an attribute
    * it is the attribute name, as a QName. Other node types return conventional names such
    * as "#text" or "#comment"
    */

    public String getNodeName() {
        // default implementation
        return getDisplayName();
    }

    /**
    * Get the prefix part of the name of this node. This is the name before the ":" if any.
    * @return the prefix part of the name. For an unnamed node, return "".
    */

    public String getPrefix() {
        int code = document.nameCode[nodeNr];
        if (code<0) return "";
        if ((code>>20 & 0xff) == 0) return "";
        return document.getNamePool().getPrefix(code);
    }

    /**
    * Get the URI part of the name of this node. This is the URI corresponding to the
    * prefix, or the URI of the default namespace if appropriate.
    * @return The URI of the namespace of this node. For an unnamed node, or for
    * an element or attribute in the default namespace, return an empty string.
    */

    public String getURI() {
        int code = document.nameCode[nodeNr];
        if (code<0) return "";
        return document.getNamePool().getURI(code);
    }

	/**
	* Get URI code. This is a number that uniquely identifies the namespace URI of the
	* node, within its NamePool
	*/
	
	public short getURICode() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return -1;
        return document.getNamePool().getURICode(nameCode);
    }		    
    
    /**
    * Get the display name of this node. For elements and attributes this is [prefix:]localname.
    * For unnamed nodes, it is an empty string.
    * @return The display name of this node. 
    * For a node with no name, return an empty string.
    */

    public String getDisplayName() {
        int code = document.nameCode[nodeNr];
        if (code<0) return "";
        return document.getNamePool().getDisplayName(code);
    }

    /**
    * Get the local name of this node. 
    * @return The local name of this node. 
    * For a node with no name, return an empty string.
    */

    public String getLocalName() {
        int code = document.nameCode[nodeNr];
        if (code<0) return "";
        return document.getNamePool().getLocalName(code);
    }

    /**
    * Test if the name of the node (including namespaces) is equivalent to the given name
    */

    //public boolean hasName(Name name) {
	//	return (document.nameCode[nodeNr]&0x3ffffff)==name.getSignature();
    //}

    /**
    * Get the column number of the node. This is not currently maintained, so return -1
    */

    //public int getColumnNumber() {
    //    return -1;
    //}

    /**
    * Get the public identifier of the document entity containing this node. This
    * is not currently maintained: return null
    */

    public String getPublicId() {
        return null;
    }

    /**
    * Get the index of this node, i.e. its position among its siblings
    */

    public final int getIndex() {
        int index = 1;
        for (int i=nodeNr-1; i>0; i--) {
            if (document.depth[i]==document.depth[nodeNr]) index++;
            if (document.depth[i]<document.depth[nodeNr]) break;
        }
        return index;
    }

    /**
     * Find the parent node of this node.
     * @return The Node object describing the containing element or root node.
     */

    public Node getParentNode()  {
        if (parent!=null) return parent;

        // if parent is unknown
        for (int i=nodeNr-1; i>=0; i--) {
            if (document.depth[i]<document.depth[nodeNr]) {
                parent = document.getNode(i);
                return parent;
            }
        }
        parent = document;
        return parent;
    }

    /**
    * Get the previous sibling of the node
    * @return The previous sibling node. Returns null if the current node is the first
    * child of its parent.
    */
    
    public Node getPreviousSibling()  {
        for (int i=nodeNr-1; i>=0; i--) {
            if (document.depth[i]==document.depth[nodeNr]) {
                TinyNodeImpl prev = document.getNode(i);
                prev.setParentNode(parent);
                return prev;
            }
            if (document.depth[i]<document.depth[nodeNr]) {
                return null;
            }
        }
        return null;
    }


   /**
    * Get next sibling node
    * @return The next sibling node. Returns null if the current node is the last
    * child of its parent.
    */

    public Node getNextSibling()  {

        // fast path to avoid scanning whole document at top level
        //if (nodeNr == document.lastLevelOneNode) return null;

        //short thisDepth = document.depth[nodeNr];
        
        //for (int i=nodeNr+1; i<=document.numberOfNodes; i++) {
        //    int diff = document.depth[i] - thisDepth;
        //    if (diff == 0) {
        //        return document.getNode(i);
        //    }
        //    if (diff < 0) {
        //        return null;
        //    }
        //}
        //return null;

        /*NEXT*/ //Code for using next pointers is commented out because
        /*NEXT*/ //it appears to be counterproductive (ie things take longer)
        
        /*NEXT*/ int n = document.next[nodeNr];
        /*NEXT*/ if (n<0) return null;
        /*NEXT*/ TinyNodeImpl next = document.getNode(n);
        /*NEXT*/ next.setParentNode(parent);
        /*NEXT*/ return next;
    }    

    /**
    * Get first child 
    * @return the first child node of this node, or null if it has no children
    */

    public Node getFirstChild()  {
        // overridden in TParentNodeImpl
        return null;
    }

    /**
    * Get last child
    * @return last child of this node, or null if it has no children
    */

    public Node getLastChild()  {
        // overridden in TinyParentNodeImpl
        return null;
    }

    /**
    * Get the number of children.
    */

    public int getNumberOfChildren() {
        // overridden in TinyParentNodeImpl
        return 0;
    }

    /**
    * Get all child nodes of the node 
    * @return 
    */

    private static NodeInfo[] emptyArray = new NodeInfo[0];
    public NodeInfo[] getAllChildNodes()  {
        // overridden in TinyParentNodeImpl
        return emptyArray;
    }

    /**
    * Get an enumeration of the children of this node
    */

    public NodeEnumeration enumerateChildren() {
        // overridden in TParentNodeImpl
        return emptyEnumeration;
    }

    /**
     * Returns whether this node (if it is an element) has any attributes.
     * @return <code>true</code> if this node has any attributes, 
     *   <code>false</code> otherwise.
     * @since DOM Level 2
     */
    
    public boolean hasAttributes() {
        // overridden in TElementImpl
        return false;
    }    
    
    /**
    * Determine whether the node is of a given type. <P>
    * Note, this can also be done by testing the node using "instanceof". But this is
    * inconvenient when passing the class as a parameter to another routine.
    * @param nodeType One of the specific node types such as ELEMENT or TEXT, or the general
    * node type NODE
    * @return true if the node is an instance of the specified node type
    */

    //public final boolean isa(int nodeType) {
    //    return (nodeType==NODE || nodeType==getNodeType());
    //}
    
    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param uri the namespace uri of an attribute
     * @param localName the local name of an attribute
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String uri, String localName ) {
        return null;
    }


    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param name the name of an attribute. This must be an unqualified attribute name,
     * i.e. one with no namespace prefix.
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String name ) {
        // overridden in TElementImpl
        return null;
    }

    
    /**
    * Get the value of a given attribute of this node
    * @param fingerprint The fingerprint of the attribute name
    * @return the attribute value if it exists or null if not
    */
    
    public String getAttributeValue(int fingerprint) {
        // overridden in TElementImpl
    	return null;
    }    

     /**
     * Get the nearest ancestor element with a given name
     * @param name The name of the required ancestor. 
     * @return The NodeInfo for the nearest ancestor with the
     * given tag; null if there is no such ancestor
     */

    public final ElementInfo getAncestor(NodeTest test) {
        NodeInfo ancestor = (NodeInfo)this.getParentNode();
        while (ancestor!=null && (ancestor instanceof ElementInfo) &&
                              !test.matches(ancestor)) {
            ancestor = (NodeInfo)ancestor.getParentNode();
        }
        if (ancestor instanceof DocumentInfo) return null;
        return (ElementInfo)ancestor;
    }

    
    /**
     * Determine whether this node is the outermost element.
     * @return True if this element is the document (outermost) element. (Note this
     * only works for a well-formed document).
     */
     
    public boolean isDocumentElement() {
        return (this instanceof ElementInfo && document.depth[nodeNr]==1);
    }

    /**
     * Get the outermost element.
     * @return the ElementInfo for the outermost element of the document. If the document is
     * not well-formed, this returns the last element child of the root if there is one, otherwise
     * null.
     */
     
    public Element getDocumentElement() {
        return document.getDocumentElement();

    }

    /**
    * Get the root (document) node
    * @return the DocumentInfo representing the containing document
    */

    public DocumentInfo getDocumentRoot() {
        return document;
    }

    /**
    * Get the next node in document order
    * @param anchor: the scan stops when it reaches a node that is not a descendant of the specified
    * anchor node
    * @return the next node in the document, or null if there is no such node
    */

    public NodeInfo getNextInDocument(NodeInfo anchor) {
        // this implementation stops when it finds any node at the same or
        // shallower depth as the anchor node
        int next = nodeNr+1;
        if (next == document.numberOfNodes) {
            return null;
        }
        if (document.depth[next] <= document.depth[((TinyNodeImpl)anchor).nodeNr]) {
            return null;
        }
        return document.getNode(next);
        // TODO: set the parent node
    }

    /**
    * Get the previous node in document order
    * @return the previous node in the document, or null if there is no such node
    */

    public NodeInfo getPreviousInDocument() {
        if (nodeNr==0) return null;
        return document.getNode(nodeNr-1);
    }

    /**
    * Copy the string-value of this node to a given outputter
    */

    public void copyStringValue(Outputter out) throws SAXException {
        out.writeContent(getValue());   // default implementation
    }
    
    /**
    * Get a character string that uniquely identifies this node and that collates nodes
    * into document order
    * @return a string. 
    */

    public String getSequentialKey() {
        NodeInfo curr = this;
        StringBuffer key = new StringBuffer();
        while(!(curr instanceof DocumentInfo)) {
            key.insert(0, Navigator.alphaKey(curr.getIndex()));
            curr = (NodeInfo)curr.getParentNode();   
        }
        key.insert(0, curr.getSequentialKey());
        return key.toString();
    }

    /**
    * Create a string that identifies the node (for diagnostic purposes only)
    */

    public String toString() {
        String s;
        s = getDisplayName();
        if (s.equals("")) {
            NodeInfo n = (NodeInfo)getParentNode();
            if (n != null) s = n.getDisplayName() + "/";
            s += "\"" + getValue().trim() + "\"";
        }
        s = s + "(" + getIndex() + ")";
        return s;
    }

    // implement DOM Node methods

    /**
    * Get the node value as defined in the DOM. This is not the same as the XPath string-value.
    */

    public String getNodeValue() {
        // default implementation
        return getValue();
    }

    /**
    * Set the node value. DOM method: always fails
    */

    public void setNodeValue(String nodeValue) throws DOMException {
        disallowUpdate();
    }

    /**
     * Return a <code>NodeList</code> that contains all children of this node. If 
     * there are no children, this is a <code>NodeList</code> containing no 
     * nodes.
     */
     
    public NodeList getChildNodes() {
        return new NodeSetExtent(getAllChildNodes());
    }

    /**
     * Return a <code>NamedNodeMap</code> containing the attributes of this node (if 
     * it is an <code>Element</code> ) or <code>null</code> otherwise. (DOM method)
     */
     
    public NamedNodeMap getAttributes() {
        // default implementation
        return null;
    }

    /**
     * Return the <code>Document</code> object associated with this node. (DOM mehod)
     */
     
    public Document getOwnerDocument() {
        return document;
    }

    /**
     * Insert the node <code>newChild</code> before the existing child node 
     * <code>refChild</code>. DOM method: always fails.
     * @param newChild  The node to insert.
     * @param refChild  The reference node, i.e., the node before which the 
     *   new node must be inserted.
     * @return  The node being inserted.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node insertBefore(Node newChild, 
                             Node refChild)
                             throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     * Replace the child node <code>oldChild</code> with 
     * <code>newChild</code> in the list of children, and returns the 
     * <code>oldChild</code> node. Always fails.
     * @param newChild  The new node to put in the child list.
     * @param oldChild  The node being replaced in the list.
     * @return  The node replaced.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node replaceChild(Node newChild, 
                             Node oldChild)
                             throws DOMException{
        disallowUpdate();
        return null;
    }
    
    /**
     * Remove the child node indicated by <code>oldChild</code> from the 
     * list of children, and returns it. DOM method: always fails.
     * @param oldChild  The node being removed.
     * @return  The node removed.
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node removeChild(Node oldChild) throws DOMException {
        disallowUpdate();
        return null;
    }
    
    /**
     *  Adds the node <code>newChild</code> to the end of the list of children 
     * of this node. DOM method: always fails.
     * @param newChild  The node to add.
     * @return  The node added.
     * @exception DOMException
     *   <br> NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node appendChild(Node newChild) throws DOMException {
        disallowUpdate();
        return null;
    }
    
    /**
     * Determine whether the node has any children.
     * @return  <code>true</code> if the node has any children, 
     *   <code>false</code> if the node has no children.
     */
     
    public boolean hasChildNodes() {
        int next = nodeNr+1;
        return (next < document.numberOfNodes &&
                    document.depth[next] > document.depth[nodeNr]);         
    }

    /**
     * Returns a duplicate of this node, i.e., serves as a generic copy 
     * constructor for nodes. The duplicate node has no parent. Not
     * implemented: always returns null. (Because trees are read-only, there
     * would be no way of using the resulting node.)
     * @param deep  If <code>true</code> , recursively clone the subtree under 
     *   the specified node; if <code>false</code> , clone only the node 
     *   itself (and its attributes, if it is an <code>Element</code> ).  
     * @return  The duplicate node.
     */
     
    public Node cloneNode(boolean deep) {
        // Not implemented 
        return null;
    }

    /**
     * Puts all <code>Text</code> nodes in the full depth of the sub-tree 
     * underneath this <code>Node</code>, including attribute nodes, into a 
     * "normal" form where only structure (e.g., elements, comments, 
     * processing instructions, CDATA sections, and entity references) 
     * separates <code>Text</code> nodes, i.e., there are neither adjacent 
     * <code>Text</code> nodes nor empty <code>Text</code> nodes.
     * @since DOM Level 2
     */
     
    public void normalize() {
        // null operation; nodes are always normalized
    }

    /**
     *  Tests whether the DOM implementation implements a specific feature and 
     * that feature is supported by this node.
     * @param feature  The name of the feature to test. This is the same name 
     *   which can be passed to the method <code>hasFeature</code> on 
     *   <code>DOMImplementation</code> .
     * @param version  This is the version number of the feature to test. In 
     *   Level 2, version 1, this is the string "2.0". If the version is not 
     *   specified, supporting any version of the feature will cause the 
     *   method to return <code>true</code> .
     * @return  Returns <code>true</code> if the specified feature is supported
     *    on this node, <code>false</code> otherwise.
     * @since DOM Level 2
     */
     
    public boolean isSupported(String feature, 
                               String version) {
        return feature.equalsIgnoreCase("xml");
    }

    /**
     * The namespace URI of this node, or <code>null</code> if it is 
     * unspecified.
     * <br> This is not a computed value that is the result of a namespace 
     * lookup based on an examination of the namespace declarations in scope. 
     * It is merely the namespace URI given at creation time.
     * <br> For nodes of any type other than <code>ELEMENT_NODE</code> and 
     * <code>ATTRIBUTE_NODE</code> and nodes created with a DOM Level 1 
     * method, such as <code>createElement</code> from the 
     * <code>Document</code> interface, this is always <code>null</code> . 
     * Per the  Namespaces in XML Specification  an attribute does not 
     * inherit its namespace from the element it is attached to. If an 
     * attribute is not explicitly given a namespace, it simply has no 
     * namespace.
     * @since DOM Level 2
     */
     
    public String getNamespaceURI() {
        String uri = getURI();
        return (uri.equals("") ? null : uri);
    }

    /**
    * Set the namespace prefix of this node. Always fails.
    */
    
    public void setPrefix(String prefix)
                            throws DOMException {
        disallowUpdate();
    }

    /**
    * Internal method used to indicate that update operations are not allowed
    */

    protected void disallowUpdate() throws DOMException {
        throw new DOMExceptionImpl(DOMException.NO_MODIFICATION_ALLOWED_ERR,
                               "The Saxon DOM cannot be updated");
    }
    

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
