package com.icl.saxon.tinytree;
import com.icl.saxon.om.*;
import com.icl.saxon.expr.NodeEnumeration;
import com.icl.saxon.expr.SingletonEnumeration;
import com.icl.saxon.LastPositionFinder;
import com.icl.saxon.output.Outputter;
import com.icl.saxon.Context;
import com.icl.saxon.Mode;

import org.xml.sax.SAXException;
import org.w3c.dom.Node;

/**
  * TinyParentNodeImpl is an implementation of a non-leaf node (specifically, an Element node
  * or a Document node)
  * @author Michael H. Kay (Michael.Kay@icl.com)
  */
  

abstract class TinyParentNodeImpl extends TinyNodeImpl {



    /**
    * Get the number of children.
    */

    public int getNumberOfChildren() {
        int count = 0;
        for (int i=nodeNr+1; i<=document.numberOfNodes; i++) {
            if (document.depth[i]==document.depth[nodeNr]) {
                return count;
            }
            if (document.depth[i]==document.depth[nodeNr]+1) {
                count++;
            }
        }
        return count;
    }

    /**
    * Get all child nodes of the element (child elements and character nodes).
    * @return a vector containing a NodeInfo for each child node
    */

    public NodeInfo[] getAllChildNodes() {
        int count = getNumberOfChildren();
        NodeInfo[] kids = new NodeInfo[count];
        int k = 0;
        for (int i=nodeNr+1; i<=document.numberOfNodes; i++) {
            if (document.depth[i]==document.depth[nodeNr]) {
                return kids;
            }
            if (document.depth[i]==document.depth[nodeNr]+1) {
                TinyNodeImpl kid = document.getNode(i);
                kid.setParentNode(this);
                kids[k++] = kid;
            }
        }
        return kids;
    }

    /**
    * Get an enumeration of the children of this node
    */

    public NodeEnumeration enumerateChildren() {
        return new ChildEnumeration();
    }

    /**
    * Get the first child node of the element 
    * @return the first child node of the required type, or null if there are no children
    */

    public Node getFirstChild() {
        int next = nodeNr+1;
        if (document.depth[next]>document.depth[nodeNr]) {
            TinyNodeImpl n = document.getNode(next);
            n.setParentNode(this);
            return n;
        }
        return null;
    }

    /**
    * Get the last child node of the element 
    * @return the last child of the element, or null if there are no children
    */

    public Node getLastChild() {
        int last = -1;
        for (int i=nodeNr+1; i<=document.numberOfNodes; i++) {
            if (document.depth[i]==document.depth[nodeNr]) {
                if (last>0) {
                    TinyNodeImpl n = document.getNode(last);
                    n.setParentNode(this);
                    return n;
                } else {
                    return null;
                }
            }
            if (document.depth[i]==document.depth[nodeNr]+1) {
                last = i;
            }
        }
        return null;
    }

    /**
    * Return the string-value of the node, that is, the concatenation
    * of the character content of all descendent elements and text nodes. 
    * @return the accumulated character content of the element, including descendant elements.
    */

    public String getValue() {
        int level = document.depth[nodeNr];
        StringBuffer sb = null;

        // note, we can't rely on the value being contiguously stored because of whitespace
        // nodes: the data for these may still be present.
        
        int next = nodeNr+1;
        while (next < document.numberOfNodes && document.depth[next] > level) {
            if (document.nodeType[next]==NodeInfo.TEXT) {
                if (sb==null) {
                    sb = new StringBuffer();
                }
                int length = document.length[next];
                int start = document.offset[next];
                char[] dest = new char[length];
                document.charBuffer.getChars(
                        start, start+length, dest, 0);
                sb.append(dest, 0, length);
            }
            next++;
        }
        if (sb==null) return "";
        return sb.toString();
    }

    /**
    * Copy the string-value of this node to a given outputter
    */

    public void copyStringValue(Outputter out) throws SAXException {
        int level = document.depth[nodeNr];

        // note, we can't rely on the value being contiguously stored because of whitespace
        // nodes: the data for these may still be present.
        
        int next = nodeNr+1;
        while (next < document.numberOfNodes && document.depth[next] > level) {
            if (document.nodeType[next]==NodeInfo.TEXT) {
                out.writeContent(document.charBuffer, document.offset[next], document.length[next]);
            }
            next++;
        }
    }

    /**
    * Perform default action for this kind of node (built-in template rule)
    */

    public void defaultAction(Context c) throws SAXException {
        Mode mode = c.getMode();
	    c.getController().applyTemplates(c, null, mode, null);        
    }

    /**
    * Get the node value as defined in the DOM. This is not the same as the XPath string-value.
    */

    public String getNodeValue() {
        return null;
    }


    class ChildEnumeration implements NodeEnumeration, LastPositionFinder {

        NodeInfo next = null;

        public ChildEnumeration() {
            next = (NodeInfo)getFirstChild();
        }

        public boolean hasMoreElements() {
            return next != null;
        }

        public NodeInfo nextElement() {
            NodeInfo n = next;
            next = (NodeInfo)next.getNextSibling();
            return n;
        }

        public boolean isSorted() {
            return true;
        }

        public boolean isReverseSorted() {
            return false;
        }

        public boolean isPeer() {
            return true;           
        }

        public int getLastPosition() {
            return getNumberOfChildren();
        }
    }
}   


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
