package com.icl.saxon.tree;
import com.icl.saxon.om.*;

//import com.icl.saxon.handlers.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.output.*;
import com.icl.saxon.pattern.NameTest;
import com.icl.saxon.pattern.NamespaceTest;

import java.util.*;

import org.xml.sax.SAXException;
import org.w3c.dom.*;

/**
  * ElementImpl implements an element with no attributes or namespace declarations.<P>
  * This class is an implementation of ElementInfo. For elements with attributes or
  * namespace declarations, class ElementWithAttributes is used.
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  * @version 8 August 2000
  */

// The name of the element and its attributes are now namespace-resolved by the
// parser. However, this class retains the ability to do namespace resolution for other
// names, for example variable and template names in a stylesheet.

public class ElementImpl extends ParentNodeImpl
    implements ElementInfo, Element, NamedNodeMap {

    private static AttributeCollection emptyAtts = new AttributeCollection((NamePool)null);
    
    protected int nameCode;
    protected DocumentImpl root;

    /**
    * Construct an empty ElementImpl
    */

    public ElementImpl() {}
    
    /**
    * Set the name code. Used when creating a dummy element in the Stripper
    */
    
    public void setNameCode(int nameCode) {
    	this.nameCode = nameCode;
    }

    /**
    * Initialise a new ElementImpl with an element name
    * @param name The element name, with namespaces resolved
    * @param atts The attribute list: always null
    * @param parent The parent node    
    */

    public void initialise(int nameCode, AttributeCollection atts, NodeInfo parent,
                            String baseURI, int lineNumber, int sequenceNumber)
    throws SAXException {
        this.nameCode = nameCode;
        this.parent = (ParentNodeImpl)parent;
        this.sequence = sequenceNumber;
        this.root = (DocumentImpl)parent.getDocumentRoot();
        root.setLineNumber(sequenceNumber, lineNumber);
        root.setBaseURI(sequenceNumber, baseURI);
    }

	/**
	* Get the root node
	*/
	
	public DocumentInfo getDocumentRoot() {
		return root;
	}

    /**
    * Get the base URI for the node.
    */

    public final String getSystemId() {
        return ((DocumentImpl)getDocumentRoot()).getBaseURI(sequence);
    }

    /**
    * Set the line number of the element within its source document entity
    */

    public void setLineNumber(int line) {
        ((DocumentImpl)getDocumentRoot()).setLineNumber(sequence, line);
    }


    /**
    * Get the line number of the node within its source document entity
    */

    public int getLineNumber() {
        return ((DocumentImpl)getDocumentRoot()).getLineNumber(sequence);
    }


	/**
	* Get the nameCode of the node. This is used to locate the name in the NamePool
	*/
	
	public int getNameCode() {
		return nameCode;
	}


    /**
    * Search the NamespaceList for a given prefix, returning the corresponding URI.
    * @param prefix The prefix to be matched. To find the default namespace, supply ""
    * @return The URI code corresponding to this namespace. If it is an unnamed default namespace,
    * return Namespace.NULL_CODE. 
    * @throws SAXException if the prefix has not been declared on this NamespaceList.
    */

    public short getURICodeForPrefix(String prefix) throws SAXException {
    	System.err.println("GetURIfor prefix " + prefix + " in " + this);
        if (prefix.equals("xml")) return Namespace.XML_CODE;
        if (parent instanceof DocumentInfo) {
            if (prefix.equals("")) {
            	return Namespace.NULL_CODE;
            }
            System.err.println("Prefix " + prefix + " not found");            
            throw new SAXException("Namespace for prefix \"" + prefix + "\" has not been declared");
        } else {
            return ((ElementInfo)parent).getURICodeForPrefix(prefix);
        }
    }

    /**
    * Search the NamespaceList for a given URI, returning the corresponding prefix.
    * @param uri The URI to be matched. 
    * @return The prefix corresponding to this URI. If not found, return null. If there is
    * more than one prefix matching the URI, the first one found is returned. If the URI matches
    * the default namespace, return an empty string.
    */

    public String getPrefixForURI(String uri) {
        if (parent instanceof DocumentInfo) {
            return null;
        } else {
            return ((ElementInfo)parent).getPrefixForURI(uri);
        }
    }

    /**
    * Make a NameCode, using this Element as the context for namespace resolution.
    * The name will be entered in the namepool: therefore this method should not be
    * called once the name pool is sealed.
    * @param qname The name as written, in the form "[prefix:]localname"
    * @boolean useDefault Defines the action when there is no prefix. If true, use
    * the default namespace URI (as for element names). If false, use no namespace URI
    * (as for attribute names).
    */

    public final int makeNameCode(String qname, boolean useDefault) throws SAXException {

		NamePool namePool = getNamePool();
		String prefix = Name.getPrefix(qname);
        if (prefix.equals("")) {
            short uriCode = 0;
            
            if (useDefault) {
                uriCode = getURICodeForPrefix(prefix);
            }
            
			return namePool.allocate(prefix, uriCode, qname);

        } else {
            String localName = Name.getLocalName(qname);
            short uriCode = getURICodeForPrefix(prefix);
			return namePool.allocate(prefix, uriCode, localName);
        }

	}   	

    /**
    * Make the set of all namespace nodes associated with this element.
    * @param owner The element owning these namespace nodes.
    * @param list a Vector containing NamespaceInfo objects representing the namespaces
    * in scope for this element; the method appends nodes to this Vector, which should
    * initially be empty. Note that the returned list will never contain the XML namespace
    * (to get this, the NamespaceEnumeration class adds it itself). The list WILL include
    * an entry for the undeclaration xmlns=""; again it is the job of NamespaceEnumeration
    * to ignore this, since it doesn't represent a true namespace node.
    * @param addXML Add the XML namespace node to the list
    */

    public void addNamespaceNodes(ElementInfo owner, Vector list, boolean addXML) throws SAXException {
        // just add the namespaces defined on the ancestor nodes

        if (!(parent instanceof DocumentInfo)) {
            ((ElementInfo)parent).addNamespaceNodes(owner, list, false);
        }
        if (addXML) {
        	int nsxml = (1<<16) + 1;
            list.addElement(
                new NamespaceImpl(this, nsxml, list.size()+1)
                );
        }
    }
    
    /**
    * Output all namespace nodes associated with this element.
    * @param out The relevant outputter
    */

    public void outputNamespaceNodes(Outputter out, boolean includeAncestors) throws SAXException {

        // just add the namespaces defined on the ancestor nodes. We rely on the outputter
        // to eliminate multiple declarations of the same prefix
        
        if (includeAncestors) {
            if (!(parent instanceof DocumentInfo)) {
                ((ElementInfo)parent).outputNamespaceNodes(out, true);
            }
        }
    }
    


    /**
    * Return the type of node.
    * @return NodeInfo.ELEMENT
    */

    public final short getNodeType() {
        return ELEMENT;
    }
    
    /**
    * Get the attribute list for this element.
    * @return The attribute list. This will not include any
    * namespace attributes. The attribute names will be in expanded form, with prefixes
    * replaced by URIs
    */
    
    public ExtendedAttributes getAttributeList() {
        return emptyAtts;
    }
    
    /**
     *  Find the value of a given attribute of this element. <BR>
     *  This is a short-cut method; the full capability to examine
     *  attributes is offered via the getAttributeList() method. <BR>
     *  The attribute may either be one that was present in the original XML document,
     *  or one that has been set by the application using setAttribute(). <BR>
     *  @param name the name of an attribute. Any prefix in the name is interpreted
     *  in the context of the namespaces applying to the current element
     *  @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( Name name ) {
        return null;
    }

    /**
     *  Find the value of a given attribute of this element. <BR>
     *  This is a short-cut method; the full capability to examine
     *  attributes is offered via the getAttributeList() method. <BR>
     *  The attribute may either be one that was present in the original XML document,
     *  or one that has been set by the application using setAttribute(). <BR>
     *  @param name the name of an attribute. There must be no prefix in the name.
     *  @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String name ) {
        return null;
    }

    /**
    * Make an attribute node for a given attribute of this element
    * @param uri The attribute name's uri
    * @param localName The local name of the attribute
    */

    public AttributeInfo makeAttributeNode(int index) {
        return null;
    }

    /**
    * Make an attribute node for a given attribute fingerprint of this element
    * @param fingerprint The attribute name's fingerprint
    */

    public AttributeInfo makeAttributeNodeFS(int fingerprint) {
        return null;
    }
    
    /**
    * Set the value of an attribute on the current element.
    * @throws DOMException (always): the Saxon tree is immutable
    */
    
    public void setAttribute(String name, String value ) throws DOMException {
        disallowUpdate();
    }
       
    /**
    * Copy this node to a given outputter (supporting xsl:copy-of)
    */

    public void copy(Outputter out) throws SAXException {
        copy(out, true);
    }

    /**
    * Copy this node to a given outputter (supporting xsl:copy-of)
    * @param out The outputter
    * @param allNamespaces true if namespaces for ancestor nodes must be output
    */

    public void copy(Outputter out, boolean allNamespaces) throws SAXException {
    	int nc = getNameCode();
        out.writeStartTag(nc);

        // output the namespaces
        
        outputNamespaceNodes(out, allNamespaces);

        // output the children

        NodeInfo next = (NodeInfo)getFirstChild();
        while (next!=null) {
            if (next instanceof ElementImpl) {
                ((ElementImpl)next).copy(out, false);
            } else {
                next.copy(out);
            }
            next = (NodeInfo)next.getNextSibling();
        }

        out.writeEndTag(nc);
    }    

    /**
    * Generate a path to this node
    */

    public String getPath() {
        String pre = ((NodeInfo)getParentNode()).getPath();
        return (pre.equals("/") ? "" : pre) +
               "/" + getDisplayName() + "[" + Navigator.getNumberSimple(this) + "]";
    }

    ////////////////////////////////////////////////////////////////////////////
    // Following interfaces are provided to implement the DOM Element interface
    ////////////////////////////////////////////////////////////////////////////
    
    /**
     *  The name of the element (DOM interface). 
     */
     
    public String getTagName() {
        return getDisplayName();
    }

    /**
     *  The value of this node (as defined in DOM). Always null.
     */

    public String getNodeValue() {
        return null;
    }
    
    /**
     * Retrieves an attribute value by name. This is a DOM method, so namespace
     * declarations count as attributes.
     * @param name  The name of the attribute to retrieve.
     * @return  The <code>Attr</code> value as a string, or the empty string if
     *    that attribute does not have a specified or default value.
     */
     
    public String getAttribute(String name) {
        return "";
    }

    /**
     * A <code>NamedNodeMap</code> containing the attributes of this element. This
     * is a DOM method, so the list of attributes includes namespace declarations.
     */
     
    public NamedNodeMap getAttributes() {
        return this;
    }

    /**
     * Removes an attribute by name.
     * @param name  The name of the attribute to remove.
     */
     
    public void removeAttribute(String name) {
        setAttribute(name, null);
    }
    
    /**
     * Retrieves an attribute node by name. This is a DOM method, so namespace
     * declarations are treated as attributes.
     * <br> To retrieve an attribute node by qualified name and namespace URI, 
     * use the <code>getAttributeNodeNS</code> method.
     * @param name  The name (<code>nodeName</code> ) of the attribute to 
     *   retrieve.
     * @return  The <code>Attr</code> node with the specified name (
     *   <code>nodeName</code> ) or <code>null</code> if there is no such 
     *   attribute.
     */
     
    public Attr getAttributeNode(String name) {
        return null;
    }

    /**
     * Adds a new attribute node. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
     
    public Attr setAttributeNode(Attr newAttr) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     * Removes the specified attribute node. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */

    public Attr removeAttributeNode(Attr oldAttr) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     *  Returns a <code>NodeList</code> of all descendant <code>Elements</code>
     *  with a given tag name, in the order in which they are encountered in 
     * a preorder traversal of this <code>Element</code> tree.
     * @param name  The name of the tag to match on. The special value "*" 
     *   matches all tags.
     * @return  A list of matching <code>Element</code> nodes.
     */
     
    public NodeList getElementsByTagName(String name) {
        Vector v = new Vector();
        NodeInfo next = this;
        while(next!=null) {
            if (next instanceof ElementInfo) {
                if (name.equals("*") || name.equals(next.getDisplayName())) {
                    v.addElement(next);
                }
            }
            next = next.getNextInDocument(this);
        }
        return new NodeSetExtent(v);        
    }

    /**
     * Retrieves an attribute value by local name and namespace URI. 
     * HTML-only DOM implementations do not need to implement this method.
     * @param namespaceURI  The  namespace URI of the attribute to retrieve.
     * @param localName  The  local name of the attribute to retrieve.
     * @return  The <code>Attr</code> value as a string, or the empty string if
     *    that attribute does not have a specified or default value.
     * @since DOM Level 2
     */
     
    public String getAttributeNS(String namespaceURI, String localName) {
        return "";
    }                                           

    /**
     * Adds a new attribute. Always fails.
     * @param namespaceURI  The  namespace URI of the attribute to create or 
     *   alter.
     * @param qualifiedName  The  qualified name of the attribute to create or 
     *   alter.
     * @param value  The value to set in string form.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
     
    public void setAttributeNS(String namespaceURI, 
                               String qualifiedName, 
                               String value)
                               throws DOMException {
        disallowUpdate();
    }

    /**
     * Removes an attribute by local name and namespace URI. Always fails
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     * @since DOM Level 2
     */
     
    public void removeAttributeNS(String namespaceURI, 
                                  String localName)
                                  throws DOMException{
        disallowUpdate();
    }
    
    /**
     * Retrieves an <code>Attr</code> node by local name and namespace URI. 
     * DOM method, so namespace declarations count as attributes.
     * @param namespaceURI  The  namespace URI of the attribute to retrieve.
     * @param localName  The  local name of the attribute to retrieve.
     * @return  The <code>Attr</code> node with the specified attribute local 
     *   name and namespace URI or <code>null</code> if there is no such 
     *   attribute.
     * @since DOM Level 2
     */
     
    public Attr getAttributeNodeNS(String namespaceURI, String localName) {
        return null;
    }                                    

    /**
     * Add a new attribute. Always fails.
     * @param newAttr  The <code>Attr</code> node to add to the attribute list.
     * @return  If the <code>newAttr</code> attribute replaces an existing 
     *   attribute with the same  local name and  namespace URI , the 
     *   replaced <code>Attr</code> node is returned, otherwise 
     *   <code>null</code> is returned.
     * @exception DOMException
     *   <br> NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     * @since DOM Level 2
     */
     
    public Attr setAttributeNodeNS(Attr newAttr)
                                   throws DOMException{
        disallowUpdate();
        return null;
    }

    /**
     * Returns a <code>NodeList</code> of all the descendant 
     * <code>Elements</code> with a given local name and namespace URI in the 
     * order in which they are encountered in a preorder traversal of this 
     * <code>Element</code> tree.
     * @param namespaceURI  The  namespace URI of the elements to match on. 
     *   The special value "*" matches all namespaces.
     * @param localName  The  local name of the elements to match on. The 
     *   special value "*" matches all local names.
     * @return  A new <code>NodeList</code> object containing all the matched 
     *   <code>Elements</code> .
     * @since DOM Level 2
     */
     
    public NodeList getElementsByTagNameNS(String namespaceURI, String localName) {
        Vector v = new Vector();
        NodeInfo next = this;
        while(next!=null) {
            if (next instanceof ElementInfo) {
                if (namespaceURI.equals("*") || namespaceURI.equals(next.getURI()) &&
                    (localName.equals("*") || localName.equals(next.getLocalName()))) {
                    v.addElement(next);
                }
            }
            next = next.getNextInDocument(this);
        }
        return new NodeSetExtent(v);   
    }

    /**
     * Returns <code>true</code> when an attribute with a given name is 
     * specified on this element or has a default value, <code>false</code> 
     * otherwise. This is a DOM method, so namespace declarations are treated as
     * attributes.
     * @param name  The name of the attribute to look for.
     * @return <code>true</code> if an attribute with the given name is 
     *   specified on this element or has a default value, <code>false</code> 
     *   otherwise.
     * @since DOM Level 2
     */
     
    public boolean hasAttribute(String name) {
        return false;
    }

    /**
     * Returns <code>true</code> when an attribute with a given local name 
     * and namespace URI is specified on this element or has a default value, 
     * <code>false</code> otherwise. This is a DOM method so namespace declarations
     * are treated as attributes.
     * @param namespaceURI  The  namespace URI of the attribute to look for.
     * @param localName  The  local name of the attribute to look for.
     * @return <code>true</code> if an attribute with the given local name and 
     *   namespace URI is specified or has a default value on this element, 
     *   <code>false</code> otherwise.
     * @since DOM Level 2
     */
     
    public boolean hasAttributeNS(String namespaceURI, String localName) {
        return false;
    }    

    //////////////////////////////////////////////////////////////////////
    // Methods to implement DOM NamedNodeMap (the set of attributes)
    //////////////////////////////////////////////////////////////////////

    /**
    * Get named attribute (DOM NamedNodeMap method)
    * Treats namespace declarations as attributes.    
    */

    public Node getNamedItem(String name) {
        return null;
    }

    /**
    * Set named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node setNamedItem(Node arg) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Remove named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node removeNamedItem(String name) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Get n'th attribute (DOM NamedNodeMap method). Namespace declarations are treated
    * as attributes.
    */

    public Node item(int index) {
        return null;
    }

    /**
    * Get number of attributes (DOM NamedNodeMap method). 
    * Treats namespace declarations as attributes.
    */

    public int getLength() {
        return 0;
    }

    /**
    * Get named attribute (DOM NamedNodeMap method)
    * Treats namespace declarations as attributes.
    */

    public Node getNamedItemNS(String uri, String localName) {
        return null;
    }

    /**
    * Set named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node setNamedItemNS(Node arg) throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
    * Remove named attribute (DOM NamedNodeMap method: always fails)
    */

    public Node removeNamedItemNS(String uri, String localName) throws DOMException {
        disallowUpdate();
        return null;
    }
            
        
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
