package com.icl.saxon.tree;
import com.icl.saxon.om.*;
import com.icl.saxon.Context;
import com.icl.saxon.pattern.Pattern;
import com.icl.saxon.expr.NodeSetExtent;
import com.icl.saxon.expr.NodeEnumeration;
import com.icl.saxon.expr.SingletonEnumeration;
import com.icl.saxon.output.Outputter;

import java.util.Vector;
import java.util.Hashtable;
import org.xml.sax.AttributeList;
import org.xml.sax.SAXException;
import org.w3c.dom.*;



/**
  * A node in the XML parse tree representing an XML element, character content, or attribute.<P>
  * This is the top-level class in the implementation class hierarchy; it essentially contains
  * all those methods that can be defined using other primitive methods, without direct access
  * to data.
  * @author <A HREF="mailto:Michael.Kay@icl.com>Michael H. Kay, ICL</A> 
  */

abstract class NodeImpl implements NodeInfo {
    
    protected static NodeInfo[] emptyArray = new NodeInfo[0];
    protected static SingletonEnumeration emptyEnumeration = new SingletonEnumeration(null);

    protected ParentNodeImpl parent;
    protected int index;

    /**
    * Determine whether this is the same node as another node
    * @return true if this Node object and the supplied Node object represent the
    * same node in the tree.
    */

    public boolean isSameNode(NodeInfo other) {
        // default implementation: differs for attribute and namespace nodes
        return this==other;
    }

	/**
	* Get the nameCode of the node. This is used to locate the name in the NamePool
	*/
	
	public int getNameCode() {
		// default implementation: return -1 for an unnamed node
		return -1;
	}

	/**
	* Get the fingerprint of the node. This is used to compare whether two nodes
	* have equivalent names.
	*/
	
	public int getFingerprint() {
		return getNameCode() & 0xfffff;
	}

    /**
    * Get the base URL for the node. Default implementation for child nodes.
    */

    public String getSystemId() {
        return parent.getSystemId();
    }

    /**
    * Get the node sequence number (in document order). Sequence numbers are monotonic but not
    * consecutive. In the current implementation, parent nodes (elements and roots) have a zero
    * least-significant word, while namespaces, attributes, text nodes, comments, and PIs have
    * the top word the same as their owner and the bottom half reflecting their relative position.
    * This is the default implementation for child nodes.
    */

    public long getSequenceNumber() {
        NodeInfo prev = this;
        for (int i=0;; i++) {
            if (prev instanceof ParentNodeImpl) {
                return prev.getSequenceNumber() + 0x10000 + i;
                // note the 0x10000 is to leave room for namespace and attribute nodes.
            }
            prev = prev.getPreviousInDocument();
        }
        
    }

	/**
	* Get the NamePool
	*/
	
	public NamePool getNamePool() {
		return getDocumentRoot().getNamePool();
	}

    /**
    * Get the prefix part of the name of this node. This is the name before the ":" if any.
    * @return the prefix part of the name. For an unnamed node, return an empty string.
    */

    public String getPrefix() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return "";
    	if ((nameCode>>20 & 0xff) == 0) return "";
        return getNamePool().getPrefix(nameCode);
    }

    /**
    * Get the URI part of the name of this node. This is the URI corresponding to the
    * prefix, or the URI of the default namespace if appropriate.
    * @return The URI of the namespace of this node. For the default namespace, return an
    * empty string. For an unnamed node, return null.
    */

    public String getURI() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return null;
        return getNamePool().getURI(nameCode);
    }

	/**
	* Get URI code. This is a number that uniquely identifies the namespace URI of the
	* node, within its NamePool
	*/
	
	public short getURICode() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return -1;
        return getNamePool().getURICode(nameCode);
    }		

    /**
    * Get the name of this node, following the DOM rules 
    * @return The name of the node. For an element this is the element name, for an attribute
    * it is the attribute name, as a QName. Other node types return conventional names such
    * as "#text" or "#comment"
    */

    public String getNodeName() {
        // default implementation
        return getDisplayName();
    }

    
    /**
    * Get the display name of this node. For elements and attributes this is [prefix:]localname.
    * For unnamed nodes, it is an empty string.
    * @return The display name of this node. 
    * For a node with no name, return an empty string.
    */

    public String getDisplayName() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return "";
        return getNamePool().getDisplayName(nameCode);
    }

    /**
    * Get the local name of this node. 
    * @return The local name of this node. 
    * For a node with no name, return an empty string.
    */

    public String getLocalName() {
    	int nameCode = getNameCode();
    	if (nameCode == -1) return "";
        return getNamePool().getLocalName(nameCode);
    }
    
    /**
    * Get the line number of the node within its source document entity
    */

    public int getLineNumber() {
        return parent.getLineNumber();
    }



    /**
    * Get the public identifier of the document entity containing this node. This
    * is not currently maintained: return null
    */

    public String getPublicId() {
        return null;
    }

    /**
    * Get the index of this node, i.e. its position among its siblings
    */

    public final int getIndex() {
        return index;
    }

    /**
     * Find the parent node of this node.
     * @return The Node object describing the containing element or root node.
     */

    public final Node getParentNode()  {
        return parent;
    }

    /**
    * Get the previous sibling of the node
    * @return The previous sibling node. Returns null if the current node is the first
    * child of its parent.
    */
    
    public Node getPreviousSibling()  {
        return parent.getNthChild(index-1);
    }


   /**
    * Get next sibling node
    * @return The next sibling node of the required type. Returns null if the current node is the last
    * child of its parent.
    */

    public Node getNextSibling()  {
        return parent.getNthChild(index+1);
    }    

    /**
    * Get first child - default implementation used for leaf nodes
    * @return null
    */

    public Node getFirstChild()  {
        return null;
    }

    /**
    * Get last child - default implementation used for leaf nodes
    * @return null
    */

    public Node getLastChild()  {
        return null;
    }

    /**
    * Get the number of children.
    */

    public int getNumberOfChildren() {
        return 0;
    }

    /**
    * Get all child nodes of the node - default implementation used for leaf nodes
    * @return an empty array
    */

    public NodeInfo[] getAllChildNodes()  {
        return emptyArray;
    }

    /**
    * Get an enumeration of the children of this node
    */

    public NodeEnumeration enumerateChildren() {
        return emptyEnumeration;
    }

    /**
    * Determine whether the node is of a given type. <P>
    * Note, this can also be done by testing the node using "instanceof". But this is
    * inconvenient when passing the class as a parameter to another routine.
    * @param nodeType One of the specific node types such as ELEMENT or TEXT, or the general
    * node type NODE
    * @return true if the node is an instance of the specified node type
    */

    public final boolean isa(int nodeType) {
        return (nodeType==NODE || nodeType==getNodeType());
    }

    /**
     * Returns whether this node (if it is an element) has any attributes.
     * @return <code>true</code> if this node has any attributes, 
     *   <code>false</code> otherwise.
     * @since DOM Level 2
     */
    
    public boolean hasAttributes() {
        return false;
    }
    
    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param uri the namespace uri of an attribute
     * @param localName the local name of an attribute
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String uri, String localName ) {
        return null;
    }


    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param name the name of an attribute. This must be an unqualified attribute name,
     * i.e. one with no namespace prefix.
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue( String name ) {
        return null;
    }
    
    /**
    * Get the value of a given attribute of this node
    * @param fingerprint The fingerprint of the attribute name
    * @return the attribute value if it exists or null if not
    */
    
    public String getAttributeValue(int fingerprint) {
    	return null;
    }    

    /**
     * Determine whether this node is the outermost element.
     * @return True if this element is the document (outermost) element.
     */
     
    public boolean isDocumentElement() {
        return (this instanceof ElementInfo && getParentNode() instanceof DocumentInfo);
    }

    /**
     * Get the outermost element.
     * @return the ElementInfo for the outermost element of the document. If the document is
     * not well-formed, this returns the last element child of the root if there is one, otherwise
     * null.
     */
     
    public Element getDocumentElement() {
        return ((DocumentImpl)getDocumentRoot()).getDocumentElement();

    }

    /**
    * Get the root (document) node
    * @return the DocumentInfo representing the containing document
    */

    public DocumentInfo getDocumentRoot() {
        NodeImpl p = this;
        while (true) {
            p = (NodeImpl)p.getParentNode();
            if (p instanceof DocumentInfo) return (DocumentInfo)p;
        }
    }

    /**
    * Get the next node in document order
    * @param anchor: the scan stops when it reaches a node that is not a descendant of the specified
    * anchor node
    * @return the next node in the document, or null if there is no such node
    */

    public NodeInfo getNextInDocument(NodeInfo anchor) {
        // find the first child node if there is one; otherwise the next sibling node
        // if there is one; otherwise the next sibling of the parent, grandparent, etc, up to the anchor element.
        // If this yields no result, return null.

        NodeInfo next = (NodeInfo)getFirstChild();
        if (next!=null) return next;
        if (this==anchor) return null;
        next = (NodeInfo)getNextSibling();
        if (next!=null) return next;
        NodeInfo parent = this;
        while (true) {
            parent = (NodeInfo)parent.getParentNode();
            if (parent==null) return null;
            if (parent==anchor) return null;
            next = (NodeInfo)parent.getNextSibling();
            if (next!=null) return next;
        }
    }


    /**
    * Get the previous node in document order
    * @return the previous node in the document, or null if there is no such node
    */

    public NodeInfo getPreviousInDocument() {

        // finds the last child of the previous sibling if there is one;
        // otherwise the previous sibling element if there is one;
        // otherwise the parent, up to the anchor element.
        // If this reaches the document root, return null.

        NodeInfo prev = (NodeInfo)getPreviousSibling();
        if (prev!=null) {
            return ((NodeImpl)prev).getLastDescendantOrSelf();
        }
        return (NodeInfo)getParentNode();
    }

    private NodeInfo getLastDescendantOrSelf() {
        NodeImpl last = (NodeImpl)getLastChild();
        if (last==null) return this;
        return last.getLastDescendantOrSelf();
    }

    /**
    * Copy the string-value of this node to a given outputter
    */

    public void copyStringValue(Outputter out) throws SAXException {
        out.writeContent(getValue());   // default implementation
    }

    /**
    * Get a character string that uniquely identifies this node and that collates nodes
    * into document order
    * @return a string. 
    */

    public String getSequentialKey() {
        NodeInfo curr = this;
        StringBuffer key = new StringBuffer();
        while(!(curr instanceof DocumentInfo)) {
            key.insert(0, Navigator.alphaKey(curr.getIndex()));
            curr = (NodeInfo)curr.getParentNode();   
        }
        key.insert(0, curr.getSequentialKey());
        return key.toString();
    }



    /**
    * Create a string that identifies the node (for diagnostic purposes only)
    */

    public String toString() {
        String s;
        s = getDisplayName();
        if (s.equals("")) {
            NodeInfo n = (NodeInfo)getParentNode();
            if (n != null) s = n.getDisplayName() + "/";
            s += "\"" + getValue().trim() + "\"";
        }
        s = s + "(" + getIndex() + ")";
        return s;
    }

    /**
    * Remove this node from the tree. For system use only.
    * When one or more nodes have been removed, renumberChildren()
    * must be called to adjust the numbering of remaining nodes.
    * PRECONDITION: The node must have a parent node.
    */

    public void removeNode() throws SAXException {
        parent.removeChild(index);
    }


    // implement DOM Node methods

    /**
    * Get the node value as defined in the DOM. This is not the same as the XPath string-value.
    */

    public String getNodeValue() {
        // default implementation
        return getValue();
    }

    /**
    * Set the node value. DOM method: always fails
    */

    public void setNodeValue(String nodeValue) throws DOMException {
        disallowUpdate();
    }

    /**
     * Return a <code>NodeList</code> that contains all children of this node. If 
     * there are no children, this is a <code>NodeList</code> containing no 
     * nodes.
     */
     
    public NodeList getChildNodes() {
        return new NodeSetExtent(getAllChildNodes());
    }

    /**
     * Return a <code>NamedNodeMap</code> containing the attributes of this node (if 
     * it is an <code>Element</code> ) or <code>null</code> otherwise. (DOM method)
     */
     
    public NamedNodeMap getAttributes() {
        // default implementation
        return null;
    }

    /**
     * Return the <code>Document</code> object associated with this node. (DOM mehod)
     */
     
    public Document getOwnerDocument() {
        return (Document)(DocumentImpl)getDocumentRoot();
    }

    /**
     * Insert the node <code>newChild</code> before the existing child node 
     * <code>refChild</code>. DOM method: always fails.
     * @param newChild  The node to insert.
     * @param refChild  The reference node, i.e., the node before which the 
     *   new node must be inserted.
     * @return  The node being inserted.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node insertBefore(Node newChild, 
                             Node refChild)
                             throws DOMException {
        disallowUpdate();
        return null;
    }

    /**
     * Replace the child node <code>oldChild</code> with 
     * <code>newChild</code> in the list of children, and returns the 
     * <code>oldChild</code> node. Always fails.
     * @param newChild  The new node to put in the child list.
     * @param oldChild  The node being replaced in the list.
     * @return  The node replaced.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node replaceChild(Node newChild, 
                             Node oldChild)
                             throws DOMException{
        disallowUpdate();
        return null;
    }
    
    /**
     * Remove the child node indicated by <code>oldChild</code> from the 
     * list of children, and returns it. DOM method: always fails.
     * @param oldChild  The node being removed.
     * @return  The node removed.
     * @exception DOMException
     *    NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node removeChild(Node oldChild) throws DOMException {
        disallowUpdate();
        return null;
    }
    
    /**
     *  Adds the node <code>newChild</code> to the end of the list of children 
     * of this node. DOM method: always fails.
     * @param newChild  The node to add.
     * @return  The node added.
     * @exception DOMException
     *   <br> NO_MODIFICATION_ALLOWED_ERR: Always raised.
     */
     
    public Node appendChild(Node newChild) throws DOMException {
        disallowUpdate();
        return null;
    }
    
    /**
     * Determine whether the node has any children.
     * @return  <code>true</code> if the node has any children, 
     *   <code>false</code> if the node has no children.
     */
     
    public boolean hasChildNodes() {
        return getFirstChild() != null;         
    }

    /**
     * Returns a duplicate of this node, i.e., serves as a generic copy 
     * constructor for nodes. The duplicate node has no parent. Not
     * implemented: always returns null. (Because trees are read-only, there
     * would be no way of using the resulting node.)
     * @param deep  If <code>true</code> , recursively clone the subtree under 
     *   the specified node; if <code>false</code> , clone only the node 
     *   itself (and its attributes, if it is an <code>Element</code> ).  
     * @return  The duplicate node.
     */
     
    public Node cloneNode(boolean deep) {
        // Not implemented 
        return null;
    }

    /**
     * Puts all <code>Text</code> nodes in the full depth of the sub-tree 
     * underneath this <code>Node</code>, including attribute nodes, into a 
     * "normal" form where only structure (e.g., elements, comments, 
     * processing instructions, CDATA sections, and entity references) 
     * separates <code>Text</code> nodes, i.e., there are neither adjacent 
     * <code>Text</code> nodes nor empty <code>Text</code> nodes.
     * @since DOM Level 2
     */
     
    public void normalize() {
        // null operation; nodes are always normalized
    }

    /**
     *  Tests whether the DOM implementation implements a specific feature and 
     * that feature is supported by this node.
     * @param feature  The name of the feature to test. This is the same name 
     *   which can be passed to the method <code>hasFeature</code> on 
     *   <code>DOMImplementation</code> .
     * @param version  This is the version number of the feature to test. In 
     *   Level 2, version 1, this is the string "2.0". If the version is not 
     *   specified, supporting any version of the feature will cause the 
     *   method to return <code>true</code> .
     * @return  Returns <code>true</code> if the specified feature is supported
     *    on this node, <code>false</code> otherwise.
     * @since DOM Level 2
     */

    public boolean isSupported(String feature, String version) {
        return feature.equalsIgnoreCase("xml");
    }

    /**
     * The namespace URI of this node, or <code>null</code> if it is 
     * unspecified.
     * <br> This is not a computed value that is the result of a namespace 
     * lookup based on an examination of the namespace declarations in scope. 
     * It is merely the namespace URI given at creation time.
     * <br> For nodes of any type other than <code>ELEMENT_NODE</code> and 
     * <code>ATTRIBUTE_NODE</code> and nodes created with a DOM Level 1 
     * method, such as <code>createElement</code> from the 
     * <code>Document</code> interface, this is always <code>null</code> . 
     * Per the  Namespaces in XML Specification  an attribute does not 
     * inherit its namespace from the element it is attached to. If an 
     * attribute is not explicitly given a namespace, it simply has no 
     * namespace.
     * @since DOM Level 2
     */
     
    public String getNamespaceURI() {
        String uri = getURI();
        return (uri.equals("") ? null : uri);
    }

    /**
    * Set the namespace prefix of this node. Always fails.
    */
    
    public void setPrefix(String prefix)
                            throws DOMException {
        disallowUpdate();
    }

    /**
    * Internal method used to indicate that update operations are not allowed
    */

    protected void disallowUpdate() throws DOMException {
        throw new DOMExceptionImpl(DOMException.NO_MODIFICATION_ALLOWED_ERR,
                               "The Saxon DOM cannot be updated");
    }
    

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/ 
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License. 
//
// The Original Code is: all this file. 
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (Michael.Kay@icl.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved. 
//
// Contributor(s): none. 
//
